/*
 * Created on 5 juil. 2003
 *
 *   Copyright 2003 Alexandre Brunel
 * 
 *   This file is part of Ghanima beta version 1.
 * 
 *   Ghanima is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package NWNDialogEditorInterface;

import java.awt.*;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.*;
import java.awt.event.*;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.BreakIterator;
import java.text.MessageFormat;
import java.util.*;

import javax.swing.*;
import javax.swing.border.BevelBorder;
import javax.swing.event.*;
import javax.swing.filechooser.FileFilter;
import javax.swing.text.JTextComponent;
import javax.swing.tree.*;

import NWNFiles.GFFFile;
import NWNFiles.NWNFile;
import NWNFiles.StringListElement;
import NWNFiles.GFFFieldTypes.FieldType;
import NWNFiles.GFFFieldTypes.FieldTypeCExoLocString;
import NWNFiles.GFFFieldTypes.InvalidType;
import NWNFiles.GFFFieldTypes.FieldTypeCExoLocString.CExoLocString;
import NWNFiles.GFFFile.Field;
import NWNFiles.GFFFile.Structure;
import NWNFiles.NWNFile.FileState;
import NWNFiles.NWNFilesExceptions.*;
import NWNLanguages.GFFFieldNames;
import NWNLanguages.KeysHolder;
import NWNLanguages.Languages;

/**
 * @author Alexandre Brunel
 *
 * This is the mane window of the application.
 */
public class MainWindow extends JFrame implements ActionListener, WindowListener
{

    /**
     * 17 juil. 2003
     * @author Alexandre Brunel
     *
     * Clipboard handle.
     */
    protected class ClipBoardElement
    {
        private DefaultMutableTreeNode node;
        private EditWindow origin;

        public ClipBoardElement(EditWindow _origin, DefaultMutableTreeNode _node)
        {
            origin = _origin;
            node = _node;
        }
    }
    /**
     * @author Alexandre Brunel
     *
     * This object represents an edition field for a single dailog file.
     */
    public class EditWindow
        extends JPanel
        implements
            ActionListener,
            WindowListener,
            Observer,
            ItemListener,
            DocumentListener,
            TreeSelectionListener,
            FocusListener,
            DragSourceListener,
            DropTargetListener,
            DragGestureListener
    {
        /**
         * @author Alexandre Brunel
         *
         * Class used to represent a dialog node in the JTree.
         */
        protected class DialogTreeNode
        {
            private boolean isEntry;
            private boolean isLink;
            private GFFFile.Structure node;
            private GFFFile.Structure nodeRef;

            /**
             * Returns a new TreeNode.
             * @param _node node The dialog node
             * @param _nodeRef The structure pointing to the node.
             * @param _isEntry A boolean value which is true if the node is a speaker entry, false if it's a PC reply
             */
            public DialogTreeNode(
                GFFFile.Structure _node,
                GFFFile.Structure _nodeRef,
                boolean _isEntry)
            {

                node = _node;
                nodeRef = _nodeRef;

                GFFFile.Field isChild = nodeRef.getField(GFFFieldNames.IsChild);

                if (isChild != null && ((Integer) isChild.getData()).intValue() != 0)
                {
                    isLink = true;
                }
                else
                {
                    isLink = false;
                }

                isEntry = _isEntry;

            }

            /**
             * Counts the words in the english phrase.
             * 
             * @return Word count.
             */
            public long countWords()
            {
                long wordCount = 0;

                BreakIterator worditerator = BreakIterator.getWordInstance();

                FieldTypeCExoLocString.CExoLocString strings =
                    (CExoLocString) node.getField(GFFFieldNames.Text).getData();

                worditerator.setText(strings.getString(0, 0).getString());

                while (worditerator.next() != BreakIterator.DONE)
                {
                    wordCount++;
                }

                return wordCount;
            }

            /**
             * @return The dialog node.
             */
            public GFFFile.Structure getNode()
            {
                return node;
            }

            /**
             * @return The structure pointing to the node.
             */
            public GFFFile.Structure getNodeRef()
            {
                return nodeRef;
            }

            /**
             * @return Returns true if the node is a speaker entry, false if it's a PC reply.
             */
            public boolean isEntry()
            {
                return isEntry;
            }

            /**
             * @return true if this node is a link to another node.
             */
            public boolean isLink()
            {
                return isLink;
            }
            /**
             * Converts the object to a string.
             * 
             * This methode is used to display the node text in dialog tree. The returned String depends on
             * the selected language and gender in the {@link MainWindow.EditWindow EditWindow} that contains the dialog tree.
             * 
             * @see java.lang.Object#toString()
             * 
             * @return This methode returns a String containing the node text.  
             */
            public String toString()
            {
                GFFFile.Field textField = node.getField(GFFFieldNames.Text);

                if (textField != null)
                {
                    FieldTypeCExoLocString.CExoLocString text = (CExoLocString) textField.getData();

                    StringListElement loc;
                    if (getCurrentLanguageId() != 0)
                    {
                        loc = text.getString(getCurrentLanguageId(), getCurrentGender());
                    }
                    else
                    {
                        loc = text.getString(getCurrentLanguageId(), 0);
                    }

                    String suffix = ""; //$NON-NLS-1$
                    boolean end = false;
                    try
                    {
                        Vector replies;

                        if (isEntry)
                        {
                            replies = (Vector) node.getField(GFFFieldNames.RepliesList).getData();
                        }
                        else
                        {
                            replies = (Vector) node.getField(GFFFieldNames.EntriesList).getData();
                        }

                        if (replies.size() > 0)
                        {
                            suffix = currentLanguage.getString(KeysHolder.ContinueSuffix);
                        }
                        else
                        {
                            suffix = currentLanguage.getString(KeysHolder.EndSuffix);
                            end = true;
                        }
                    }
                    catch (NullPointerException e)
                    {
                        e.printStackTrace();
                        suffix = ""; //$NON-NLS-1$
                    }

                    if (loc != null)
                    {
                        if (loc.getString().equalsIgnoreCase("")) //$NON-NLS-1$
                        {
                            return suffix;
                        }
                        else if (end)
                        {
                            return loc.getString() + suffix;
                        }
                        else
                        {
                            return loc.getString();
                        }

                    }
                    else
                    {

                        loc = text.getString(0);
                        if (loc == null)
                        {
                            return currentLanguage.getString(KeysHolder.LanguageNotAvailable)
                                + suffix;
                        }
                        else if (loc.getString().equalsIgnoreCase("")) //$NON-NLS-1$
                        {
                            return suffix;
                        }
                        else if (end)
                        {
                            return currentLanguage.getString(KeysHolder.LanguageNotAvailable)
                                + loc.getString()
                                + suffix;
                        }
                        else
                        {
                            return currentLanguage.getString(KeysHolder.LanguageNotAvailable)
                                + loc.getString();
                        }
                    }
                }
                else
                {
                    return ""; //$NON-NLS-1$
                }

            }

        }

        /**
         * 16 juil. 2003
         * @author Alexandre Brunel
         *
         * This class handles the undo/redo.
         */
        protected class FieldChangeUndo implements UndoRedo
        {

            private String fieldLabel;

            private DialogTreeNode node;

            private Object oldValue;

            private JComponent viewComponent;

            /**
             * Constructs a FieldChangeUndo object.
             * 
             * @param _node The node containing the field.
             * @param _fieldLabel The label of the field. 
             * @param _oldValue The old value of the field.
             */
            public FieldChangeUndo(
                DialogTreeNode _node,
                String _fieldLabel,
                Object _oldValue,
                JComponent _viewComponent)
            {

                node = _node;
                fieldLabel = _fieldLabel;
                oldValue = _oldValue;
                viewComponent = _viewComponent;

            }

            private void doAction()
            {
                GFFFile.Field field;

                if (node == null)
                {
                    field = nwnDlgFile.getStruct(0).getField(fieldLabel);

                }
                else
                {
                    field = node.getNode().getField(fieldLabel);

                    if (field == null)
                    {
                        field = node.getNodeRef().getField(fieldLabel);
                    }

                }

                Object temp = field.getData();
                field.setData(oldValue);

                if (viewComponent != null)
                {

                    try
                    {
                        if (Class.forName("javax.swing.text.JTextComponent") //$NON-NLS-1$
                        .isInstance(viewComponent))
                        {
                            ((JTextComponent) viewComponent).setText((String) oldValue);
                        }
                        else if (Class.forName("javax.swing.JComboBox").isInstance(viewComponent)) //$NON-NLS-1$
                        {
                            ((JComboBox) viewComponent).setSelectedIndex(
                                ((Number) oldValue).intValue());
                        }
                        else if (Class.forName("javax.swing.JCheckBox").isInstance(viewComponent)) //$NON-NLS-1$
                        {
                            ((JCheckBox) viewComponent).setSelected(
                                ((Number) oldValue).intValue() != 0);
                        }
                    }
                    catch (ClassNotFoundException e)
                    {

                        e.printStackTrace();
                    }
                }
                else if (fieldLabel.equals(GFFFieldNames.Text))
                {
                    long oldWordCount = 0;
                    long newWorldcount = 0;

                    BreakIterator worditerator = BreakIterator.getWordInstance();

                    FieldTypeCExoLocString.CExoLocString oldStrings = (CExoLocString) oldValue;
                    FieldTypeCExoLocString.CExoLocString newStrings = (CExoLocString) temp;

                    worditerator.setText(oldStrings.getString(0, 0).getString());

                    while (worditerator.next() != BreakIterator.DONE)
                    {
                        oldWordCount++;
                    }

                    worditerator.setText(newStrings.getString(0, 0).getString());

                    while (worditerator.next() != BreakIterator.DONE)
                    {
                        newWorldcount++;
                    }

                    nwnDlgFile.getStruct(0).getField(GFFFieldNames.NumWords).setData(
                        new Long(
                            ((Number) nwnDlgFile
                                .getStruct(0)
                                .getField(GFFFieldNames.NumWords)
                                .getData())
                                .longValue()
                                - oldWordCount
                                + newWorldcount));
                }
                oldValue = temp;
            }

            /* (non-Javadoc)
             * @see NWNDialogEditorInterface.MainWindow.UndoRedo#redo()
             */
            public void redo()
            {
                doAction();
                modifyUndo(this);
            }

            /* (non-Javadoc)
             * @see java.lang.Object#toString()
             */
            public String toString()
            {
                Object[] args = { fieldLabel };
                return MessageFormat.format(
                    currentRessouceBundle.getString(KeysHolder.FieldModificationFormat),
                    args);
            }

            /* (non-Javadoc)
             * @see NWNDialogEditorInterface.MainWindow.UndoRedo#undo()
             */
            public void undo()
            {
                doAction();
                modifyRedo(this);
            }

        }
        /**
         * 10 aot 2003
         * @author Alexandre Brunel
         *
         * TODO NWNDialogEditorInterface.MovingNode documentation
         */
        protected class MovingNode implements Transferable
        {

            private DataFlavor[] dataFlavors;

            private DefaultMutableTreeNode node;

            public MovingNode(DefaultMutableTreeNode data)
            {
                node = data;

                dataFlavors = MovingNodeDataFlavors;
            }
            public DefaultMutableTreeNode getTransferData()
            {

                return node;
            }

            /* (non-Javadoc)
             * @see java.awt.datatransfer.Transferable#getTransferData(java.awt.datatransfer.DataFlavor)
             */
            public Object getTransferData(DataFlavor flavor)
                throws UnsupportedFlavorException, IOException
            {

                if (TreeNodeDataFlavor.equals(flavor))
                {
                    return node;
                }
                else if (DataFlavor.stringFlavor.equals(flavor))
                {
                    return node.toString();
                }
                else if (TreeNodePathDataFlavor.equals(flavor))
                {

                    return new TreePath(node.getPath());
                }
                else
                {
                    throw new UnsupportedFlavorException(flavor);
                }

            }
            /* (non-Javadoc)
             * @see java.awt.datatransfer.Transferable#getTransferDataFlavors()
             */
            public DataFlavor[] getTransferDataFlavors()
            {

                return dataFlavors;
            }

            /* (non-Javadoc)
             * @see java.awt.datatransfer.Transferable#isDataFlavorSupported(java.awt.datatransfer.DataFlavor)
             */
            public boolean isDataFlavorSupported(DataFlavor flavor)
            {

                return dataFlavors[0].equals(flavor) || dataFlavors[1].equals(flavor);
            }

        }
        /**
         * 2 aot 2003
         * @author Alexandre Brunel
         *
         * TODO NWNDialogEditorInterface.NodeMoover documentation
         */
        protected class NodeMoover extends TransferHandler
        {

            /**
             * TODO NWNDialogEditorInterface.NodeMoover Constructor documentation.
             * Constructs a NodeMoover object.
             * 
             */
            public NodeMoover()
            {
                super();
                // TODO Auto-generated constructor NodeMoover
            }

            /* (non-Javadoc)
             * @see javax.swing.TransferHandler#canImport(javax.swing.JComponent, java.awt.datatransfer.DataFlavor[])
             */
            public boolean canImport(JComponent comp, DataFlavor[] transferFlavors)
            {

                return true;
            }

            /* (non-Javadoc)
             * @see javax.swing.TransferHandler#createTransferable(javax.swing.JComponent)
             */
            protected Transferable createTransferable(JComponent c)
            {
                // TODO Auto-generated method createTransferable
                return super.createTransferable(c);
            }

            /* (non-Javadoc)
             * @see javax.swing.TransferHandler#exportAsDrag(javax.swing.JComponent, java.awt.event.InputEvent, int)
             */
            public void exportAsDrag(JComponent comp, InputEvent e, int action)
            {
                // TODO Auto-generated method exportAsDrag
                super.exportAsDrag(comp, e, action);
            }

            /* (non-Javadoc)
             * @see javax.swing.TransferHandler#exportDone(javax.swing.JComponent, java.awt.datatransfer.Transferable, int)
             */
            protected void exportDone(JComponent source, Transferable data, int action)
            {
                // TODO Auto-generated method exportDone
                super.exportDone(source, data, action);
            }

        }
        /**
         * 15 juil. 2003
         * @author Alexandre Brunel
         *
         * This class allows to edit a dialog tree.
         */
        protected class TreeEditor implements TreeCellEditor, KeyListener
        {

            private boolean canceled;

            private int cellRow;

            private JPanel editingField;

            private JTree editingTree;

            private DefaultMutableTreeNode editingValue;

            private JTextArea editingZone;

            private Vector listeners;

            private JScrollPane scrollPane;

            /**
             * Constructs a TreeEditor object.
             * 
             */
            public TreeEditor()
            {
                editingField = null;

                listeners = new Vector();
            }

            /* (non-Javadoc)
             * @see javax.swing.CellEditor#addCellEditorListener(javax.swing.event.CellEditorListener)
             */
            public void addCellEditorListener(CellEditorListener l)
            {
                listeners.add(l);
                listenerList.add(l.getClass(), l);
            }

            /* (non-Javadoc)
             * @see javax.swing.CellEditor#cancelCellEditing()
             */
            public void cancelCellEditing()
            {
                canceled = true;
                Iterator i = listeners.iterator();

                while (i.hasNext())
                {
                    CellEditorListener l = (CellEditorListener) i.next();

                    l.editingCanceled(new ChangeEvent(this));

                }

            }

            /* (non-Javadoc)
             * @see javax.swing.CellEditor#getCellEditorValue()
             */
            public Object getCellEditorValue()
            {
                DialogTreeNode node = (DialogTreeNode) editingValue.getUserObject();

                return node;
            }

            /* (non-Javadoc)
             * @see javax.swing.tree.TreeCellEditor#getTreeCellEditorComponent(javax.swing.JTree, java.lang.Object, boolean, boolean, boolean, int)
             */
            public Component getTreeCellEditorComponent(
                JTree tree,
                Object value,
                boolean isSelected,
                boolean expanded,
                boolean leaf,
                int row)
            {
                editingTree = tree;
                editingValue = (DefaultMutableTreeNode) value;

                cellRow = row;

                if (!expanded)
                {
                    tree.expandRow(row);
                }

                canceled = false;

                tree.setRowHeight(0);

                FieldTypeCExoLocString.CExoLocString strings =
                    (CExoLocString) ((DialogTreeNode) editingValue.getUserObject())
                        .getNode()
                        .getField(GFFFieldNames.Text)
                        .getData();

                StringListElement temp =
                    strings.getString(getCurrentLanguageId(), getCurrentGender());

                String text = null;

                if (temp != null)
                {
                    text = temp.getString();
                }

                if (editingField == null)
                {

                    editingField = new JPanel();
                    editingZone = new JTextArea(text);
                    scrollPane = new JScrollPane(editingField);

                    editingField.add(editingZone);
                    editingField.setBackground(editingZone.getBackground());

                    editingZone.addKeyListener(this);
                    editingZone.setRows(3);
                    editingZone.invalidate();

                }
                else
                {
                    editingZone.setText(text);
                    //editingZone.setRows((editingZone.getRows() > 2 ? editingZone.getRows():2));

                }

                return scrollPane;
            }

            /* (non-Javadoc)
             * @see javax.swing.CellEditor#isCellEditable(java.util.EventObject)
             */
            public boolean isCellEditable(EventObject anEvent)
            {
                JTree test = (JTree) anEvent.getSource();

                TreePath nodePath = test.getSelectionPath();

                if (nodePath != null && nodePath.getPathCount() > 1)
                {
                    try
                    {
                        if (Class.forName("java.awt.event.MouseEvent").isInstance(anEvent) //$NON-NLS-1$
                        && ((MouseEvent) anEvent).isPopupTrigger())
                        {
                            return false;
                        }
                        else if (((MouseEvent) anEvent).getClickCount() == 2)
                        {
                            if (nodePath
                                .equals(
                                    test.getPathForLocation(
                                        ((MouseEvent) anEvent).getX(),
                                        ((MouseEvent) anEvent).getY()))
                                && !((DialogTreeNode) ((DefaultMutableTreeNode) nodePath
                                    .getLastPathComponent())
                                    .getUserObject())
                                    .isLink())
                            {
                                return true;
                            }
                        }
                        else
                        {
                            return false;
                        }
                    }
                    catch (ClassNotFoundException e)
                    {

                    }

                }

                return false;
            }

            /* (non-Javadoc)
             * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
             */
            public void keyPressed(KeyEvent e)
            {

            }

            /* (non-Javadoc)
             * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
             */
            public void keyReleased(KeyEvent e)
            {
                if (e.getSource() == editingZone
                    && e.getKeyCode() == KeyEvent.VK_ENTER
                    && ((e.getModifiers() == (InputEvent.CTRL_DOWN_MASK | InputEvent.CTRL_MASK))
                        || (e.getModifiers() == InputEvent.CTRL_DOWN_MASK)
                        || (e.getModifiers() == InputEvent.CTRL_MASK)))
                {
                    editingTree.stopEditing();
                }

            }

            /* (non-Javadoc)
             * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
             */
            public void keyTyped(KeyEvent e)
            {

            }

            /* (non-Javadoc)
             * @see javax.swing.CellEditor#removeCellEditorListener(javax.swing.event.CellEditorListener)
             */
            public void removeCellEditorListener(CellEditorListener l)
            {
                listeners.remove(l);
                listenerList.remove(l.getClass(), l);

            }

            /* (non-Javadoc)
             * @see javax.swing.CellEditor#shouldSelectCell(java.util.EventObject)
             */
            public boolean shouldSelectCell(EventObject anEvent)
            {

                return true;
            }

            /* (non-Javadoc)
             * @see javax.swing.CellEditor#stopCellEditing()
             */
            public boolean stopCellEditing()
            {

                DialogTreeNode node = (DialogTreeNode) editingValue.getUserObject();

                if (!canceled)
                {
                    GFFFile.Field field = node.getNode().getField(GFFFieldNames.Text);

                    if (field != null)
                    {
                        FieldTypeCExoLocString.CExoLocString string =
                            (CExoLocString) field.getData();

                        modifyUndo(new FieldChangeUndo(node, GFFFieldNames.Text, string, null));

                        FieldTypeCExoLocString.CExoLocString newString;

                        if (string != null)
                        {
                            long language = getCurrentLanguageId();
                            int gender = getCurrentGender();

                            long wordCount = 0;

                            BreakIterator worditerator = BreakIterator.getWordInstance();

                            worditerator.setText(editingZone.getText());

                            while (worditerator.next() != BreakIterator.DONE)
                            {
                                wordCount++;
                            }

                            StringListElement temp =
                                ((FieldTypeCExoLocString.CExoLocString) field.getData()).getString(
                                    getCurrentLanguageId(),
                                    getCurrentGender());

                            long oldwordCount = 0;

                            if (temp != null)
                            {
                                worditerator.setText(temp.getString());

                                while (worditerator.next() != BreakIterator.DONE)
                                {
                                    oldwordCount++;
                                }
                            }

                            if (getCurrentLanguageId() == 0)
                            {
                                nwnDlgFile.getStruct(0).getField(GFFFieldNames.NumWords).setData(
                                    new Long(
                                        ((Number) nwnDlgFile
                                            .getStruct(0)
                                            .getField(GFFFieldNames.NumWords)
                                            .getData())
                                            .longValue()
                                            - oldwordCount
                                            + wordCount));
                            }
                            StringListElement newValue = null;

                            try
                            {
                                newValue =
                                    new StringListElement(
                                        editingZone.getText(),
                                        (int) (2 * language + gender));
                            }
                            catch (InvalidLanguageId e)
                            {
                                try
                                {
                                    newValue = new StringListElement(editingZone.getText(), 0);
                                }
                                catch (InvalidLanguageId e1)
                                {

                                    e1.printStackTrace();
                                }
                                e.printStackTrace();
                            }

                            StringListElement[] result;
                            if (temp == null)
                            {
                                result = new StringListElement[string.getStringCount() + 1];

                            }
                            else
                            {
                                result = new StringListElement[string.getStringCount()];
                            }

                            for (int i = 0; i < string.getStringCount(); ++i)
                            {
                                StringListElement cur = string.getString(i);

                                if (cur.getGender() == gender && cur.getLanguageID() == language)
                                {
                                    result[i] = newValue;
                                }
                                else
                                {
                                    result[i] = cur;
                                }
                            }

                            if (temp == null)
                            {
                                result[string.getStringCount()] = newValue;
                            }

                            newString = new CExoLocString(string.getStringRef(), result);

                            field.setData(newString);
                        }
                        else
                        {
                            long language = getCurrentLanguageId();
                            int gender = getCurrentGender();

                            long wordCount = 0;

                            BreakIterator worditerator = BreakIterator.getWordInstance();

                            worditerator.setText(editingZone.getText());

                            if (getCurrentLanguageId() == 0)
                            {
                                while (worditerator.next() != BreakIterator.DONE)
                                {
                                    wordCount++;
                                }

                                nwnDlgFile.getStruct(0).getField(GFFFieldNames.NumWords).setData(
                                    new Long(
                                        ((Number) nwnDlgFile
                                            .getStruct(0)
                                            .getField(GFFFieldNames.NumWords)
                                            .getData())
                                            .longValue()
                                            + wordCount));
                            }
                            StringListElement newValue = null;

                            try
                            {
                                newValue =
                                    new StringListElement(
                                        editingZone.getText(),
                                        (int) (2 * language + gender));
                            }
                            catch (InvalidLanguageId e)
                            {
                                try
                                {
                                    newValue = new StringListElement(editingZone.getText(), 0);
                                }
                                catch (InvalidLanguageId e1)
                                {

                                    e1.printStackTrace();
                                }
                                e.printStackTrace();
                            }

                            StringListElement[] result = { newValue };

                            newString = new CExoLocString(0, result);

                            field.setData(newString);
                        }
                    }
                    else
                    {

                        long language = getCurrentLanguageId();
                        int gender = getCurrentGender();

                        StringListElement newValue = null;
                        try
                        {
                            newValue =
                                new StringListElement(
                                    editingZone.getText(),
                                    (int) (2 * language + gender));
                        }
                        catch (InvalidLanguageId e)
                        {
                            try
                            {
                                newValue = new StringListElement(editingZone.getText(), 0);
                            }
                            catch (InvalidLanguageId e1)
                            {

                                e1.printStackTrace();
                            }
                            e.printStackTrace();
                        }

                        StringListElement[] result = { newValue };

                        FieldTypeCExoLocString.CExoLocString newString =
                            new CExoLocString(0, result);

                        try
                        {
                            node.getNode().addField(
                                FieldType.CExoLocString_TYPE,
                                GFFFieldNames.Text,
                                newString);
                        }
                        catch (NameTooLong e1)
                        {

                            e1.printStackTrace();
                        }
                    }
                }

                Iterator i = listeners.iterator();
                while (i.hasNext())
                {
                    CellEditorListener l = (CellEditorListener) i.next();

                    l.editingStopped(new ChangeEvent(this));

                }

                return true;
            }

        }
        private JTextField abortScriptField;

        private JLabel abortScriptLabel;
        private JButton abortScriptPick;
        private JButton activationScriptButton;
        private JTextField activationScriptField;
        private JLabel activationScriptLabel;
        private String[] animationComboList;
        private JLabel animationLabel;
        private JCheckBox animationLoop;
        private JCheckBox cameraZoom;
        private JTextArea commentArea;
        private JTree dialog;
        private File dialogFile;
        private Vector entryList;
        private JTextField exitScriptField;
        private JLabel exitScriptLabel;
        private JButton exitScriptPick;
        private JComboBox genderComboBox;
        private JLabel genderLabel;
        private JTextField journalEntryName;
        private JLabel journalEntryNameLabel;
        private JTextField journalEntryNumber;
        private JLabel journalEntryNumberLabel;
        private JComboBox languageComboBox;
        private JLabel languageLabel;
        private Languages languagesHandler;
        private boolean modified;
        private GFFFile nwnDlgFile;
        private JPanel options;
        private MainWindow owner;
        private JButton pickSpeakerButton;
        protected Stack redoStack;
        private JComboBox replyAnimation;
        private Vector replyList;
        private JTextField replySound;
        private DefaultMutableTreeNode rootNode;
        private JLabel soundLabel;
        private JTextField speakerEditField;
        private JLabel speakersTagLabel;
        private Vector startingList;
        private JTextField textAppearWhenField;
        private JLabel textAppearWhenLabel;
        private JButton textAppearWhenSelectButton;

        protected Stack undoStack;

        /**
         * 
         */
        public EditWindow(MainWindow _owner)
            throws
                FileNotFoundException,
                FileDoesNotExist,
                InvalidFile,
                NameTooLong,
                IOException,
                InvalidType
        {
            super(new GridLayout(2, 1), true);
            dialogFile = null;
            owner = _owner;
            init();
        }

        /**
         * @param isDoubleBuffered
         */
        public EditWindow(MainWindow _owner, boolean isDoubleBuffered)
            throws
                FileNotFoundException,
                FileDoesNotExist,
                InvalidFile,
                NameTooLong,
                IOException,
                InvalidType
        {
            super(new GridLayout(2, 1), isDoubleBuffered);
            dialogFile = null;
            owner = _owner;
            init();
        }

        /**
         * 
         */
        public EditWindow(MainWindow _owner, File _dialogFile)
            throws
                FileNotFoundException,
                FileDoesNotExist,
                InvalidFile,
                NameTooLong,
                IOException,
                InvalidType
        {
            super(new GridLayout(2, 1), true);
            dialogFile = _dialogFile;
            owner = _owner;
            init();
        }

        /**
         * @param layout
         */
        public EditWindow(MainWindow _owner, LayoutManager layout)
            throws
                FileNotFoundException,
                FileDoesNotExist,
                InvalidFile,
                NameTooLong,
                IOException,
                InvalidType
        {
            super(layout, true);
            dialogFile = null;
            owner = _owner;
            init();
        }

        /**
        * @param layout
        * @param isDoubleBuffered
        */
        public EditWindow(MainWindow _owner, LayoutManager layout, boolean isDoubleBuffered)
            throws
                FileNotFoundException,
                FileDoesNotExist,
                InvalidFile,
                NameTooLong,
                IOException,
                InvalidType
        {
            super(layout, isDoubleBuffered);
            dialogFile = null;
            owner = _owner;
            init();
        }

        /* (non-Javadoc)
         * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
         */
        public synchronized void actionPerformed(ActionEvent e)
        {
            if (e.getSource() == cameraZoom)
            {
                GFFFile.Structure struct = nwnDlgFile.getStruct(0);

                GFFFile.Field field = struct.getField(GFFFieldNames.PreventZoomIn);

                if ((cameraZoom.isSelected() && ((Number) field.getData()).intValue() != 0)
                    || (!cameraZoom.isSelected() && ((Number) field.getData()).intValue() == 0))
                {
                    modifyUndo(
                        new FieldChangeUndo(
                            null,
                            GFFFieldNames.PreventZoomIn,
                            field.getData(),
                            cameraZoom));

                    field.setData(new Integer((cameraZoom.isSelected() ? 1 : 0)));
                }
            }
            else if (e.getSource() == animationLoop)
            {
                DefaultMutableTreeNode treeNode =
                    (DefaultMutableTreeNode) dialog.getSelectionPath().getLastPathComponent();

                if (!treeNode.isRoot())
                {
                    DialogTreeNode node = (DialogTreeNode) treeNode.getUserObject();

                    GFFFile.Structure struct = node.getNode();

                    GFFFile.Field field = struct.getField(GFFFieldNames.AnimLoop);

                    if ((cameraZoom.isSelected() && ((Number) field.getData()).intValue() != 0)
                        || (!cameraZoom.isSelected() && ((Number) field.getData()).intValue() == 0))
                    {
                        modifyUndo(
                            new FieldChangeUndo(
                                null,
                                GFFFieldNames.AnimLoop,
                                field.getData(),
                                animationLoop));

                        field.setData(new Integer((cameraZoom.isSelected() ? 1 : 0)));
                    }
                }
            }
        }

        /**
         * Adds a node to the selected node.
         */
        protected synchronized void addNode()
        {
            DefaultMutableTreeNode node =
                (DefaultMutableTreeNode) dialog.getSelectionPath().getLastPathComponent();

            GFFFile.Structure structure = null;

            String fieldLabel;

            boolean isEntry = false;

            if (!node.isRoot())
            {
                DialogTreeNode treeNode = (DialogTreeNode) node.getUserObject();

                if (isEntry = treeNode.isEntry)
                    fieldLabel = GFFFieldNames.RepliesList;
                else
                    fieldLabel = GFFFieldNames.EntriesList;

                if (!treeNode.isLink())
                {
                    structure = treeNode.getNode();

                }
            }
            else
            {
                structure = nwnDlgFile.getStruct(0);
                fieldLabel = GFFFieldNames.StartingList;
            }

            if (structure != null)
            {
                Vector child = (Vector) structure.getField(fieldLabel).getData();

                Vector newNode = createNewNode(!isEntry);

                child.add(newNode.get(1));

                DialogTreeNode newTreeNode =
                    new DialogTreeNode(
                        (GFFFile.Structure) newNode.get(0),
                        (GFFFile.Structure) newNode.get(1),
                        !isEntry);

                if (isEntry)
                {
                    replyList.add(newTreeNode);
                }
                else
                {
                    entryList.add(newTreeNode);
                }

                //node.add(new DefaultMutableTreeNode(newTreeNode));
                ((DefaultTreeModel) dialog.getModel()).insertNodeInto(
                    new DefaultMutableTreeNode(newTreeNode),
                    node,
                    node.getChildCount());
            }
        }

        /* (non-Javadoc)
         * @see javax.swing.event.DocumentListener#changedUpdate(javax.swing.event.DocumentEvent)
         */
        public void changedUpdate(DocumentEvent e)
        {
            updateText(e);

        }

        /**
         * Ferfomes some cleaning operations.
         * @return Returns true if the edition tab can be closed. 
         */
        protected synchronized boolean close()
        {

            speakerEditField.getDocument().removeDocumentListener(this);

            cameraZoom.removeActionListener(this);
            abortScriptPick.removeActionListener(this);

            abortScriptField.getDocument().removeDocumentListener(this);

            exitScriptPick.removeActionListener(this);

            exitScriptField.getDocument().removeDocumentListener(this);

            journalEntryNumber.getDocument().removeDocumentListener(this);

            journalEntryName.getDocument().removeDocumentListener(this);

            replySound.removeActionListener(this);
            animationLoop.removeActionListener(this);

            replyAnimation.removeItemListener(this);
            genderComboBox.removeItemListener(this);
            languageComboBox.removeItemListener(this);

            activationScriptButton.removeActionListener(this);

            activationScriptField.getDocument().removeDocumentListener(this);

            textAppearWhenSelectButton.removeActionListener(this);

            textAppearWhenField.getDocument().removeDocumentListener(this);

            pickSpeakerButton.removeActionListener(this);

            speakerEditField.getDocument().removeDocumentListener(this);

            nwnDlgFile.deleteObserver(this);

            rootNode.removeAllChildren();

            dialog.setCellEditor(null);

            return true;
        }

        /**
         * 
         */
        public synchronized void copy()
        {
            TreePath path = dialog.getSelectionPath();

            DefaultMutableTreeNode treeNode;

            if (path != null
                && !(treeNode = (DefaultMutableTreeNode) path.getLastPathComponent()).isRoot())
            {
                DialogTreeNode node = (DialogTreeNode) treeNode.getUserObject();

                //menuEditPaste.setEnabled(true);

                if (node.isLink())
                {
                    //menuEditPasteLink.setEnabled(false);
                    clipboard = new ClipBoardElement(this, treeNode);

                }
                else
                {
                    //menuEditPasteLink.setEnabled(true);
                    DefaultMutableTreeNode newNode = (DefaultMutableTreeNode) treeNode.clone();

                    clipboard = new ClipBoardElement(this, newNode);

                    Vector currentNodes;
                    Vector nextNodes = new Vector(1);
                    Vector currentCoppy;
                    Vector nextCopy = new Vector(1);

                    nextNodes.add(treeNode);
                    nextCopy.add(newNode);

                    while (!nextNodes.isEmpty())
                    {
                        currentCoppy = nextCopy;
                        currentNodes = nextNodes;

                        Iterator i = currentNodes.iterator();
                        Iterator j = currentCoppy.iterator();

                        nextCopy = new Vector();
                        nextNodes = new Vector();

                        while (i.hasNext())
                        {
                            DefaultMutableTreeNode original = (DefaultMutableTreeNode) i.next();
                            DefaultMutableTreeNode copy = (DefaultMutableTreeNode) j.next();

                            Enumeration enum = original.children();

                            while (enum.hasMoreElements())
                            {
                                DefaultMutableTreeNode element =
                                    (DefaultMutableTreeNode) enum.nextElement();

                                nextNodes.add(element);

                                newNode = (DefaultMutableTreeNode) element.clone();

                                nextCopy.add(newNode);

                                copy.add(newNode);
                            }
                        }
                    }
                }
            }
        }

        /**
         * Creates a new node.
         * 
         * @param isEntry Tells if the node to be created is a speaker entry or a PC reply
         * 
         * @return A vector containing the new node, the structure ponting on the node. 
         */
        private Vector createNewNode(boolean isEntry)
        {
            Vector retval = new Vector(2);

            GFFFile.Field[] nodeFields;
            GFFFile.Field[] pointerFields;

            Vector nodeList;

            if (isEntry)
            {
                nodeFields = new GFFFile.Field[10];
                nodeList =
                    (Vector) nwnDlgFile.getStruct(0).getField(GFFFieldNames.EntryList).getData();

                try
                {
                    nodeFields[9] = nwnDlgFile.new Field(FieldType.CExoString_TYPE, GFFFieldNames.Speaker, ""); //$NON-NLS-1$
                    nodeFields[8] =
                        nwnDlgFile.new Field(
                            FieldType.List_TYPE,
                            GFFFieldNames.RepliesList,
                            new Vector());
                }
                catch (NameTooLong e)
                {

                    e.printStackTrace();
                }
            }
            else
            {
                nodeFields = new GFFFile.Field[9];
                nodeList =
                    (Vector) nwnDlgFile.getStruct(0).getField(GFFFieldNames.ReplyList).getData();
                try
                {
                    nodeFields[8] =
                        nwnDlgFile.new Field(
                            FieldType.List_TYPE,
                            GFFFieldNames.EntriesList,
                            new Vector());
                }
                catch (NameTooLong e)
                {

                    e.printStackTrace();
                }
            }

            StringListElement[] temp = new StringListElement[1];
            try
            {
                temp[0] =
                    new StringListElement(currentLanguage.getString(KeysHolder.NewNodeText), 0);
            }
            catch (InvalidLanguageId e)
            {

                e.printStackTrace();
            }

            try
            {
                nodeFields[0] =
                    nwnDlgFile.new Field(
                        FieldType.DWORD_TYPE,
                        GFFFieldNames.Animation,
                        new Long(0));

                nodeFields[1] =
                    nwnDlgFile.new Field(
                        FieldType.BYTE_TYPE,
                        GFFFieldNames.AnimLoop,
                        new Integer(1));
                nodeFields[2] =
                    nwnDlgFile.new Field(
                        FieldType.CExoLocString_TYPE,
                        GFFFieldNames.Text,
                        new FieldTypeCExoLocString.CExoLocString(0l, temp));
                nodeFields[3] = nwnDlgFile.new Field(FieldType.RESREF_TYPE, GFFFieldNames.Script, ""); //$NON-NLS-1$
                nodeFields[4] =
                    nwnDlgFile.new Field(
                        FieldType.DWORD_TYPE,
                        GFFFieldNames.Delay,
                        new Long(0xFFFFFFFFl));
                nodeFields[5] = nwnDlgFile.new Field(FieldType.CExoString_TYPE, GFFFieldNames.Comment, ""); //$NON-NLS-2$ //$NON-NLS-1$
                nodeFields[6] = nwnDlgFile.new Field(FieldType.RESREF_TYPE, GFFFieldNames.Sound, ""); //$NON-NLS-1$
                nodeFields[7] = nwnDlgFile.new Field(FieldType.CExoString_TYPE, GFFFieldNames.Quest, ""); //$NON-NLS-1$
            }
            catch (NameTooLong e1)
            {

                e1.printStackTrace();
            }

            retval.add(nwnDlgFile.addStructure(0, nodeFields));

            nodeList.add(retval.get(0));

            int index = nodeList.indexOf(retval.get(0));

            pointerFields = new GFFFile.Field[3];

            Long indexValue = new Long(index);

            try
            {
                pointerFields[0] =
                    nwnDlgFile.new Field(FieldType.DWORD_TYPE, GFFFieldNames.Index, indexValue);

                pointerFields[1] = nwnDlgFile.new Field(FieldType.RESREF_TYPE, GFFFieldNames.Active, ""); //$NON-NLS-1$
                pointerFields[2] =
                    nwnDlgFile.new Field(
                        FieldType.BYTE_TYPE,
                        GFFFieldNames.IsChild,
                        new Integer(0));
            }
            catch (NameTooLong e2)
            {

                e2.printStackTrace();
            }

            retval.add(nwnDlgFile.addStructure(0, pointerFields));
            //retval.add(indexValue);

            return retval;

        }

        /**
         * 
         */
        public synchronized void cut()
        {
            copy();
            deleteNode();

        }

        /**
         * Delete the selected node
         */
        protected synchronized void deleteNode()
        {
            TreePath path = dialog.getSelectionPath();

            DefaultMutableTreeNode node;

            if (path != null
                && !(node = ((DefaultMutableTreeNode) path.getLastPathComponent())).isRoot())
            {
                DialogTreeNode treeNode = (DialogTreeNode) node.getUserObject();

                Stack nodeStack = new Stack();

                nodeStack.push(node);

                boolean deleteOk = true;

                if (!treeNode.isLink())
                {
                    deleteOk =
                        JOptionPane.showConfirmDialog(
                            this,
                            languagesHandler.getObject(KeysHolder.NodeDeleteConfirm),
                            languagesHandler.getString(KeysHolder.NodeDeleteConfirmTitle),
                            JOptionPane.YES_NO_OPTION,
                            JOptionPane.INFORMATION_MESSAGE)
                            == JOptionPane.YES_OPTION;

                    if (deleteOk)
                    {
                        if (clipboard != null
                            && ((DialogTreeNode) clipboard.node.getUserObject()).isLink()
                            && ((DialogTreeNode) clipboard.node.getUserObject()).getNode()
                                == treeNode.getNode())
                        {
                            clipboard = null;
                            menuEditPaste.setEnabled(false);
                        }
                        Enumeration enum = node.depthFirstEnumeration();

                        Vector deleted = new Vector();

                        deleted.add(((DialogTreeNode) node.getUserObject()).getNode());

                        while (enum.hasMoreElements())
                        {
                            DefaultMutableTreeNode element =
                                (DefaultMutableTreeNode) enum.nextElement();

                            if (!nodeStack.contains(element))
                            {

                                nodeStack.push(element);
                                deleted.add(((DialogTreeNode) element.getUserObject()).getNode());
                            }
                        }

                        enum = rootNode.depthFirstEnumeration();

                        while (enum.hasMoreElements())
                        {
                            DefaultMutableTreeNode element =
                                (DefaultMutableTreeNode) enum.nextElement();

                            DialogTreeNode cur = (DialogTreeNode) element.getUserObject();

                            if (cur != null
                                && cur.isLink
                                && deleted.contains(cur.getNode())
                                && !nodeStack.contains(element))
                            {
                                nodeStack.push(element);
                            }
                        }

                    }

                }

                if (deleteOk)
                {
                    while (!nodeStack.isEmpty())
                    {
                        DefaultMutableTreeNode cur = (DefaultMutableTreeNode) nodeStack.pop();

                        DefaultMutableTreeNode parent = (DefaultMutableTreeNode) cur.getParent();

                        Vector children;

                        if (parent.isRoot())
                        {

                            children =
                                (Vector) nwnDlgFile
                                    .getStruct(0)
                                    .getField(GFFFieldNames.StartingList)
                                    .getData();
                        }
                        else
                        {
                            DialogTreeNode parentTreeNode = (DialogTreeNode) parent.getUserObject();
                            children =
                                (Vector) parentTreeNode
                                    .getNode()
                                    .getField(
                                        (parentTreeNode.isEntry()
                                            ? GFFFieldNames.RepliesList
                                            : GFFFieldNames.EntriesList))
                                    .getData();
                        }

                        DialogTreeNode currentTreeNode = (DialogTreeNode) cur.getUserObject();

                        children.remove(currentTreeNode.getNodeRef());
                        nwnDlgFile.removeStruct(currentTreeNode.getNodeRef());

                        if (!currentTreeNode.isLink())
                        {
                            if (currentTreeNode.isEntry())
                            {
                                (
                                    (Vector) nwnDlgFile
                                        .getStruct(0)
                                        .getField(GFFFieldNames.EntryList)
                                        .getData())
                                        .remove(
                                    currentTreeNode.getNode());
                                entryList.remove(currentTreeNode);
                            }
                            else
                            {
                                (
                                    (Vector) nwnDlgFile
                                        .getStruct(0)
                                        .getField(GFFFieldNames.ReplyList)
                                        .getData())
                                        .remove(
                                    currentTreeNode.getNode());
                                replyList.remove(currentTreeNode);
                            }

                            nwnDlgFile.getStruct(0).getField(GFFFieldNames.NumWords).setData(
                                new Long(
                                    ((Long) nwnDlgFile
                                        .getStruct(0)
                                        .getField(GFFFieldNames.NumWords)
                                        .getData())
                                        .longValue()
                                        - currentTreeNode.countWords()));
                            nwnDlgFile.removeStruct(currentTreeNode.getNode());
                        }

                        ((DefaultTreeModel) dialog.getModel()).removeNodeFromParent(cur);

                    }

                    Enumeration enum = rootNode.depthFirstEnumeration();

                    while (enum.hasMoreElements())
                    {
                        DefaultMutableTreeNode element =
                            (DefaultMutableTreeNode) enum.nextElement();

                        if (!element.isRoot())
                        {
                            DialogTreeNode temp = (DialogTreeNode) element.getUserObject();

                            if (temp.isEntry())
                            {
                                temp.getNodeRef().getField(GFFFieldNames.Index).setData(
                                    new Long(
                                        (
                                            (Vector) nwnDlgFile
                                                .getStruct(0)
                                                .getField(GFFFieldNames.EntryList)
                                                .getData())
                                                .indexOf(
                                            temp.getNode())));
                            }
                            else
                            {
                                temp.getNodeRef().getField(GFFFieldNames.Index).setData(
                                    new Long(
                                        (
                                            (Vector) nwnDlgFile
                                                .getStruct(0)
                                                .getField(GFFFieldNames.ReplyList)
                                                .getData())
                                                .indexOf(
                                            temp.getNode())));
                            }

                        }
                    }
                }

            }

        }
        /* (non-Javadoc)
                                                                 * @see java.awt.dnd.DragSourceListener#dragDropEnd(java.awt.dnd.DragSourceDropEvent)
                                                                 */
        public void dragDropEnd(DragSourceDropEvent dsde)
        { // TODO Auto-generated method dragDropEnd

        }
        /* (non-Javadoc)
         * @see java.awt.dnd.DragSourceListener#dragEnter(java.awt.dnd.DragSourceDragEvent)
         */
        public void dragEnter(DragSourceDragEvent dsde)
        {
            // TODO Auto-generated method dragEnter

        }
        /* (non-Javadoc)
                                                                 * @see java.awt.dnd.DropTargetListener#dragEnter(java.awt.dnd.DropTargetDragEvent)
                                                                 */
        public void dragEnter(DropTargetDragEvent dtde)
        { // TODO Auto-generated method dragEnter

        } /* (non-Javadoc)
                                                                                                                  * @see java.awt.dnd.DragSourceListener#dragExit(java.awt.dnd.DragSourceEvent)
                                                                                                                  */
        public void dragExit(DragSourceEvent dse)
        { // TODO Auto-generated method dragExit

        } /* (non-Javadoc)
                                                                                                                  * @see java.awt.dnd.DropTargetListener#dragExit(java.awt.dnd.DropTargetEvent)
                                                                                                                  */
        public void dragExit(DropTargetEvent dte)
        { // TODO Auto-generated method dragExit

        }
        /* (non-Javadoc)
         * @see java.awt.dnd.DragGestureListener#dragGestureRecognized(java.awt.dnd.DragGestureEvent)
         */
        public void dragGestureRecognized(DragGestureEvent dge)
        {
            TreePath path = dialog.getSelectionPath();
            DefaultMutableTreeNode treeNode;
            if (path != null
                && !(treeNode = (DefaultMutableTreeNode) path.getLastPathComponent()).isRoot())
            {

                dge.startDrag(DragSource.DefaultMoveNoDrop, new MovingNode(treeNode), this);
            }

        }

        /* (non-Javadoc)
         * @see java.awt.dnd.DragSourceListener#dragOver(java.awt.dnd.DragSourceDragEvent)
         */
        public void dragOver(DragSourceDragEvent dsde)
        {

            Point point = dsde.getLocation();
            Point origin = dialog.getLocationOnScreen();

            point = new Point(point.x - origin.x, point.y - origin.y);

            TreePath path = dialog.getPathForLocation(point.x, point.y);

            try
            {
                if (path != null)
                {

                    DefaultMutableTreeNode treeNode =
                        (DefaultMutableTreeNode) dsde
                            .getDragSourceContext()
                            .getTransferable()
                            .getTransferData(
                            TreeNodeDataFlavor);

                    DefaultMutableTreeNode target =
                        (DefaultMutableTreeNode) path.getLastPathComponent();

                    if (target != treeNode
                        && ((target.isRoot() && ((DialogTreeNode) treeNode.getUserObject()).isEntry())
                            || (!target.isRoot()
                                && !((DialogTreeNode) target.getUserObject()).isLink())))
                        //                        && (
                        //                        (target.isRoot() && ((DialogTreeNode) treeNode.getUserObject()).isEntry())
                        //                        ||
                        //                        (!target.isRoot() && ((DialogTreeNode) treeNode.getUserObject()).isEntry()
                        //                                    == !((DialogTreeNode) ((DefaultMutableTreeNode) path
                        //                                        .getLastPathComponent())
                        //                                        .getUserObject())
                        //                                        .isEntry())))
                    {
                        dsde.getDragSourceContext().setCursor(DragSource.DefaultMoveDrop);

                    }
                    else
                    {
                        dsde.getDragSourceContext().setCursor(DragSource.DefaultMoveNoDrop);
                    }
                }
                else
                {
                    dsde.getDragSourceContext().setCursor(DragSource.DefaultMoveNoDrop);
                }
            }
            catch (UnsupportedFlavorException e)
            {

                e.printStackTrace();
            }
            catch (IOException e)
            {

                e.printStackTrace();
            }

        } /* (non-Javadoc)
                                                                                                                  * @see java.awt.dnd.DropTargetListener#dragOver(java.awt.dnd.DropTargetDragEvent)
                                                                                                                  */
        public void dragOver(DropTargetDragEvent dtde)
        { // TODO Auto-generated method dragOver

        } /* (non-Javadoc)
                                                                                                                  * @see java.awt.dnd.DropTargetListener#drop(java.awt.dnd.DropTargetDropEvent)
                                                                                                                  */
        public void drop(DropTargetDropEvent dtde)
        {

            DefaultMutableTreeNode target =
                (DefaultMutableTreeNode) dialog
                    .getPathForLocation(dtde.getLocation().x, dtde.getLocation().y)
                    .getLastPathComponent();

            DefaultMutableTreeNode node;
            try
            {
                /*Transferable transferable = dtde.getTransferable(); 
                TreePath path = (TreePath) transferable.getTransferData(TreeNodePathDataFlavor);
                */
                node = (DefaultMutableTreeNode) dialog.getSelectionPath().getLastPathComponent();

                if (target.isRoot()
                    && !node.equals(target)
                    && !node.getParent().equals(target)
                    && ((DialogTreeNode) node.getUserObject()).isEntry())
                {
                    DefaultTreeModel model = (DefaultTreeModel) dialog.getModel();

                    (
                        (Vector) nwnDlgFile
                            .getStruct(0)
                            .getField(GFFFieldNames.StartingList)
                            .getData())
                            .add(
                        ((DialogTreeNode) node.getUserObject()).getNodeRef());

                    DefaultMutableTreeNode oldParent = (DefaultMutableTreeNode) node.getParent();

                    DialogTreeNode parentTreeNode = (DialogTreeNode) oldParent.getUserObject();

                    if (parentTreeNode.isEntry())
                    {

                        (
                            (Vector) parentTreeNode
                                .getNode()
                                .getField(GFFFieldNames.RepliesList)
                                .getData())
                                .remove(
                            ((DialogTreeNode) node.getUserObject()).getNodeRef());

                    }
                    else
                    {
                        (
                            (Vector) parentTreeNode
                                .getNode()
                                .getField(GFFFieldNames.EntriesList)
                                .getData())
                                .remove(
                            ((DialogTreeNode) node.getUserObject()).getNodeRef());
                    }

                    model.removeNodeFromParent(node);
                    model.insertNodeInto(node, target, target.getChildCount());

                }
                else if (
                    !node.equals(target)
                        && !node.getParent().equals(target)
                        && (((DialogTreeNode) node.getUserObject()).isEntry()
                            != ((DialogTreeNode) target.getUserObject()).isEntry()))
                {
                    DefaultTreeModel model = (DefaultTreeModel) dialog.getModel();

                    if (((DialogTreeNode) target.getUserObject()).isEntry())
                    {
                        (
                            (Vector) ((DialogTreeNode) target.getUserObject())
                                .getNode()
                                .getField(GFFFieldNames.RepliesList)
                                .getData())
                                .add(
                            ((DialogTreeNode) node.getUserObject()).getNodeRef());
                    }
                    else
                    {
                        (
                            (Vector) ((DialogTreeNode) target.getUserObject())
                                .getNode()
                                .getField(GFFFieldNames.EntriesList)
                                .getData())
                                .add(
                            ((DialogTreeNode) node.getUserObject()).getNodeRef());
                    }
                    DefaultMutableTreeNode oldParent = (DefaultMutableTreeNode) node.getParent();

                    DialogTreeNode parentTreeNode = (DialogTreeNode) oldParent.getUserObject();

                    if (parentTreeNode.isEntry())
                    {

                        (
                            (Vector) parentTreeNode
                                .getNode()
                                .getField(GFFFieldNames.RepliesList)
                                .getData())
                                .remove(
                            ((DialogTreeNode) node.getUserObject()).getNodeRef());

                    }
                    else
                    {
                        (
                            (Vector) parentTreeNode
                                .getNode()
                                .getField(GFFFieldNames.EntriesList)
                                .getData())
                                .remove(
                            ((DialogTreeNode) node.getUserObject()).getNodeRef());
                    }

                    model.removeNodeFromParent(node);
                    model.insertNodeInto(node, target, target.getChildCount());
                }
                else if (
                    !node.equals(target)
                        && (((DialogTreeNode) node.getUserObject()).isEntry()
                            == ((DialogTreeNode) target.getUserObject()).isEntry()))
                {
                    DefaultTreeModel model = (DefaultTreeModel) dialog.getModel();

                    int index = target.getParent().getIndex(target);

                    target = (DefaultMutableTreeNode) target.getParent();

                    DefaultMutableTreeNode oldParent = (DefaultMutableTreeNode) node.getParent();

                    if (oldParent.isRoot())
                    {
                        (
                            (Vector) nwnDlgFile
                                .getStruct(0)
                                .getField(GFFFieldNames.StartingList)
                                .getData())
                                .remove(
                            ((DialogTreeNode) node.getUserObject()).getNodeRef());
                    }
                    else
                    {

                        DialogTreeNode parentTreeNode = (DialogTreeNode) oldParent.getUserObject();

                        if (parentTreeNode.isEntry())
                        {

                            (
                                (Vector) parentTreeNode
                                    .getNode()
                                    .getField(GFFFieldNames.RepliesList)
                                    .getData())
                                    .remove(
                                ((DialogTreeNode) node.getUserObject()).getNodeRef());

                        }
                        else
                        {
                            (
                                (Vector) parentTreeNode
                                    .getNode()
                                    .getField(GFFFieldNames.EntriesList)
                                    .getData())
                                    .remove(
                                ((DialogTreeNode) node.getUserObject()).getNodeRef());
                        }
                    }

                    if (target.isRoot())
                    {
                        (
                            (Vector) nwnDlgFile
                                .getStruct(0)
                                .getField(GFFFieldNames.StartingList)
                                .getData())
                                .insertElementAt(
                            ((DialogTreeNode) node.getUserObject()).getNodeRef(),
                            index);

                    }
                    else
                    {

                        if (((DialogTreeNode) node.getUserObject()).isEntry())
                        {
                            (
                                (Vector) ((DialogTreeNode) node.getUserObject())
                                    .getNode()
                                    .getField(GFFFieldNames.RepliesList)
                                    .getData())
                                    .insertElementAt(
                                ((DialogTreeNode) node.getUserObject()).getNodeRef(),
                                index);
                        }
                        else
                        {
                            (
                                (Vector) ((DialogTreeNode) node.getUserObject())
                                    .getNode()
                                    .getField(GFFFieldNames.EntriesList)
                                    .getData())
                                    .insertElementAt(
                                ((DialogTreeNode) node.getUserObject()).getNodeRef(),
                                index);
                        }

                    }

                    model.removeNodeFromParent(node);
                    model.insertNodeInto(node, target, index);
                }
            }
            catch (ClassCastException e)
            {
                e.printStackTrace();
            }

        } /* (non-Javadoc)
                                                                                                                  * @see java.awt.dnd.DragSourceListener#dropActionChanged(java.awt.dnd.DragSourceDragEvent)
                                                                                                                  */
        public void dropActionChanged(DragSourceDragEvent dsde)
        { // TODO Auto-generated method dropActionChanged

        } /* (non-Javadoc)
                                                                                                                  * @see java.awt.dnd.DropTargetListener#dropActionChanged(java.awt.dnd.DropTargetDragEvent)
                                                                                                                  */
        public void dropActionChanged(DropTargetDragEvent dtde)
        { // TODO Auto-generated method dropActionChanged

        }
        /* (non-Javadoc)
         * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
         */
        public void focusGained(FocusEvent e)
        {
            if (e.getID() == FocusEvent.FOCUS_GAINED)
            {
                if (e.getSource() == this || this.dialog == e.getSource())
                {
                    valueChanged(
                        new TreeSelectionEvent(
                            this,
                            dialog.getSelectionPath(),
                            false,
                            dialog.getLeadSelectionPath(),
                            dialog.getLeadSelectionPath()));
                }
            }

        }

        /* (non-Javadoc)
         * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
         */
        public void focusLost(FocusEvent e)
        {
            // TODO Auto-generated method focusLost

        }

        /**
         * @return Returns an array containing the animation names
         */
        protected Object[] generateAnimationComboList()
        {

            if (animationComboList == null)
            {

                animationComboList = new String[19];

                int i = 0;

                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameDefault);
                animationComboList[i++] = currentLanguage.getString(KeysHolder.AnimationNamenone);
                animationComboList[i++] = currentLanguage.getString(KeysHolder.AnimationNameTaunt);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameSalutation);
                animationComboList[i++] = currentLanguage.getString(KeysHolder.AnimationNameListen);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameAdoring);
                animationComboList[i++] = currentLanguage.getString(KeysHolder.AnimationNameHello);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameLeaning);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNamePickPocketing);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameTalkingNormaly);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameBegging);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameTalkingWithAuthority);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameLaughting);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameVictory1);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameVictory2);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameVictory3);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameLonkingFarAway);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameDrinking);
                animationComboList[i++] =
                    currentLanguage.getString(KeysHolder.AnimationNameReading);
            }

            return animationComboList;
        }

        /**
         * 
         */
        public int getCurrentGender()
        {

            return genderComboBox.getSelectedIndex();
        }

        /**
         * @return The currently used languaged id.
         */
        public long getCurrentLanguageId()
        {

            return languagesHandler.getLanguageId((String) languageComboBox.getSelectedItem());
        }

        /**
         * @return Returns a String array containing the gender names.
         */
        protected String[] getGenderNames()
        {
            String[] retval = new String[2];

            retval[0] = currentLanguage.getString(KeysHolder.GenderMale);
            retval[1] = currentLanguage.getString(KeysHolder.GenderFemale);
            return retval;
        }

        /**
         * @return Returns a list of languages supported by NWN.
         */
        protected String[] getLanguageNames()
        {

            return languagesHandler.getLanguagesList();
        }

        /**
         * Initializes the edit zone.
         * 
         * @throws FileNotFoundException
         * @throws FileDoesNotExist
         * @throws InvalidFile
         * @throws NameTooLong
         * @throws IOException
         * @throws InvalidType
         */
        protected synchronized void init()
            throws
                FileNotFoundException,
                FileDoesNotExist,
                InvalidFile,
                NameTooLong,
                IOException,
                InvalidType
        {

            undoStack = new Stack();
            redoStack = new Stack();

            addFocusListener(this);

            if (dialogFile == null)
            {
                setName(currentLanguage.getString(KeysHolder.UntitledDialog));
                nwnDlgFile = new GFFFile(this);
                nwnDlgFile.setFileType("DLG "); //$NON-NLS-1$
                GFFFile.Structure root = nwnDlgFile.getStruct(0);

                root.addField(FieldType.DWORD_TYPE, GFFFieldNames.DelayEntry, new Long(0l));
                root.addField(FieldType.DWORD_TYPE, GFFFieldNames.DelayReply, new Long(0l));
                root.addField(FieldType.DWORD_TYPE, GFFFieldNames.NumWords, new Long(0l));
                root.addField(
                    FieldType.RESREF_TYPE,
                    GFFFieldNames.EndConversation,
                    owner.getDefaultEndConv());
                root.addField(
                    FieldType.RESREF_TYPE,
                    GFFFieldNames.EndConverAbort,
                    owner.getDefaultAbortConv());
                root.addField(FieldType.BYTE_TYPE, GFFFieldNames.PreventZoomIn, new Integer(0));
                root.addField(FieldType.List_TYPE, GFFFieldNames.EntryList, new Vector());
                root.addField(FieldType.List_TYPE, GFFFieldNames.ReplyList, new Vector());
                root.addField(FieldType.List_TYPE, GFFFieldNames.StartingList, new Vector());

            }
            else
            {
                setName(dialogFile.getAbsolutePath());

                String fileName = dialogFile.getName();

                if (fileName.length() >= 3)
                {
                    String ext = fileName.substring(fileName.length() - 3);

                    if (ext.equalsIgnoreCase("dlg")) //$NON-NLS-1$
                    {
                        nwnDlgFile = new GFFFile(dialogFile, this);
                        if (!nwnDlgFile.getFileType().equalsIgnoreCase("dlg ")) //$NON-NLS-1$
                        {
                            throw new InvalidFile(dialogFile);
                        }
                    }
                    else if (ext.equalsIgnoreCase("gdf")) //$NON-NLS-1$
                    {
                        throw new InvalidFile(dialogFile);
                    }
                    else
                    {
                        nwnDlgFile = new GFFFile(dialogFile, this);
                        if (!nwnDlgFile.getFileType().equalsIgnoreCase("dlg ")) //$NON-NLS-1$
                        {
                            throw new InvalidFile(dialogFile);
                        }
                    }
                }
                else
                {
                    nwnDlgFile = new GFFFile(dialogFile, this);
                    if (!nwnDlgFile.getFileType().equalsIgnoreCase("dlg ")) //$NON-NLS-1$
                    {
                        throw new InvalidFile(dialogFile);
                    }
                }
            }

            languagesHandler = (Languages) ResourceBundle.getBundle(KeysHolder.LanguagesBaseName);

            rootNode = new DefaultMutableTreeNode();

            dialog = new JTree(rootNode);

            JTabbedPane replyOption =
                new JTabbedPane(JTabbedPane.TOP, JTabbedPane.SCROLL_TAB_LAYOUT);
            JPanel dialogOption = new JPanel();

            options = new JPanel();

            options.setLayout(new GridLayout(1, 2));

            dialog.setBorder(BorderFactory.createBevelBorder(BevelBorder.LOWERED));

            options.add(replyOption);

            JScrollPane temp = new JScrollPane(dialogOption);
            temp.setBorder(BorderFactory.createEtchedBorder());

            options.add(temp);

            temp = new JScrollPane(dialog);
            temp.setBorder(BorderFactory.createEmptyBorder());

            DefaultTreeSelectionModel selectionModel = new DefaultTreeSelectionModel();

            selectionModel.setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);

            dialog.setSelectionModel(selectionModel);
            dialog.addTreeSelectionListener(this);
            dialog.setEditable(true);
            dialog.setCellEditor(new TreeEditor());
            dialog.addFocusListener(this);
            //dialog.setDragEnabled(true);
            //dialog.setTransferHandler(new NodeMoover());
            DragSource dragSource = DragSource.getDefaultDragSource();

            dragSource.createDefaultDragGestureRecognizer(dialog, DnDConstants.ACTION_MOVE, this);
            new DropTarget(dialog, DnDConstants.ACTION_COPY_OR_MOVE, this);

            add(temp);
            add(options);

            // Pannels

            int x, y;
            x = y = 0;

            GridBagLayout layout = new GridBagLayout();
            GridBagConstraints contraints = new GridBagConstraints();
            contraints.anchor = GridBagConstraints.NORTHWEST;
            JPanel tabPanel;
            Component tempComponent;

            tabPanel = new JPanel();
            temp = new JScrollPane(tabPanel);
            temp.setBorder(BorderFactory.createEmptyBorder());

            y = 0;

            temp.setName(currentLanguage.getString(KeysHolder.SpeakerTabTitle));
            tabPanel.setLayout(layout);

            replyOption.add(temp);

            tempComponent =
                speakersTagLabel =
                    new JLabel(currentLanguage.getString(KeysHolder.SpeakersTagLabel));

            contraints.gridy = y++;
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridy = y++;
            tempComponent = speakerEditField = new JTextField(16);
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridx = 1;
            tempComponent =
                pickSpeakerButton =
                    new JButton(currentLanguage.getString(KeysHolder.PickButtonLabel));
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);
            contraints.gridx = 0;

            tabPanel = new JPanel();
            temp = new JScrollPane(tabPanel);
            temp.setBorder(BorderFactory.createEmptyBorder());

            y = 0;

            temp.setName(currentLanguage.getString(KeysHolder.ScriptTabTitle));
            tabPanel.setLayout(layout);

            replyOption.add(temp);

            tempComponent =
                textAppearWhenLabel =
                    new JLabel(currentLanguage.getString(KeysHolder.TextAppearsWhenLabel));

            contraints.gridy = y++;
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridy = y++;
            tempComponent = textAppearWhenField = new JTextField(16);
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridx = 1;
            tempComponent =
                textAppearWhenSelectButton =
                    new JButton(currentLanguage.getString(KeysHolder.PickButtonLabel));
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);
            contraints.gridx = 0;

            tempComponent =
                activationScriptLabel =
                    new JLabel(currentLanguage.getString(KeysHolder.ActivationscriptLabel));

            contraints.gridy = y++;
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridy = y++;
            tempComponent = activationScriptField = new JTextField(16);
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridx = 1;
            tempComponent =
                activationScriptButton =
                    new JButton(currentLanguage.getString(KeysHolder.PickButtonLabel));
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);
            contraints.gridx = 0;

            tabPanel = new JPanel();
            temp = new JScrollPane(tabPanel);
            temp.setBorder(BorderFactory.createEmptyBorder());

            y = 0;

            contraints.gridy = y++;

            temp.setName(currentLanguage.getString(KeysHolder.OtherTabTitle));
            tabPanel.setLayout(layout);

            replyOption.add(temp);

            tempComponent =
                animationLabel = new JLabel(currentLanguage.getString(KeysHolder.AnimationLabel));

            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridy = y++;
            tempComponent = replyAnimation = new JComboBox(generateAnimationComboList());
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridy = y++;
            tempComponent =
                animationLoop =
                    new JCheckBox(currentLanguage.getString(KeysHolder.AnimationLoopLabel), true);
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            tempComponent =
                soundLabel = new JLabel(currentLanguage.getString(KeysHolder.SoundLabel));

            contraints.gridy = y++;
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridy = y++;
            tempComponent = replySound = new JTextField(16);
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            tempComponent =
                journalEntryNameLabel =
                    new JLabel(currentLanguage.getString(KeysHolder.JournalEntryNameLabel));

            contraints.gridy = y;
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridy = y + 1;
            tempComponent = journalEntryName = new JTextField(16);
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridx = 1;
            tempComponent =
                journalEntryNumberLabel =
                    new JLabel(currentLanguage.getString(KeysHolder.JournalEntryNumberLabel));

            contraints.gridy = y++;
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridy = y++;
            tempComponent = journalEntryNumber = new JTextField(4);
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);
            contraints.gridx = 0;

            tabPanel = new JPanel();
            temp = new JScrollPane(tabPanel);
            temp.setBorder(BorderFactory.createEmptyBorder());

            y = 0;

            temp.setName(currentLanguage.getString(KeysHolder.CommentTabTitle));
            tabPanel.setLayout(layout);

            replyOption.add(temp);

            contraints.gridy = y++;
            tempComponent = commentArea = new JTextArea(10, 50);
            commentArea.setBorder(BorderFactory.createLoweredBevelBorder());
            layout.setConstraints(tempComponent, contraints);
            tabPanel.add(tempComponent);

            contraints.gridx = 0;

            // General options

            dialogOption.setLayout(layout);

            y = 0;
            x = 0;

            contraints.gridx = x;
            tempComponent =
                languageLabel = new JLabel(currentLanguage.getString(KeysHolder.LanguageLabel));

            contraints.gridy = y;
            layout.setConstraints(tempComponent, contraints);
            dialogOption.add(tempComponent);

            tempComponent = languageComboBox = new JComboBox(getLanguageNames());

            contraints.gridy = y + 1;
            layout.setConstraints(tempComponent, contraints);
            dialogOption.add(tempComponent);

            contraints.gridx = x + 1;
            tempComponent =
                genderLabel = new JLabel(currentLanguage.getString(KeysHolder.GenderLabel));

            contraints.gridy = y;
            layout.setConstraints(tempComponent, contraints);
            dialogOption.add(tempComponent);

            tempComponent = genderComboBox = new JComboBox(getGenderNames());

            contraints.gridy = ++y;
            ++y;
            layout.setConstraints(tempComponent, contraints);
            dialogOption.add(tempComponent);

            contraints.gridx = x++;
            tempComponent =
                exitScriptLabel = new JLabel(currentLanguage.getString(KeysHolder.ExitScriptLabel));

            contraints.gridy = y;
            layout.setConstraints(tempComponent, contraints);
            dialogOption.add(tempComponent);

            contraints.gridy = y + 1;
            tempComponent = exitScriptField = new JTextField(16);
            layout.setConstraints(tempComponent, contraints);
            dialogOption.add(tempComponent);

            contraints.gridx = x++;
            tempComponent =
                exitScriptPick = new JButton(currentLanguage.getString(KeysHolder.PickButtonLabel));
            layout.setConstraints(tempComponent, contraints);
            dialogOption.add(tempComponent);

            contraints.gridx = x++;
            tempComponent =
                abortScriptLabel =
                    new JLabel(currentLanguage.getString(KeysHolder.AbortScriptLabel));

            contraints.gridy = y;
            layout.setConstraints(tempComponent, contraints);
            dialogOption.add(tempComponent);

            contraints.gridy = y + 1;
            tempComponent = abortScriptField = new JTextField(16);
            layout.setConstraints(tempComponent, contraints);
            dialogOption.add(tempComponent);

            contraints.gridx = x++;
            tempComponent =
                abortScriptPick =
                    new JButton(currentLanguage.getString(KeysHolder.PickButtonLabel));
            layout.setConstraints(tempComponent, contraints);
            dialogOption.add(tempComponent);

            contraints.gridx = 0;
            contraints.gridy = contraints.gridy + 1;
            tempComponent =
                cameraZoom =
                    new JCheckBox(
                        currentLanguage.getString(KeysHolder.PreventCameraZoomLabel),
                        false);
            layout.setConstraints(tempComponent, contraints);
            dialogOption.add(tempComponent);

            //listeners

            speakerEditField.getDocument().addDocumentListener(this);

            cameraZoom.addActionListener(this);
            abortScriptPick.addActionListener(this);

            abortScriptField.getDocument().addDocumentListener(this);

            exitScriptPick.addActionListener(this);

            exitScriptField.getDocument().addDocumentListener(this);

            journalEntryNumber.getDocument().addDocumentListener(this);

            journalEntryName.getDocument().addDocumentListener(this);

            replySound.addActionListener(this);
            animationLoop.addActionListener(this);

            replyAnimation.addItemListener(this);
            genderComboBox.addItemListener(this);
            languageComboBox.addItemListener(this);

            activationScriptButton.addActionListener(this);

            activationScriptField.getDocument().addDocumentListener(this);

            textAppearWhenSelectButton.addActionListener(this);

            textAppearWhenField.getDocument().addDocumentListener(this);

            pickSpeakerButton.addActionListener(this);

            speakerEditField.getDocument().addDocumentListener(this);

            loadDialog();

            dialog.expandPath(new TreePath(rootNode));

        }

        /* (non-Javadoc)
         * @see javax.swing.event.DocumentListener#insertUpdate(javax.swing.event.DocumentEvent)
         */
        public void insertUpdate(DocumentEvent e)
        {
            updateText(e);

        }

        /**
         * Returns the modification state of that dialog.
         * @return true if the dialog has been modified
         */
        public synchronized boolean isModified()
        {
            return modified;
        }

        /* (non-Javadoc)
         * @see java.awt.event.ItemListener#itemStateChanged(java.awt.event.ItemEvent)
         */
        public synchronized void itemStateChanged(ItemEvent e)
        {
            TreePath path = dialog.getSelectionPath();

            synchronized (nwnDlgFile)
            {

                if (e.getStateChange() == ItemEvent.SELECTED)
                {
                    if (e.getItemSelectable() == replyAnimation)
                    {
                        int i = 0;
                        Object selected = e.getItem();

                        if (path.getPathCount() > 1)
                        {

                            while (i < replyAnimation.getItemCount()
                                && replyAnimation.getItemAt(i) != selected)
                            {
                                ++i;
                            }

                            DialogTreeNode node =
                                (DialogTreeNode) ((DefaultMutableTreeNode) path
                                    .getLastPathComponent())
                                    .getUserObject();

                            GFFFile.Field field = node.getNode().getField(GFFFieldNames.Animation);

                            if (i < replyAnimation.getItemCount()
                                && ((Long) field.getData()).intValue() != i)
                            {

                                modifyUndo(
                                    new FieldChangeUndo(
                                        node,
                                        GFFFieldNames.Animation,
                                        field.getData(),
                                        replyAnimation));
                                field.setData(new Long(i));
                            }
                        }
                    }
                    else if (
                        e.getItemSelectable() == genderComboBox
                            || e.getItemSelectable() == languageComboBox)
                    {
                        //((DefaultTreeCellRenderer)dialog.getCellRenderer()).repaint(((DefaultTreeCellRenderer)dialog.getCellRenderer()).getVisibleRect());

                    }
                }
            }

        }

        /**
         * 
         */
        public synchronized void link()
        {
            TreePath path = dialog.getSelectionPath();

            DefaultMutableTreeNode treeNode;

            if (path != null)
                //&& !(treeNode = (DefaultMutableTreeNode) path.getLastPathComponent()).isRoot())
            {
                treeNode = (DefaultMutableTreeNode) path.getLastPathComponent();

                DialogTreeNode node = (DialogTreeNode) treeNode.getUserObject();

                DialogTreeNode clipNode = (DialogTreeNode) clipboard.node.getUserObject();

                if ((node == null && clipNode.isEntry())
                    || (node != null && node.isEntry() != clipNode.isEntry()))
                {
                    if (clipNode.isLink() && clipboard.origin == this)
                    {
                        Field[] pointerFields = new GFFFile.Field[4];

                        Object indexValue = node.getNodeRef().getField(GFFFieldNames.Index);

                        try
                        {
                            pointerFields[0] =
                                nwnDlgFile.new Field(
                                    FieldType.DWORD_TYPE,
                                    GFFFieldNames.Index,
                                    indexValue);

                            pointerFields[1] = nwnDlgFile.new Field(FieldType.RESREF_TYPE, GFFFieldNames.Active, ""); //$NON-NLS-1$
                            pointerFields[2] =
                                nwnDlgFile.new Field(
                                    FieldType.BYTE_TYPE,
                                    GFFFieldNames.IsChild,
                                    new Integer(0));
                            pointerFields[3] =
                                nwnDlgFile.new Field(
                                    FieldType.CExoString_TYPE,
                                    GFFFieldNames.LinkComment,
                                    new String());
                        }
                        catch (NameTooLong e2)
                        {

                            e2.printStackTrace();
                        }
                        GFFFile.Structure pointer = nwnDlgFile.addStructure(0, pointerFields);

                        ((DefaultTreeModel) dialog.getModel()).insertNodeInto(
                            treeNode,
                            new DefaultMutableTreeNode(
                                new DialogTreeNode(
                                    clipNode.getNode(),
                                    pointer,
                                    clipNode.isEntry())),
                            treeNode.getChildCount());

                        Vector children;

                        if (treeNode.isRoot())
                        {
                            children =
                                (Vector) nwnDlgFile
                                    .getStruct(0)
                                    .getField(GFFFieldNames.StartingList)
                                    .getData();
                        }
                        else
                        {
                            if (clipNode.isEntry())
                            {
                                children =
                                    (Vector) node
                                        .getNode()
                                        .getField(GFFFieldNames.EntriesList)
                                        .getData();
                            }
                            else
                            {
                                children =
                                    (Vector) node
                                        .getNode()
                                        .getField(GFFFieldNames.RepliesList)
                                        .getData();
                            }

                        }

                        children.add(pointer);
                    }
                }
            }

        }

        /**
         * Loads the dialog from the current dialog file.
         */
        protected void loadDialog()
        {

            synchronized (nwnDlgFile)
            {

                GFFFile.Structure root = nwnDlgFile.getStruct(0);
                GFFFile.Field field = root.getField(GFFFieldNames.EndConversation);

                if (field != null)
                {
                    exitScriptField.setText((String) field.getData());
                }

                field = root.getField(GFFFieldNames.EndConverAbort);

                if (field != null)
                {
                    abortScriptField.setText((String) field.getData());
                }

                field = root.getField(GFFFieldNames.PreventZoomIn);

                if (field != null)
                {
                    Integer value = (Integer) field.getData();

                    if (value.intValue() == 0)
                    {
                        cameraZoom.setSelected(false);
                    }
                    else
                    {
                        cameraZoom.setSelected(true);
                    }

                }

                try
                {
                    Vector currentLevel = new Vector(1);
                    Vector currentLevelChildLists = new Vector(1);
                    Vector nextLevelChildLists;

                    currentLevel.add(rootNode);

                    Vector nextLevel;

                    Vector fileEntryList =
                        (Vector) root.getField(GFFFieldNames.EntryList).getData();

                    entryList = new Vector(fileEntryList.size());

                    Vector fileReplyList =
                        (Vector) root.getField(GFFFieldNames.ReplyList).getData();

                    replyList = new Vector(fileReplyList.size());

                    Vector childList;

                    startingList = (Vector) root.getField(GFFFieldNames.StartingList).getData();

                    currentLevelChildLists.add(startingList);

                    boolean isEntry = true;

                    while (!currentLevel.isEmpty())
                    {

                        Iterator j = currentLevel.iterator();
                        Iterator k = currentLevelChildLists.iterator();
                        Vector nodeList;
                        Vector nodeSourceList;
                        String listFieldName;

                        nextLevel = new Vector();
                        nextLevelChildLists = new Vector();

                        if (isEntry)
                        {

                            nodeSourceList = fileEntryList;

                            nodeList = entryList;
                            listFieldName = GFFFieldNames.RepliesList;
                        }
                        else
                        {
                            nodeSourceList = fileReplyList;
                            nodeList = replyList;
                            listFieldName = GFFFieldNames.EntriesList;
                        }

                        while (j.hasNext())
                        {

                            childList = (Vector) k.next();
                            DefaultMutableTreeNode currentNode = (DefaultMutableTreeNode) j.next();

                            Iterator i = childList.iterator();

                            while (i.hasNext())
                            {
                                GFFFile.Structure cur = (Structure) i.next();

                                long index =
                                    ((Long) cur.getField(GFFFieldNames.Index).getData())
                                        .longValue();

                                GFFFile.Structure node =
                                    (Structure) nodeSourceList.get((int) index);

                                DialogTreeNode newNode = new DialogTreeNode(node, cur, isEntry);

                                DefaultMutableTreeNode temp = new DefaultMutableTreeNode(newNode);

                                if (!newNode.isLink())
                                {
                                    if (index >= nodeList.size())
                                    {
                                        nodeList.setSize((int) (index + 1));
                                    }
                                    nodeList.set((int) index, newNode);
                                    nextLevel.add(temp);
                                    nextLevelChildLists.add(node.getField(listFieldName).getData());
                                }

                                currentNode.add(temp);
                            }
                        }
                        currentLevel = nextLevel;
                        currentLevelChildLists = nextLevelChildLists;
                        isEntry = !isEntry;
                    }

                }
                catch (NullPointerException e)
                {
                    e.printStackTrace();
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
            }
        }

        /**
         * Adds an operation to the redo stack.
         * @param operation Operation to be added to the redo stack.
         */
        protected synchronized void modifyRedo(UndoRedo operation)
        {
            redoStack.add(operation);
            menuEditRedo.setEnabled(true);
        }

        /**
         * Adds an operation to the undo stack.
         * @param operation Operation to be added to the undo stack
         */
        protected synchronized void modifyUndo(UndoRedo operation)
        {
            undoStack.add(operation);
            menuEditUndo.setEnabled(true);
        }

        /**
         * 
         */
        public synchronized void paste()
        {
            TreePath path = dialog.getSelectionPath();

            DefaultMutableTreeNode treeNode;

            if (path != null)
                //&& !(treeNode = (DefaultMutableTreeNode) path.getLastPathComponent()).isRoot())
            {
                treeNode = (DefaultMutableTreeNode) path.getLastPathComponent();

                DialogTreeNode node = (DialogTreeNode) treeNode.getUserObject();

                DialogTreeNode clipNode = (DialogTreeNode) clipboard.node.getUserObject();

                if ((node == null && clipNode.isEntry())
                    || (node != null && node.isEntry() != clipNode.isEntry()))
                {
                    if (clipNode.isLink())
                    {
                        if (clipboard.origin == this)
                        {
                            Field[] pointerFields = new GFFFile.Field[4];

                            Object indexValue = node.getNodeRef().getField(GFFFieldNames.Index);

                            try
                            {
                                pointerFields[0] =
                                    nwnDlgFile.new Field(
                                        FieldType.DWORD_TYPE,
                                        GFFFieldNames.Index,
                                        indexValue);

                                pointerFields[1] = nwnDlgFile.new Field(FieldType.RESREF_TYPE, GFFFieldNames.Active, ""); //$NON-NLS-1$
                                pointerFields[2] =
                                    nwnDlgFile.new Field(
                                        FieldType.BYTE_TYPE,
                                        GFFFieldNames.IsChild,
                                        new Integer(0));
                                pointerFields[3] =
                                    nwnDlgFile.new Field(
                                        FieldType.CExoString_TYPE,
                                        GFFFieldNames.LinkComment,
                                        new String());
                            }
                            catch (NameTooLong e2)
                            {

                                e2.printStackTrace();
                            }
                            GFFFile.Structure pointer = nwnDlgFile.addStructure(0, pointerFields);

                            ((DefaultTreeModel) dialog.getModel()).insertNodeInto(
                                treeNode,
                                new DefaultMutableTreeNode(
                                    new DialogTreeNode(
                                        clipNode.getNode(),
                                        pointer,
                                        clipNode.isEntry())),
                                treeNode.getChildCount());

                            Vector children;

                            if (treeNode.isRoot())
                            {
                                children =
                                    (Vector) nwnDlgFile
                                        .getStruct(0)
                                        .getField(GFFFieldNames.StartingList)
                                        .getData();
                            }
                            else
                            {
                                if (clipNode.isEntry())
                                {
                                    children =
                                        (Vector) node
                                            .getNode()
                                            .getField(GFFFieldNames.EntriesList)
                                            .getData();
                                }
                                else
                                {
                                    children =
                                        (Vector) node
                                            .getNode()
                                            .getField(GFFFieldNames.RepliesList)
                                            .getData();
                                }

                            }

                            children.add(pointer);
                        }
                    }
                    else
                    {
                        Vector currentNodes, nextNodes, nextCopy, currentCopy;

                        nextNodes = new Vector();
                        nextCopy = new Vector();

                        GFFFile.Structure pointer =
                            (Structure) clipNode.getNodeRef().clone(nwnDlgFile);
                        GFFFile.Structure nodeStruct =
                            (Structure) clipNode.getNode().clone(nwnDlgFile);

                        Vector structList;

                        if (clipNode.isEntry())
                        {
                            structList =
                                (Vector) nwnDlgFile
                                    .getStruct(0)
                                    .getField(GFFFieldNames.EntryList)
                                    .getData();
                        }
                        else
                        {
                            structList =
                                (Vector) nwnDlgFile
                                    .getStruct(0)
                                    .getField(GFFFieldNames.ReplyList)
                                    .getData();
                        }

                        structList.add(nodeStruct);
                        nwnDlgFile.addStructure(nodeStruct);

                        pointer.getField(GFFFieldNames.Index).setData(
                            new Long(structList.indexOf(nodeStruct)));
                        nwnDlgFile.addStructure(pointer);

                        DefaultMutableTreeNode cur =
                            new DefaultMutableTreeNode(
                                new DialogTreeNode(nodeStruct, pointer, clipNode.isEntry()));

                        ((DefaultTreeModel) dialog.getModel()).insertNodeInto(
                            cur,
                            treeNode,
                            treeNode.getChildCount());

                        Vector children;

                        if (treeNode.isRoot())
                        {
                            children =
                                (Vector) nwnDlgFile
                                    .getStruct(0)
                                    .getField(GFFFieldNames.StartingList)
                                    .getData();
                        }
                        else
                        {
                            if (clipNode.isEntry())
                            {
                                children =
                                    (Vector) node
                                        .getNode()
                                        .getField(GFFFieldNames.EntriesList)
                                        .getData();
                            }
                            else
                            {
                                children =
                                    (Vector) node
                                        .getNode()
                                        .getField(GFFFieldNames.RepliesList)
                                        .getData();
                            }

                        }

                        children.add(pointer);

                        nextCopy.add(cur);
                        nextNodes.add(clipboard.node);

                        while (!nextNodes.isEmpty())
                        {
                            currentCopy = nextCopy;
                            currentNodes = nextNodes;

                            nextCopy = new Vector();
                            nextNodes = new Vector();

                            Iterator i, j;

                            i = currentCopy.iterator();
                            j = currentNodes.iterator();

                            while (i.hasNext())
                            {
                                DefaultMutableTreeNode copy = (DefaultMutableTreeNode) i.next();
                                DefaultMutableTreeNode original = (DefaultMutableTreeNode) j.next();

                                Enumeration e = original.children();

                                while (e.hasMoreElements())
                                {
                                    DefaultMutableTreeNode element =
                                        (DefaultMutableTreeNode) e.nextElement();

                                    clipNode = (DialogTreeNode) element.getUserObject();

                                    if (clipNode.isLink())
                                    {
                                        if (clipboard.origin == this)
                                        {
                                            pointer = (Structure) clipNode.getNodeRef().clone();

                                            nodeStruct = clipNode.getNode();

                                            nwnDlgFile.addStructure(pointer);

                                            cur =
                                                new DefaultMutableTreeNode(
                                                    new DialogTreeNode(
                                                        nodeStruct,
                                                        pointer,
                                                        clipNode.isEntry()));

                                            ((DefaultTreeModel) dialog.getModel()).insertNodeInto(
                                                cur,
                                                copy,
                                                copy.getChildCount());

                                            node = (DialogTreeNode) copy.getUserObject();
                                            /*if (copy.isRoot())
                                            {
                                                children =
                                                    (Vector) nwnDlgFile
                                                        .getStruct(0)
                                                        .getField(GFFFieldNames.StartingList)
                                                        .getData();
                                            }
                                            else
                                            {*/
                                            if (clipNode.isEntry())
                                            {
                                                children =
                                                    (Vector) node
                                                        .getNode()
                                                        .getField(GFFFieldNames.EntriesList)
                                                        .getData();
                                            }
                                            else
                                            {
                                                children =
                                                    (Vector) node
                                                        .getNode()
                                                        .getField(GFFFieldNames.RepliesList)
                                                        .getData();
                                            }

                                            //}

                                            children.add(pointer);

                                            nextCopy.add(cur);
                                            nextNodes.add(element);
                                        }
                                    }
                                    else
                                    {
                                        pointer =
                                            (Structure) clipNode.getNodeRef().clone(nwnDlgFile);
                                        nodeStruct =
                                            (Structure) clipNode.getNode().clone(nwnDlgFile);

                                        if (clipNode.isEntry())
                                        {
                                            structList =
                                                (Vector) nwnDlgFile
                                                    .getStruct(0)
                                                    .getField(GFFFieldNames.EntryList)
                                                    .getData();
                                        }
                                        else
                                        {
                                            structList =
                                                (Vector) nwnDlgFile
                                                    .getStruct(0)
                                                    .getField(GFFFieldNames.ReplyList)
                                                    .getData();
                                        }

                                        structList.add(nodeStruct);
                                        nwnDlgFile.addStructure(nodeStruct);

                                        pointer.getField(GFFFieldNames.Index).setData(
                                            new Long(structList.indexOf(nodeStruct)));
                                        nwnDlgFile.addStructure(pointer);

                                        cur =
                                            new DefaultMutableTreeNode(
                                                new DialogTreeNode(
                                                    nodeStruct,
                                                    pointer,
                                                    clipNode.isEntry()));

                                        ((DefaultTreeModel) dialog.getModel()).insertNodeInto(
                                            cur,
                                            copy,
                                            copy.getChildCount());

                                        node = (DialogTreeNode) copy.getUserObject();

                                        if (clipNode.isEntry())
                                        {
                                            children =
                                                (Vector) node
                                                    .getNode()
                                                    .getField(GFFFieldNames.EntriesList)
                                                    .getData();
                                        }
                                        else
                                        {
                                            children =
                                                (Vector) node
                                                    .getNode()
                                                    .getField(GFFFieldNames.RepliesList)
                                                    .getData();
                                        }

                                        children.add(pointer);

                                        nextCopy.add(cur);
                                        nextNodes.add(element);
                                    }
                                }

                            }

                        }

                    }
                }
            }

        }
        /**
         * Restores the last canceled operation.
         */
        public synchronized void redo()
        {
            ((UndoRedo) redoStack.pop()).redo();
            if (redoStack.isEmpty())
            {
                menuEditRedo.setEnabled(false);
            }
        }

        /* (non-Javadoc)
         * @see javax.swing.event.DocumentListener#removeUpdate(javax.swing.event.DocumentEvent)
         */
        public void removeUpdate(DocumentEvent e)
        {
            updateText(e);

        }

        /**
         * Saves the dialog file to the GFF format. 
         */
        protected void save()
        {
            setEnabled(false);

            synchronized (nwnDlgFile)
            {

                try
                {
                    nwnDlgFile.save();
                }
                catch (IOException e)
                {
                    JOptionPane optionPane =
                        new JOptionPane(
                            e.getMessage(),
                            JOptionPane.ERROR_MESSAGE,
                            JOptionPane.OK_OPTION);
                    JDialog errorMessage =
                        optionPane.createDialog(
                            this,
                            currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
                    errorMessage.setModal(true);
                    errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
                    errorMessage.show();
                }
                catch (InvalidType e)
                {
                    JOptionPane optionPane =
                        new JOptionPane(
                            e.getMessage(),
                            JOptionPane.ERROR_MESSAGE,
                            JOptionPane.OK_OPTION);
                    JDialog errorMessage =
                        optionPane.createDialog(
                            this,
                            currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
                    errorMessage.setModal(true);
                    errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
                    errorMessage.show();
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
            }
            setEnabled(true);

        }

        /**
         * Saves the dialog file to the GFF format. 
         * @param fileName New file name.
         */
        protected void save(File fileName)
        {
            setEnabled(false);
            setName(fileName.getAbsolutePath());
            synchronized (nwnDlgFile)
            {
                try
                {

                    nwnDlgFile.save(fileName);

                }
                catch (IOException e)
                {
                    JOptionPane optionPane =
                        new JOptionPane(
                            e.getMessage(),
                            JOptionPane.ERROR_MESSAGE,
                            JOptionPane.OK_OPTION);
                    JDialog errorMessage =
                        optionPane.createDialog(
                            this,
                            currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
                    errorMessage.setModal(true);
                    errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
                    errorMessage.show();
                }
                catch (InvalidType e)
                {
                    JOptionPane optionPane =
                        new JOptionPane(
                            e.getMessage(),
                            JOptionPane.ERROR_MESSAGE,
                            JOptionPane.OK_OPTION);
                    JDialog errorMessage =
                        optionPane.createDialog(
                            this,
                            currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
                    errorMessage.setModal(true);
                    errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
                    errorMessage.show();
                }
            }
            setEnabled(true);

        }

        /* (non-Javadoc)
         * @see java.awt.Component#setEnabled(boolean)
         */
        public synchronized void setEnabled(boolean enabled)
        {

            genderComboBox.setEnabled(enabled);
            languageComboBox.setEnabled(enabled);
            cameraZoom.setEnabled(enabled);
            abortScriptPick.setEnabled(enabled);
            abortScriptField.setEnabled(enabled);
            exitScriptPick.setEnabled(enabled);
            exitScriptField.setEnabled(enabled);
            commentArea.setEnabled(enabled);
            journalEntryNumber.setEnabled(enabled);
            journalEntryName.setEnabled(enabled);
            replySound.setEnabled(enabled);
            animationLoop.setEnabled(enabled);
            replyAnimation.setEnabled(enabled);

            activationScriptButton.setEnabled(enabled);
            activationScriptField.setEnabled(enabled);
            textAppearWhenSelectButton.setEnabled(enabled);
            textAppearWhenField.setEnabled(enabled);
            pickSpeakerButton.setEnabled(enabled);
            speakerEditField.setEnabled(enabled);
            dialog.setEnabled(enabled);
            options.setEnabled(enabled);

            super.setEnabled(enabled);
        }

        /**
         * Sets the modification state of the dialog
         * @param b new state.
         */
        protected synchronized void setModified(boolean b)
        {
            modified = b;
        }

        /**
         * Cancels the last operation. 
         */
        public synchronized void undo()
        {
            ((UndoRedo) undoStack.pop()).undo();

            if (undoStack.isEmpty())
            {
                menuEditUndo.setEnabled(false);
            }

        }

        /* (non-Javadoc)
         * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
         */
        public synchronized void update(Observable o, Object arg)
        {
            NWNFile.FileState fileState = (FileState) arg;

            if (arg == null)
            {
                try
                {

                    fileState = (FileState) nwnDlgFile.getFileState().clone();
                }
                catch (CloneNotSupportedException e)
                {
                    e.printStackTrace();
                }
            }

            if (fileState != null)
            {
                Object[] args =
                    { fileState, getName(), new Integer(editorZone.indexOfComponent(this))};

                (new InterfaceThread("doUpdate", owner, args)).start(); //$NON-NLS-1$

            }
        }

        /**
         * Handles the text events.
         * 
         * @param e Event fired by the text component
         */
        protected synchronized void updateText(DocumentEvent e)
        {
            TreePath treeNodePath = dialog.getSelectionPath();

            if (e.getDocument() == speakerEditField.getDocument())
            {
                String text = speakerEditField.getText();

                if (text.length() > 32)
                {
                    speakerEditField.setText(text.substring(0, 32));
                }

                if (treeNodePath.getPathCount() > 1)
                {
                    DialogTreeNode node =
                        (DialogTreeNode) ((DefaultMutableTreeNode) treeNodePath
                            .getLastPathComponent())
                            .getUserObject();

                    GFFFile.Structure structure = node.getNode();

                    Field field = structure.getField(GFFFieldNames.Speaker);

                    if (!text.equals(field.getData()))
                    {

                        modifyUndo(
                            new FieldChangeUndo(
                                node,
                                GFFFieldNames.Speaker,
                                field.getData(),
                                speakerEditField));

                        field.setData(text);
                    }
                }

            }
            else if (e.getDocument() == activationScriptField.getDocument())
            {
                String text = activationScriptField.getText();

                if (text.length() > 16)
                {
                    activationScriptField.setText(text.substring(0, 16));
                }

                if (treeNodePath != null && treeNodePath.getPathCount() > 1)
                {
                    DialogTreeNode node =
                        (DialogTreeNode) ((DefaultMutableTreeNode) treeNodePath
                            .getLastPathComponent())
                            .getUserObject();

                    GFFFile.Structure structure = node.getNode();

                    Field field = structure.getField(GFFFieldNames.Script);

                    if (!text.equals(field.getData()))
                    {
                        modifyUndo(
                            new FieldChangeUndo(
                                node,
                                GFFFieldNames.Script,
                                field.getData(),
                                activationScriptField));

                        field.setData(text);
                    }
                }
            }
            else if (e.getDocument() == textAppearWhenField.getDocument())
            {
                String text = textAppearWhenField.getText();

                if (text.length() > 16)
                {
                    textAppearWhenField.setText(text.substring(0, 16));
                }

                if (treeNodePath != null && treeNodePath.getPathCount() > 1)
                {
                    DialogTreeNode node =
                        (DialogTreeNode) ((DefaultMutableTreeNode) treeNodePath
                            .getLastPathComponent())
                            .getUserObject();

                    GFFFile.Structure structure = node.getNodeRef();

                    Field field = structure.getField(GFFFieldNames.Active);

                    if (!text.equals(field.getData()))
                    {
                        modifyUndo(
                            new FieldChangeUndo(
                                node,
                                GFFFieldNames.Active,
                                field.getData(),
                                textAppearWhenField));

                        field.setData(text);
                    }
                }
            }
            else if (e.getDocument() == commentArea.getDocument())
            {
                String text = commentArea.getText();

                if (text.length() > 1024)
                {
                    commentArea.setText(text.substring(0, 1024));
                }

                if (treeNodePath != null && treeNodePath.getPathCount() > 1)
                {
                    DialogTreeNode node =
                        (DialogTreeNode) ((DefaultMutableTreeNode) treeNodePath
                            .getLastPathComponent())
                            .getUserObject();

                    GFFFile.Structure structure;

                    Field field;
                    String fieldLabel;

                    if (node.isLink())
                    {
                        structure = node.getNodeRef();
                        fieldLabel = GFFFieldNames.LinkComment;
                        field = structure.getField(fieldLabel);
                    }
                    else
                    {
                        structure = node.getNode();
                        fieldLabel = GFFFieldNames.Comment;
                        field = structure.getField(fieldLabel);
                    }

                    if (!text.equals(field.getData()))
                    {
                        modifyUndo(
                            new FieldChangeUndo(node, fieldLabel, field.getData(), commentArea));

                        field.setData(text);
                    }
                }
            }
            else if (e.getDocument() == replySound.getDocument())
            {
                String text = replySound.getText();

                if (text.length() > 16)
                {
                    replySound.setText(text.substring(0, 16));
                }

                if (treeNodePath != null && treeNodePath.getPathCount() > 1)
                {
                    DialogTreeNode node =
                        (DialogTreeNode) ((DefaultMutableTreeNode) treeNodePath
                            .getLastPathComponent())
                            .getUserObject();

                    GFFFile.Structure structure = node.getNodeRef();

                    Field field = structure.getField(GFFFieldNames.Sound);

                    if (!text.equals(field.getData()))
                    {
                        modifyUndo(
                            new FieldChangeUndo(
                                node,
                                GFFFieldNames.Sound,
                                field.getData(),
                                replySound));

                        field.setData(text);
                    }
                }
            }
            else if (e.getDocument() == journalEntryName.getDocument())
            {
                String text = journalEntryName.getText();

                if (text.length() > 16)
                {
                    journalEntryName.setText(text.substring(0, 16));
                }

                if (treeNodePath != null && treeNodePath.getPathCount() > 1)
                {
                    DialogTreeNode node =
                        (DialogTreeNode) ((DefaultMutableTreeNode) treeNodePath
                            .getLastPathComponent())
                            .getUserObject();

                    GFFFile.Structure structure = node.getNodeRef();

                    Field field = structure.getField(GFFFieldNames.Quest);

                    if (!text.equals(field.getData()))
                    {
                        modifyUndo(
                            new FieldChangeUndo(
                                node,
                                GFFFieldNames.Quest,
                                field.getData(),
                                journalEntryName));

                        field.setData(text);
                    }
                }
            }
            else if (e.getDocument() == exitScriptField.getDocument())
            {
                String text = exitScriptField.getText();

                if (text.length() > 16)
                {
                    exitScriptField.setText(text.substring(0, 16));
                }

                GFFFile.Structure structure = nwnDlgFile.getStruct(0);

                Field field = structure.getField(GFFFieldNames.EndConversation);

                if (!text.equals(field.getData()))
                {
                    modifyUndo(
                        new FieldChangeUndo(
                            null,
                            GFFFieldNames.EndConversation,
                            field.getData(),
                            exitScriptField));

                    field.setData(text);
                }

            }
            else if (e.getDocument() == abortScriptField.getDocument())
            {
                String text = abortScriptField.getText();

                if (text.length() > 16)
                {
                    abortScriptField.setText(text.substring(0, 16));
                }

                GFFFile.Structure structure = nwnDlgFile.getStruct(0);

                Field field = structure.getField(GFFFieldNames.EndConverAbort);

                if (!text.equals(field.getData()))
                {
                    modifyUndo(
                        new FieldChangeUndo(
                            null,
                            GFFFieldNames.EndConverAbort,
                            field.getData(),
                            abortScriptField));

                    field.setData(text);
                }

            }

        }

        /* (non-Javadoc)
         * @see javax.swing.event.TreeSelectionListener#valueChanged(javax.swing.event.TreeSelectionEvent)
         */
        public synchronized void valueChanged(TreeSelectionEvent e)
        {

            if (e.getPath() == null)
            {
                menuEditPasteLink.setEnabled(false);
                menuEditPaste.setEnabled(false);
            }
            else
            {

                DefaultMutableTreeNode jTreeNode =
                    (DefaultMutableTreeNode) e.getPath().getLastPathComponent();

                DialogTreeNode node = (DialogTreeNode) jTreeNode.getUserObject();

                menuEditAddNode.setEnabled(true);
                if (jTreeNode.isRoot())
                {
                    menuEditDeleteNode.setEnabled(false);
                }
                else
                {
                    menuEditDeleteNode.setEnabled(true);
                    menuEditCopy.setEnabled(true);
                    menuEditCut.setEnabled(true);

                }
                if (clipboard != null)
                {

                    if (((DialogTreeNode) clipboard.node.getUserObject()).isLink())
                    {
                        menuEditPasteLink.setEnabled(false);
                        if (clipboard.origin == this)
                        {
                            menuEditPaste.setEnabled(true);
                        }
                        else
                        {
                            menuEditPaste.setEnabled(false);
                        }
                    }
                    else
                    {
                        menuEditPasteLink.setEnabled(true);
                        menuEditPaste.setEnabled(true);
                    }
                }
                else
                {
                    menuEditPaste.setEnabled(false);
                    menuEditPasteLink.setEnabled(false);
                }

                if (node != null)
                {

                    GFFFile.Structure dialogNode = node.getNode();
                    GFFFile.Structure pointingStructure = node.getNodeRef();

                    GFFFile.Field field;

                    field = dialogNode.getField(GFFFieldNames.Speaker);

                    if (field != null && field.getType() == FieldType.CExoString_TYPE)
                    {
                        speakerEditField.setText((String) field.getData());
                        speakerEditField.setEnabled(true && !node.isLink());
                        pickSpeakerButton.setEnabled(true && !node.isLink());
                    }
                    else
                    {
                        speakerEditField.setEnabled(false);
                        pickSpeakerButton.setEnabled(false);
                    }

                    field = dialogNode.getField(GFFFieldNames.Animation);

                    if (field != null && field.getType() == FieldType.DWORD_TYPE)
                    {
                        replyAnimation.setSelectedIndex(((Long) field.getData()).intValue());
                        replyAnimation.setEnabled(true && !node.isLink());
                    }
                    else
                    {
                        replyAnimation.setEnabled(false);
                    }

                    field = dialogNode.getField(GFFFieldNames.AnimLoop);

                    if (field != null && field.getType() == FieldType.BYTE_TYPE)
                    {
                        animationLoop.setSelected(((Integer) field.getData()).intValue() != 0);
                        animationLoop.setEnabled(true && !node.isLink());
                    }
                    else
                    {
                        animationLoop.setEnabled(false);
                    }

                    field = dialogNode.getField(GFFFieldNames.Script);

                    if (field != null && field.getType() == FieldType.RESREF_TYPE)
                    {
                        activationScriptField.setText((String) field.getData());
                        activationScriptField.setEnabled(true && !node.isLink());
                        activationScriptButton.setEnabled(true && !node.isLink());
                    }
                    else
                    {
                        activationScriptField.setEnabled(false);
                        abortScriptPick.setEnabled(false);
                    }

                    if (node.isLink())
                    {
                        field = pointingStructure.getField(GFFFieldNames.LinkComment);
                    }
                    else
                    {
                        field = dialogNode.getField(GFFFieldNames.Comment);
                    }

                    if (field != null && field.getType() == FieldType.CExoString_TYPE)
                    {
                        commentArea.setText((String) field.getData());
                        commentArea.setEnabled(true);
                    }
                    else
                    {
                        commentArea.setEnabled(false);
                    }

                    field = dialogNode.getField(GFFFieldNames.Sound);

                    if (field != null && field.getType() == FieldType.RESREF_TYPE)
                    {
                        replySound.setText((String) field.getData());
                        replySound.setEnabled(true && !node.isLink());
                        //journalEntryNumber.setEnabled(true&&!node.isLink());
                    }
                    else
                    {
                        replySound.setEnabled(false);

                    }

                    field = dialogNode.getField(GFFFieldNames.Quest);

                    if (field != null && field.getType() == FieldType.CExoString_TYPE)
                    {
                        journalEntryName.setText((String) field.getData());
                        journalEntryName.setEnabled(true && !node.isLink());
                        journalEntryNumber.setEnabled(true && !node.isLink());
                    }
                    else
                    {
                        journalEntryName.setEnabled(false);
                        journalEntryNumber.setEnabled(false);
                    }

                    if (node.isLink())
                    {
                        field = pointingStructure.getField(GFFFieldNames.Index);
                        if (field != null && field.getType() == FieldType.DWORD_TYPE)
                        {
                            DialogTreeNode temp;
                            if (node.isEntry())
                            {
                                try
                                {
                                    temp =
                                        (DialogTreeNode) entryList.get(
                                            ((Long) field.getData()).intValue());
                                }
                                catch (Exception excep)
                                {
                                    temp = null;
                                    excep.printStackTrace();
                                }
                            }
                            else
                            {
                                try
                                {
                                    temp =
                                        (DialogTreeNode) replyList.get(
                                            ((Long) field.getData()).intValue());
                                }
                                catch (Exception excep)
                                {
                                    temp = null;
                                    excep.printStackTrace();
                                }
                            }
                            if (temp != null)
                            {
                                field = temp.getNodeRef().getField(GFFFieldNames.Active);
                            }
                            else
                            {
                                field = null;
                            }
                        }
                        else
                        {
                            field = null;
                        }
                    }
                    else
                    {
                        field = pointingStructure.getField(GFFFieldNames.Active);
                    }

                    if (field != null && field.getType() == FieldType.RESREF_TYPE)
                    {
                        textAppearWhenField.setText((String) field.getData());
                        textAppearWhenField.setEnabled(true && !node.isLink());
                        textAppearWhenSelectButton.setEnabled(true && !node.isLink());
                    }
                    else
                    {
                        textAppearWhenField.setEnabled(false);
                        textAppearWhenSelectButton.setEnabled(false);
                    }
                }
            }
        }

        /* (non-Javadoc)
         * @see java.awt.event.WindowListener#windowActivated(java.awt.event.WindowEvent)
         */
        public synchronized void windowActivated(WindowEvent e)
        {
            // TODO Auto-generated windowActivated

        }

        /* (non-Javadoc)
         * @see java.awt.event.WindowListener#windowClosed(java.awt.event.WindowEvent)
         */
        public synchronized void windowClosed(WindowEvent e)
        {
            // TODO Auto-generated windowClosed

        }

        /* (non-Javadoc)
         * @see java.awt.event.WindowListener#windowClosing(java.awt.event.WindowEvent)
         */
        public synchronized void windowClosing(WindowEvent e)
        {
            // TODO Auto-generated windowClosing

        }

        /* (non-Javadoc)
         * @see java.awt.event.WindowListener#windowDeactivated(java.awt.event.WindowEvent)
         */
        public synchronized void windowDeactivated(WindowEvent e)
        {
            // TODO Auto-generated windowDeactivated

        }

        /* (non-Javadoc)
         * @see java.awt.event.WindowListener#windowDeiconified(java.awt.event.WindowEvent)
         */
        public synchronized void windowDeiconified(WindowEvent e)
        {
            // TODO Auto-generated windowDeiconified

        }

        /* (non-Javadoc)
         * @see java.awt.event.WindowListener#windowIconified(java.awt.event.WindowEvent)
         */
        public synchronized void windowIconified(WindowEvent e)
        {
            // TODO Auto-generated windowIconified

        }

        /* (non-Javadoc)
         * @see java.awt.event.WindowListener#windowOpened(java.awt.event.WindowEvent)
         */
        public synchronized void windowOpened(WindowEvent e)
        {
            // TODO Auto-generated windowOpened

        }

    } /**
                                                  * 16 juil. 2003
                                                  * @author Alexandre Brunel
                                                  *
                                                  * This interface Handles the undo/redo operations.
                                                  */
    public interface UndoRedo
    { /**
                                                  * The operation is restored.
                                                  * 
                                                  * This method is to be used when the object is part of a redo stack.
                                                  */
        public abstract void redo();
        /**
         * The operation is canceled.
         * 
         * This method is to be used when the object is part of an undo stack.
         * 
         */
        public abstract void undo();
    }
    private ERFFileImportDialog importdialog;
    protected static final DataFlavor TreeNodeDataFlavor =
        new DataFlavor(DefaultMutableTreeNode.class, "TreeNode");
    protected static final DataFlavor TreeNodePathDataFlavor =
        new DataFlavor(TreePath.class, "TreeNodePath");
    protected static final DataFlavor[] MovingNodeDataFlavors =
        { TreeNodeDataFlavor, DataFlavor.stringFlavor, TreeNodePathDataFlavor };

    private JPanel aboutDialog;
    private FileFilter allSuportedFormatFilter;
    private ClipBoardElement clipboard;
    private ResourceBundle currentLanguage;
    private File currentPath;
    private ResourceBundle currentRessouceBundle;
    private FileFilter dlgFileFilter;
    private JTabbedPane editorZone;
    private JFileChooser fileChooser;
    private JFileChooser fileSaveChooser;
    private FileFilter gdfFileFilter;
    private JMenu menuEdit;
    private JMenuItem menuEditAddNode;
    private JMenuItem menuEditCopy;
    private JMenuItem menuEditCut;
    private JMenuItem menuEditDeleteNode;
    private JMenuItem menuEditPaste;
    private JMenuItem menuEditPasteLink;
    private JMenuItem menuEditRedo;
    private JMenuItem menuEditUndo;
    private JMenu menuFile;
    private JMenuItem menuFileClose;
    private JMenuItem menuFileCloseAll;
    private JMenuItem menuFileNew;
    private JMenuItem menuFileOpen;
    private JMenuItem menuFileOpenFromERF;
    private JMenuItem menuFileQuit;
    private JMenuItem menuFileSave;
    private JMenuItem menuFileSaveAll;
    private JMenuItem menuFileSaveAs;
    private JMenuItem menuFileSaveAsXml;
    private JMenu menuHelp;
    private JMenuItem menuHelpAbout;
    private JMenu menuOption;
    private JToolBar toolBar;
    /**
     * @throws java.awt.HeadlessException
     */
    public MainWindow() throws HeadlessException
    {
        super();
        init();
    } /**
                                                      * @param gc
                                                      */
    public MainWindow(GraphicsConfiguration gc)
    {
        super(gc);
        init();
    } /**
                                                      * @param title
                                                      * @throws java.awt.HeadlessException
                                                      */
    public MainWindow(String title) throws HeadlessException
    {
        super(title);
        init();
    } /**
                                                      * @param title
                                                      * @param gc
                                                      */
    public MainWindow(String title, GraphicsConfiguration gc)
    {
        super(title, gc);
        init();
    } /* (non-Javadoc)
                                                      * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
                                                      */
    public void actionPerformed(ActionEvent e)
    {
        Object source = e.getSource();
        if (source == menuFileQuit)
        { //(new InterfaceThread("quit", this, null)).start();
            quit();
        }
        else if (source == menuFileClose)
        { //(new InterfaceThread("closeSelected", this, null)).start();
            closeSelected();
        }
        else if (source == menuFileCloseAll)
        { //(new InterfaceThread("closeAll", this, null)).start();
            closeAll();
        }
        else if (source == menuFileNew)
        { //(new InterfaceThread("createNew", this, null)).start();
            createNew();
        }
        else if (source == menuFileOpenFromERF)
        {
            if (importdialog == null)
                importdialog = new ERFFileImportDialog(new File("."), this);

            importdialog.setModal(true);
            importdialog.show();

            if (importdialog.getExtractedFile() != null)
            {

                Object[] args = { importdialog.getExtractedFile()};
                (new InterfaceThread("openInportedFile", this, args)).start(); //$NON-NLS-1$
            }
        }
        else if (source == menuFileSaveAsXml)
        {
        }
        else if (source == menuFileSaveAs)
        {

            saveAs();
        }
        else if (source == menuFileSaveAll)
        {
            (new InterfaceThread("saveAll", this, null)).start(); //$NON-NLS-1$
        }
        else if (source == menuFileSave)
        {
            if (((EditWindow) editorZone.getSelectedComponent()).nwnDlgFile.getFilePath() != null)
            {
                (new InterfaceThread("save", editorZone.getSelectedComponent(), null)).start(); //$NON-NLS-1$
            }
            else
            {
                saveAs();
            }
        }
        else if (source == menuFileOpen)
        {

            if (fileChooser == null)
            {
                fileChooser = new JFileChooser(currentPath);
                fileChooser.setMultiSelectionEnabled(true);
                if (dlgFileFilter == null)
                {
                    dlgFileFilter = new FileFilter()
                    {

                        public boolean accept(File f)
                        {
                            return (f.getName().matches(".*\\.dlg") || f.isDirectory()); //$NON-NLS-1$
                        }

                        public String getDescription()
                        {
                            return currentLanguage.getString(KeysHolder.DLGFileFilterName);
                        }
                    };
                }

                if (gdfFileFilter == null)
                {
                    gdfFileFilter = new FileFilter()
                    {

                        public boolean accept(File f)
                        {
                            return (f.getName().matches(".*\\.gdf") || f.isDirectory()); //$NON-NLS-1$
                        }

                        public String getDescription()
                        {
                            return currentLanguage.getString(KeysHolder.GDFFileFilterName);
                        }
                    };
                }

                if (allSuportedFormatFilter != null)
                {
                    allSuportedFormatFilter = new FileFilter()
                    {

                        public boolean accept(File f)
                        {
                            return (f.getName().matches(".*\\.(dlg)|(gdf)") || f.isDirectory()); //$NON-NLS-1$
                        }

                        public String getDescription()
                        {
                            return currentLanguage.getString(KeysHolder.AllFormatsFileFilterName);
                        }
                    };
                }
                fileChooser.addChoosableFileFilter(allSuportedFormatFilter);
                fileChooser.setAcceptAllFileFilterUsed(true);
                fileChooser.addChoosableFileFilter(gdfFileFilter);
                fileChooser.addChoosableFileFilter(dlgFileFilter);
            }

            if (fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION)
            {

                File[] selectedFiles = fileChooser.getSelectedFiles();
                Object[] args = { selectedFiles };
                (new InterfaceThread("openFiles", this, args)).start(); //$NON-NLS-1$
                //openFiles(selectedFiles);
            }
        }
        else if (source == menuEditUndo)
        {
            (new InterfaceThread("undo", editorZone.getSelectedComponent(), null)).start(); //$NON-NLS-1$
        }
        else if (source == menuEditRedo)
        {

            (new InterfaceThread("redo", editorZone.getSelectedComponent(), null)).start(); //$NON-NLS-1$
        }
        else if (source == menuEditAddNode)
        {
            (new InterfaceThread("addNode", editorZone.getSelectedComponent(), null)).start(); //$NON-NLS-1$
        }
        else if (source == menuEditDeleteNode)
        {
            (new InterfaceThread("deleteNode", editorZone.getSelectedComponent(), null)).start(); //$NON-NLS-1$
        }
        else if (source == menuHelpAbout)
        { /*JOptionPane optionPane =
                                                                                                                     new JOptionPane(ex.getMessage(), JOptionPane.ERROR_MESSAGE, JOptionPane.OK_OPTION);
                                                                                                                 JDialog errorMessage = optionPane.createDialog(this, currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
                                                                                                                 errorMessage.setModal(true);
                                                                                                                 errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
                                                                                                                 errorMessage.show();*/
            if (aboutDialog == null)
            {

                aboutDialog = new JPanel();
                aboutDialog.setLayout(new BoxLayout(aboutDialog, BoxLayout.PAGE_AXIS));
                aboutDialog.add(new JLabel(currentLanguage.getString(KeysHolder.ApplicationName)));
                aboutDialog.add(new JLabel(" ")); //$NON-NLS-1$
                aboutDialog.add(new JLabel("Copyright 2003 Alexandre Brunel")); //$NON-NLS-1$
                aboutDialog.add(new JLabel(" ")); //$NON-NLS-1$
                String temp = currentLanguage.getString(KeysHolder.AboutEmailAddressFormat);
                Object[] args = { "ghanima2@wanadoo.fr" }; //$NON-NLS-1$
                aboutDialog.add(new JLabel(MessageFormat.format(temp, args)));
            }

            JOptionPane.showMessageDialog(this, aboutDialog);
        }
        else if (source == menuEditCopy)
        {
            ((EditWindow) editorZone.getSelectedComponent()).copy();
        }
        else if (source == menuEditCut)
        {
            ((EditWindow) editorZone.getSelectedComponent()).cut();
        }
        else if (source == menuEditPaste)
        {
            ((EditWindow) editorZone.getSelectedComponent()).paste();
        }
        else if (source == menuEditPasteLink)
        {
            ((EditWindow) editorZone.getSelectedComponent()).link();
        }

    } /**
                                                  *	Closes all tabs. 
                                                  */
    protected void closeAll()
    {
        int i = 0;
        synchronized (editorZone)
        {

            while (i < editorZone.getTabCount())
            {

                EditWindow selected = (EditWindow) editorZone.getComponentAt(i);

                if (selected.close())
                {
                    transferFocusUpCycle();
                    editorZone.remove(selected);

                }
                else
                {
                    ++i;
                }
            }

            if (editorZone.getTabCount() == 0)
            {
                menuFileClose.setEnabled(false);
                menuFileCloseAll.setEnabled(false);
                menuEditUndo.setEnabled(false);
                menuEditRedo.setEnabled(false);
                menuFileSaveAll.setEnabled(false);
                menuFileSaveAs.setEnabled(false);
                menuEditAddNode.setEnabled(false);
                //            menuEditPaste.setEnabled(false);
                //            menuEditPasteLink.setEnabled(false);
                //            menuEditCopy.setEnabled(false);
                //			menuEditCut.setEnabled(false);
            }
            else
            {
                EditWindow selected = (EditWindow) editorZone.getSelectedComponent();
                //
                //			boolean temp = !selected.undoStack.isEmpty();
                //                
                //			menuEditUndo.setEnabled(temp);
                //			menuEditRedo.setEnabled(temp);
                //                
                //			temp =
                //				selected.dialog.getSelectionPath().getLastPathComponent() != selected.rootNode;
                //                    
                //			menuEditCopy.setEnabled(temp);
                //			menuEditCut.setEnabled(temp);
                //			temp = temp && clipboard != null;
                //			menuEditPaste.setEnabled(temp);
                //			menuEditPasteLink.setEnabled(temp);
                menuEditAddNode.setEnabled(selected.dialog.getSelectionPath() != null);
            }
        }

    }

    /**
     * Closes selected tab
     */
    protected void closeSelected()
    {
        synchronized (editorZone)
        {
            EditWindow selected = (EditWindow) editorZone.getSelectedComponent();

            if (selected != null && selected.close())
            {
                transferFocusUpCycle();
                editorZone.remove(selected);

                if (editorZone.getTabCount() == 0)
                {
                    menuFileClose.setEnabled(false);
                    menuFileCloseAll.setEnabled(false);
                    menuEditUndo.setEnabled(false);
                    menuEditRedo.setEnabled(false);
                    //                menuEditPaste.setEnabled(false);
                    //                menuEditPasteLink.setEnabled(false);
                    menuEditAddNode.setEnabled(false);
                    menuFileSaveAll.setEnabled(false);
                    menuFileSaveAs.setEnabled(false);

                }
                else
                {
                    selected = (EditWindow) editorZone.getSelectedComponent();
                    //
                    //                boolean temp = !selected.undoStack.isEmpty();
                    //
                    //                menuEditUndo.setEnabled(temp);
                    //                menuEditRedo.setEnabled(temp);
                    //
                    //                temp =
                    //                    selected.dialog.getSelectionPath().getLastPathComponent() != selected.rootNode;
                    //
                    //                menuEditPaste.setEnabled(temp);
                    //                menuEditPasteLink.setEnabled(temp);
                    menuEditAddNode.setEnabled(selected.dialog.getSelectionPath() != null);
                }

            }
        }
    }

    /**
     * Create a new dialog.
     */
    protected void createNew()
    {
        EditWindow temp;

        try
        {

            temp = new EditWindow(this);
            synchronized (editorZone)
            {
                editorZone.add(temp);
                menuFileClose.setEnabled(true);
                menuFileCloseAll.setEnabled(true);
                editorZone.setSelectedComponent(temp);
                menuFileSaveAll.setEnabled(true);
                menuFileSaveAs.setEnabled(true);
            }

        }
        catch (InvalidType ex)
        {

            JOptionPane optionPane =
                new JOptionPane(ex.getMessage(), JOptionPane.ERROR_MESSAGE, JOptionPane.OK_OPTION);
            JDialog errorMessage =
                optionPane.createDialog(
                    this,
                    currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
            errorMessage.setModal(true);
            errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
            errorMessage.show();

        }
        catch (FileNotFoundException ex)
        {
            JOptionPane optionPane =
                new JOptionPane(ex.getMessage(), JOptionPane.ERROR_MESSAGE, JOptionPane.OK_OPTION);
            JDialog errorMessage =
                optionPane.createDialog(
                    this,
                    currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
            errorMessage.setModal(true);
            errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
            errorMessage.show();
        }
        catch (IOException ex)
        {
            JOptionPane optionPane =
                new JOptionPane(ex.getMessage(), JOptionPane.ERROR_MESSAGE, JOptionPane.OK_OPTION);
            JDialog errorMessage =
                optionPane.createDialog(
                    this,
                    currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
            errorMessage.setModal(true);
            errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
            errorMessage.show();
        }
        catch (NWNFileException ex)
        {
            JOptionPane optionPane =
                new JOptionPane(ex.getMessage(), JOptionPane.ERROR_MESSAGE, JOptionPane.OK_OPTION);
            JDialog errorMessage =
                optionPane.createDialog(
                    this,
                    currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
            errorMessage.setModal(true);
            errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
            errorMessage.show();
        }

    }

    synchronized void doUpdate(NWNFile.FileState arg, String title, Integer index)
    {

        FileState fileState = null;

        try
        {

            fileState = (FileState) arg.clone();
        }
        catch (CloneNotSupportedException e)
        {
            e.printStackTrace();
        }

        if (fileState.isModified())
        {

            if (index.intValue() >= 0)
            {
                editorZone.setTitleAt(index.intValue(), title + "*"); //$NON-NLS-1$
                if (index.intValue() == editorZone.getSelectedIndex())
                {
                    menuFileSave.setEnabled(true);
                }
            }

        }
        else if (!fileState.isModified())
        {
            if (index.intValue() >= 0)
            {
                editorZone.setTitleAt(index.intValue(), title);
                if (index.intValue() == editorZone.getSelectedIndex())
                {
                    menuFileSave.setEnabled(false);
                }
            }
        }

    }

    /**
     * @return
     */
    public String getDefaultAbortConv()
    {
        return KeysHolder.DefaultAbortConv;
    }

    /**
     * @return
     */
    public String getDefaultEndConv()
    {

        return KeysHolder.DefaultEndConv;

    }

    /**
     * Iitialize the window.
     *  
     */
    protected void init()
    {

        currentLanguage = ResourceBundle.getBundle(KeysHolder.LanguagesBaseName);

        setTitle(currentLanguage.getString(KeysHolder.ApplicationName));

        JMenuBar mainMenu = getJMenuBar();
        fileChooser = null;
        getRootPane().setDoubleBuffered(true);
        setSize(500, 500);

        setIconImage((new ImageIcon(getClass().getResource("NWNIcon.png"))).getImage()); //$NON-NLS-1$

        currentPath = new File("."); //$NON-NLS-1$

        setDefaultLookAndFeelDecorated(true);

        if (mainMenu == null)
        {
            mainMenu = new JMenuBar();
            setJMenuBar(mainMenu);
        }

        menuFile = new JMenu(currentLanguage.getString(KeysHolder.FileMenuLabel));
        mainMenu.add(menuFile);
        menuFile.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.FileMenuMnemonic)).intValue());

        menuFileNew = menuFile.add(currentLanguage.getString(KeysHolder.FileMenuNewLabel));
        menuFileNew.addActionListener(this);
        menuFileNew.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.FileMenuNewMnemonic)).intValue());
        menuFileNew.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N, InputEvent.CTRL_MASK));

        menuFile.addSeparator();

        menuFileOpen = menuFile.add(currentLanguage.getString(KeysHolder.FileMenuOpenLabel));
        menuFileOpen.addActionListener(this);
        menuFileOpen.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.FileMenuOpenMnemonic)).intValue());
        menuFileOpen.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, InputEvent.CTRL_MASK));

        menuFileOpenFromERF =
            menuFile.add(currentLanguage.getString(KeysHolder.FileMenuOpenFromERFLabel));
        menuFileOpenFromERF.addActionListener(this);
        menuFileOpenFromERF.setEnabled(true);
        menuFileOpenFromERF.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.FileMenuOpenFromERFMnemonic))
                .intValue());

        menuFile.addSeparator();

        menuFileClose = menuFile.add(currentLanguage.getString(KeysHolder.FileMenuCloseLabel));
        menuFileClose.addActionListener(this);
        menuFileClose.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.FileMenuCloseMnemonic)).intValue());
        menuFileClose.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_W, InputEvent.CTRL_MASK));
        menuFileClose.setEnabled(false);

        menuFileCloseAll =
            menuFile.add(currentLanguage.getString(KeysHolder.FileMenuCloseAllLabel));
        menuFileCloseAll.addActionListener(this);
        menuFileCloseAll.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.FileMenuCloseAllMnemonic)).intValue());
        menuFileCloseAll.setAccelerator(
            KeyStroke.getKeyStroke(KeyEvent.VK_W, InputEvent.CTRL_MASK | InputEvent.SHIFT_MASK));
        menuFileCloseAll.setEnabled(false);

        menuFile.addSeparator();

        menuFileSave = menuFile.add(currentLanguage.getString(KeysHolder.FileMenuSaveLabel));
        menuFileSave.setEnabled(false);
        menuFileSave.addActionListener(this);
        menuFileSave.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.FileMenuSaveMnemonic)).intValue());
        menuFileSave.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, InputEvent.CTRL_MASK));

        menuFileSaveAll = menuFile.add(currentLanguage.getString(KeysHolder.FileMenuSaveAllLabel));
        menuFileSaveAll.setEnabled(false);
        menuFileSaveAll.addActionListener(this);
        menuFileSaveAll.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.FileMenuSaveAllMnemonic)).intValue());
        menuFileSaveAll.setAccelerator(
            KeyStroke.getKeyStroke(KeyEvent.VK_S, InputEvent.CTRL_MASK | InputEvent.SHIFT_MASK));

        menuFileSaveAs = menuFile.add(currentLanguage.getString(KeysHolder.FileMenuSaveAsLabel));
        menuFileSaveAs.setEnabled(false);
        menuFileSaveAs.addActionListener(this);
        menuFileSaveAs.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.FileMenuSaveAsMnemonic)).intValue());
        menuFileSaveAs.setAccelerator(
            KeyStroke.getKeyStroke(KeyEvent.VK_S, InputEvent.CTRL_MASK | InputEvent.ALT_MASK));

        menuFileSaveAsXml =
            menuFile.add(currentLanguage.getString(KeysHolder.FileMenuSaveAsXMLLabel));
        menuFileSaveAsXml.setEnabled(false);
        menuFileSaveAsXml.addActionListener(this);
        menuFileSaveAsXml.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.FileMenuSaveAsXMLMnemonic)).intValue());
        menuFileSaveAsXml.setAccelerator(
            KeyStroke.getKeyStroke(KeyEvent.VK_X, InputEvent.CTRL_MASK | InputEvent.ALT_MASK));

        menuFile.add(new JSeparator());
        menuFileQuit = menuFile.add(currentLanguage.getString(KeysHolder.FileMenuQuitLabel));
        menuFileQuit.addActionListener(this);
        menuFileQuit.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.FileMenuQuitMnemonic)).intValue());

        menuEdit = new JMenu(currentLanguage.getString(KeysHolder.EditMenuLabel));
        mainMenu.add(menuEdit);
        menuEdit.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.EditMenuMnemonic)).intValue());

        menuEditUndo = menuEdit.add(currentLanguage.getString(KeysHolder.EditMenuUndoLabel));
        menuEditUndo.addActionListener(this);
        menuEditUndo.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.EditMenuUndoMnemonic)).intValue());
        menuEditUndo.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Z, InputEvent.CTRL_MASK));
        menuEditUndo.setEnabled(false);

        menuEditRedo = menuEdit.add(currentLanguage.getString(KeysHolder.EditMenuRedoLabel));
        menuEditRedo.addActionListener(this);
        menuEditRedo.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.EditMenuRedoMnemonic)).intValue());
        menuEditRedo.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Y, InputEvent.CTRL_MASK));
        menuEditRedo.setEnabled(false);

        menuEdit.addSeparator();

        menuEditAddNode = menuEdit.add(currentLanguage.getString(KeysHolder.EditMenuAddNodeLabel));
        menuEditAddNode.addActionListener(this);
        menuEditAddNode.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.EditMenuAddNodeMnemonic)).intValue());
        menuEditAddNode.setAccelerator(
            KeyStroke.getKeyStroke(KeyEvent.VK_PLUS, InputEvent.CTRL_MASK));
        menuEditAddNode.setEnabled(false);

        menuEditDeleteNode =
            menuEdit.add(currentLanguage.getString(KeysHolder.EditMenuDeleteNodeLabel));
        menuEditDeleteNode.addActionListener(this);
        menuEditDeleteNode.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.EditMenuDeleteNodeMnemonic)).intValue());
        menuEditDeleteNode.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
        menuEditDeleteNode.setEnabled(false);

        menuEdit.addSeparator();

        menuEditCopy = menuEdit.add(currentLanguage.getString(KeysHolder.EditMenuCopyLabel));
        menuEditCopy.addActionListener(this);
        menuEditCopy.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.EditMenuCopyMnemonic)).intValue());
        menuEditCopy.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, InputEvent.CTRL_MASK));
        menuEditCopy.setEnabled(false);

        menuEditPaste = menuEdit.add(currentLanguage.getString(KeysHolder.EditMenuPasteLabel));
        menuEditPaste.addActionListener(this);
        menuEditPaste.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.EditMenuPasteMnemonic)).intValue());
        menuEditPaste.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_V, InputEvent.CTRL_MASK));
        menuEditPaste.setEnabled(false);

        menuEditPasteLink =
            menuEdit.add(currentLanguage.getString(KeysHolder.EditMenuPasteAsLinkLabel));
        menuEditPasteLink.addActionListener(this);
        menuEditPasteLink.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.EditMenuPasteAsLinkMnemonic))
                .intValue());
        menuEditPasteLink.setAccelerator(
            KeyStroke.getKeyStroke(KeyEvent.VK_V, InputEvent.CTRL_MASK | InputEvent.ALT_MASK));
        menuEditPasteLink.setEnabled(false);

        menuEditCut = menuEdit.add(currentLanguage.getString(KeysHolder.EditMenuCutLabel));
        menuEditCut.addActionListener(this);
        menuEditCut.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.EditMenuCutMnemonic)).intValue());
        menuEditCut.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, InputEvent.CTRL_MASK));
        menuEditCut.setEnabled(false);

        menuOption = new JMenu(currentLanguage.getString(KeysHolder.OptionMenuLabel));
        mainMenu.add(menuOption);
        menuOption.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.OptionMenuMnemonic)).intValue());

        menuHelp = new JMenu(currentLanguage.getString(KeysHolder.HelpMenuLabel));
        mainMenu.add(menuHelp);
        menuHelp.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.HelpMenuMnemonic)).intValue());

        menuHelpAbout = menuHelp.add(currentLanguage.getString(KeysHolder.HelpMenuAboutLabel));
        menuHelpAbout.addActionListener(this);
        menuHelpAbout.setMnemonic(
            ((Number) currentLanguage.getObject(KeysHolder.HelpMenuAboutMnemonic)).intValue());
        menuHelpAbout.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F1, 0));

        getRootPane().add(mainMenu);

        editorZone = new JTabbedPane();
        editorZone.setDoubleBuffered(true);
        editorZone.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);

        getContentPane().add(editorZone);

    }

    protected void openInportedFile(File selectedFile)
    {
        int i = editorZone.getTabCount() - 1;
        File[] temp = { selectedFile };

        openFiles(temp);

        selectedFile.deleteOnExit();

        EditWindow selected = (EditWindow) editorZone.getComponentAt(i);

        selected.nwnDlgFile.setFileType(selected.nwnDlgFile.getFileType());

    }

    /**
     * Opens the given dialog files
     * @param selectedFiles Files to be opened.
     */
    protected void openFiles(File[] selectedFiles)
    {
        for (int i = 0; i < selectedFiles.length; ++i)
        {

            EditWindow temp;
            try
            {

                temp = new EditWindow(this, selectedFiles[i]);
                synchronized (editorZone)
                {
                    editorZone.add(temp);
                    menuFileClose.setEnabled(true);
                    menuFileCloseAll.setEnabled(true);
                    editorZone.setSelectedComponent(temp);
                    menuFileSaveAll.setEnabled(true);
                    menuFileSaveAs.setEnabled(true);
                }

            }
            catch (InvalidType ex)
            {

                JOptionPane optionPane =
                    new JOptionPane(
                        ex.getMessage(),
                        JOptionPane.ERROR_MESSAGE,
                        JOptionPane.OK_OPTION);
                JDialog errorMessage =
                    optionPane.createDialog(
                        this,
                        currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
                errorMessage.setModal(true);
                errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
                errorMessage.show();

            }
            catch (FileNotFoundException ex)
            {
                JOptionPane optionPane =
                    new JOptionPane(
                        ex.getMessage(),
                        JOptionPane.ERROR_MESSAGE,
                        JOptionPane.OK_OPTION);
                JDialog errorMessage =
                    optionPane.createDialog(
                        this,
                        currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
                errorMessage.setModal(true);
                errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
                errorMessage.show();
            }
            catch (IOException ex)
            {
                JOptionPane optionPane =
                    new JOptionPane(
                        ex.getMessage(),
                        JOptionPane.ERROR_MESSAGE,
                        JOptionPane.OK_OPTION);
                JDialog errorMessage =
                    optionPane.createDialog(
                        this,
                        currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
                errorMessage.setModal(true);
                errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
                errorMessage.show();
            }
            catch (NWNFileException ex)
            {
                JOptionPane optionPane =
                    new JOptionPane(
                        ex.getMessage(),
                        JOptionPane.ERROR_MESSAGE,
                        JOptionPane.OK_OPTION);
                JDialog errorMessage =
                    optionPane.createDialog(
                        this,
                        currentLanguage.getString(KeysHolder.ErrorDialogsTitle));
                errorMessage.setModal(true);
                errorMessage.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
                errorMessage.show();
            }

        }

    }

    protected synchronized void quit()
    {
        processWindowEvent(new WindowEvent((Window) this, WindowEvent.WINDOW_CLOSING));
    }

    /**
     * Saves all dialogs.
     */
    protected void saveAll()
    {
        synchronized (editorZone)
        {
            for (int i = 0; i < editorZone.getTabCount(); i++)
            {
                ((EditWindow) editorZone.getComponentAt(i)).save();
            }
        }
    }

    /**
     * 
     */
    private void saveAs()
    {
        if (fileSaveChooser == null)
        {

            fileSaveChooser = new JFileChooser(currentPath);

            if (dlgFileFilter == null)
            {
                dlgFileFilter = new FileFilter()
                {

                    public boolean accept(File f)
                    {
                        return (f.getName().matches(".*\\.dlg") || f.isDirectory()); //$NON-NLS-1$
                    }

                    public String getDescription()
                    {
                        return currentLanguage.getString(KeysHolder.DLGFileFilterName);

                    }
                };
            }

            fileSaveChooser.addChoosableFileFilter(dlgFileFilter);
            fileSaveChooser.setAcceptAllFileFilterUsed(false);

        }

        fileSaveChooser.setMultiSelectionEnabled(false);
        if (fileSaveChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION)
        {
            File selected = new File(fileSaveChooser.getSelectedFile().getAbsolutePath());

            Object[] args = { selected };

            (new InterfaceThread("save", editorZone.getSelectedComponent(), args)).start(); //$NON-NLS-1$
        }

    }

    /* (non-Javadoc)
     * @see java.awt.event.WindowListener#windowActivated(java.awt.event.WindowEvent)
     */
    public void windowActivated(WindowEvent e)
    {
        // TODO Auto-generated method windowActivated

    }

    /* (non-Javadoc)
     * @see java.awt.event.WindowListener#windowClosed(java.awt.event.WindowEvent)
     */
    public void windowClosed(WindowEvent e)
    {
        // TODO Auto-generated method windowClosed

    }

    /* (non-Javadoc)
     * @see java.awt.event.WindowListener#windowClosing(java.awt.event.WindowEvent)
     */
    public void windowClosing(WindowEvent e)
    {
        // TODO Auto-generated method windowClosing

    }

    /* (non-Javadoc)
     * @see java.awt.event.WindowListener#windowDeactivated(java.awt.event.WindowEvent)
     */
    public void windowDeactivated(WindowEvent e)
    {
        // TODO Auto-generated method windowDeactivated

    }

    /* (non-Javadoc)
     * @see java.awt.event.WindowListener#windowDeiconified(java.awt.event.WindowEvent)
     */
    public void windowDeiconified(WindowEvent e)
    {
        // TODO Auto-generated method windowDeiconified

    }

    /* (non-Javadoc)
     * @see java.awt.event.WindowListener#windowIconified(java.awt.event.WindowEvent)
     */
    public void windowIconified(WindowEvent e)
    {
        // TODO Auto-generated method windowIconified

    }

    /* (non-Javadoc)
     * @see java.awt.event.WindowListener#windowOpened(java.awt.event.WindowEvent)
     */
    public void windowOpened(WindowEvent e)
    {
        // TODO Auto-generated method windowOpened

    }

}
