
/*
 *   Copyright 2003 Alexandre Brunel
 * 
 *   This file is part of Ghanima beta version 1.
 * 
 *   Ghanima is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */
package NWNFiles;

import java.io.*;
import java.nio.ByteOrder;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;
import java.text.Collator;
import java.util.*;

import javax.imageio.stream.FileImageOutputStream;

import NWNFiles.NWNFilesExceptions.*;
import NWNFiles.language.KeysHolder;

/**
 * <p>
 * This class allows to handle bioware ERF files.
 * </p>
 * <p>
 * 
 * </p>
 */
public class ERFFile extends NWNFile
{
    /**
     * <p>
     * The default file version.
     * </p>
     */
    protected static final String DefaultVersion = "V1.0"; //$NON-NLS-1$
    protected static final int HeaderSize = 160;

    ///////////////////////////////////////
    // attributes

    /**
     * <p>
     * All known ERF file types.
     * </p>
     */
    protected static final int KeyStructureSize = 24;
    public static final String[] KownERFFileTypes = { "ERF ", "MOD ", "SAV ", "HAK " }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
    protected static final int ResourceListEntrySize = 8;

    private static long ReadInt(byte[] data, int offset)
    {
        DataInputStream dataStream = new DataInputStream(new ByteArrayInputStream(data, offset, 4));
        try
        {
            long temp = (long) dataStream.readUnsignedByte();

            temp |= ((long) dataStream.readUnsignedByte()) << 8;
            temp |= ((long) dataStream.readUnsignedByte()) << 16;
            temp |= ((long) dataStream.readUnsignedByte()) << 24;

            return temp;
        }
        catch (java.io.IOException e)
        {
            return 0;
        }
    }
    /**
     * <p>
     * Number of days since january the 1st
     * </p>
     */
    private int buildDay;

    /**
     * <p>
     * Number of years since 1900.
     * </p>
     */
    private int buildYear;

    /**
     * <p>
     * Represents ...
     * </p>
     */
    private File file;

    /**
     * <p>
     * Represents ...
     * </p>
     */
    private String fileName;

    /**
     * <p>
     * Number of file packed in this ERF file.
     * </p>
     */
    private int fileNumber;

    /**
     * <p>
     * The type of the file.
     * </p>
     */
    private String fileType;

    /**
     * <p>
     * Curent file version.
     * </p>
     */
    private String fileVersion;

    private boolean modified;

    /**
     * <p>
     * Represents ...
     * </p>
     */
    private Collection resources; // of type Collection

    private int resRef;

    /**
     * <p>
     * Elements of localized string table.
     * </p>
     */
    private Collection stringListElement; // of type Collection

    /**
     * <p>
     * The number of entry in the localized string table.
     * </p>
     */
    private int stringTableElementNumber;

    public ERFFile()
    {
        super();
        fileVersion = DefaultVersion;
        buildDay = 0;
        buildYear = 0;
        file = null;
        fileName = ""; //$NON-NLS-1$
        fileNumber = 0;
        fileType = "ERF "; //$NON-NLS-1$
        stringTableElementNumber = 0;
        setModified(false);

    }
    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * 
     * 
     * 
     * @return a ERFFile with ...
     * </p><p>
     * @param filePath ...
     * </p>
     */
    public ERFFile(File filePath)
        throws FileDoesNotExist, InvalidFile, NameTooLong, FileNotFoundException
    {

        super();
        load(filePath);
    } // end ERFFile        

    public ERFFile(File filePath, Observer observer)
        throws FileDoesNotExist, InvalidFile, NameTooLong, FileNotFoundException
    {

        super();
        addObserver(observer);
        load(filePath);
    } // end ERFFile

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * </p><p>
     * 
     * 
     * 
     * 
     * @param filePath ...
     * </p>
     */
    public ERFResourceEntry addFile(File filePath)
        throws FileDoesNotExist, InvalidFile, NameTooLong, FileNotFoundException
    {

        if (!filePath.exists())
        {
            throw new FileDoesNotExist(filePath);
        }
        if (!filePath.isFile())
        {
            throw new InvalidFile(filePath);
        }

        String fileName = filePath.getName();

        if (fileName.length() > ERFResourceEntry.MaxFileNameLength + 4)
        {
            throw new NameTooLong(fileName);
        }

        ERFResourceEntry newResource = new ERFResourceEntry(filePath, this);

        if (resources == null)
        {
            resources = new Vector();
        }

        resources.add(newResource);

        fileNumber = resources.size();
        setModified(true);
        return newResource;
    } // end addFile        

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * </p><p>
     * 
     * 
     * 
     * 
     * @param newElement ...
     * </p>
     */
    public void addStringListElement(StringListElement newElement)
    {

        // your code here
        if (stringListElement == null)
        {
            stringListElement = new Vector();

        }

        stringListElement.add(newElement);
        stringTableElementNumber = stringListElement.size();
        setModified(true);
    } // end addStringListElement        

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * </p>
     */
    public void finalize()
    {
        // your code here
    } // end finalize        
    /**
     * <p>
     * Number of days since january the 1st
     * </p>
     */
    public int getBuildDay()
    {

        return buildDay;
    } // end getBuildDay        

    /**
     * <p>
     * Number of years since 1900.
     * </p>
     */
    public int getBuildYear()
    {
        return buildYear;
    } // end getBuildYear        

    public File getFile()
    {
        return file;
    }

    /**
     * <p>
     * Represents ...
     * </p>
     */
    public String getFileName()
    {
        return fileName;
    } // end getFileName        

    /**
     * <p>
     * Represents ...
     * </p>
     */
    public int getFileNumber()
    {

        return fileNumber;
    } // end getFileNumber        

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * 
     * 
     * 
     * @return a String with ...
     * </p>
     */
    public String getFileType()
    {

        return fileType;
    } // end getFileType        
    /**
     * <p>
     * Represents ...
     * </p>
     */

    ///////////////////////////////////////
    // operations

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * 
     * 
     * 
     * @return a String with ...
     * </p>
     */
    public String getFileVersion()
    {

        return fileVersion;
    } // end getFileVersion        

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * 
     * 
     * 
     * @return a ERFResourceEntry with ...
     * </p><p>
     * @param index ...
     * </p>
     */
    public ERFResourceEntry getResourceEntry(int index) throws ArrayIndexOutOfBoundsException
    {

        if (resources != null && !resources.isEmpty() && index < resources.size())
        {

            return (ERFResourceEntry) (resources.toArray())[index];
        }
        else
        {
            throw new ArrayIndexOutOfBoundsException();
        }
    } // end getResourceEntry        

    public int getResRef()
    {
        return resRef;
    }

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * 
     * 
     * 
     * @return a StringListElement with ...
     * </p><p>
     * @param index ...
     * </p>
     */
    public StringListElement getStringListElement(int index) throws ArrayIndexOutOfBoundsException
    {

        // your code here
        if (!stringListElement.isEmpty() && index < stringListElement.size())
        {

            return (StringListElement) (stringListElement.toArray())[index];
        }
        else
        {
            throw new ArrayIndexOutOfBoundsException();
        }
    } // end getStringListElement        

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * 
     * 
     * 
     * @return a int with ...
     * </p>
     */
    public int getStringTableElementNumber()
    {

        return stringTableElementNumber;
    } // end getStringTableElementNumber        

    public boolean isModified()
    {
        return modified;
    }

    public void load(File filePath)
        throws FileDoesNotExist, InvalidFile, NameTooLong, FileNotFoundException
    {

        FileInputStream sourceFile = new FileInputStream(filePath);

        file = filePath;

        if (!filePath.exists())
        {
            throw new FileDoesNotExist(filePath);
        }
        if (!filePath.isFile())
        {
            throw new InvalidFile(filePath);
        }

        fileName = file.getName();

        byte[] header = new byte[HeaderSize];

        try
        {
            sourceFile.read(header);
        }
        catch (IOException e)
        {
            throw new InvalidFile(filePath);
        }

        //InputStreamReader iterpretor = new InputStreamReader(new ByteArrayInputStream(header),Charset.forName(currentLanguage.getString(KeysHolder.BinaryCarset)));

        fileState.setOperation(currentLanguage.getString(KeysHolder.LoadOperation));
        fileState.setProgress((float) 0.0);
        fileState.setModified(false);
        setChanged();
        notifyObservers(fileState);
        try
        {
            fileType =
                new String(
                    header,
                    0,
                    4,
                    Charset.forName(currentLanguage.getString(KeysHolder.BinaryCarset)).name());
        }
        catch (java.io.UnsupportedEncodingException e)
        {
            throw new InvalidFile(filePath);
        }

        try
        {
            fileVersion =
                new String(
                    header,
                    4,
                    4,
                    Charset.forName(currentLanguage.getString(KeysHolder.BinaryCarset)).name());
        }
        catch (java.io.UnsupportedEncodingException e)
        {
            throw new InvalidFile(filePath);
        }

        stringTableElementNumber = (int) ReadInt(header, 8);

        int totalStringSize = (int) ReadInt(header, 12);

        fileNumber = (int) ReadInt(header, 16);

        int stringOffset = (int) ReadInt(header, 20);
        int keyOffset = (int) ReadInt(header, 24);
        int resourceOffset = (int) ReadInt(header, 28);

        buildYear = (int) ReadInt(header, 32);
        buildDay = (int) ReadInt(header, 36);

        resRef = (int) ReadInt(header, 40);

        int keySegmentSize = KeyStructureSize * fileNumber;

        if (stringOffset != HeaderSize
            || keyOffset != stringOffset + totalStringSize
            || resourceOffset != keyOffset + keySegmentSize)
        {
            throw new InvalidFile(filePath);
        }

        byte[] temp = new byte[totalStringSize];

        try
        {
            sourceFile.read(temp);
        }
        catch (IOException e)
        {
            throw new InvalidFile(filePath);
        }

        int offset = 0;
        int i = 0;

        stringListElement = new Vector(stringTableElementNumber);

        try
        {
            while (i < stringTableElementNumber && offset < totalStringSize)
            {
                int languageID = (int) ReadInt(temp, offset);
                offset += 4;

                int stringSize = (int) ReadInt(temp, offset);
                offset += 4;

                String localizedString;

                try
                {
                    localizedString =
                        new String(
                            temp,
                            offset,
                            stringSize,
                            Charset
                                .forName(currentLanguage.getString(KeysHolder.BinaryCarset))
                                .name());
                    offset += stringSize;
                }
                catch (java.io.UnsupportedEncodingException e)
                {
                    throw new InvalidFile(filePath);
                }
                stringListElement.add(new StringListElement(localizedString, languageID));
                ++i;

                fileState.setProgress((float) ((((float) i) / ((float) fileNumber)) * 0.5));

                setChanged();
                notifyObservers(fileState);
            }
        }
        catch (InvalidLanguageId e)
        {
            throw new InvalidFile(filePath);
        }
        if (i < stringTableElementNumber || offset < totalStringSize)
        {
            throw new InvalidFile(filePath);
        }

        temp = new byte[keySegmentSize];

        try
        {
            if (sourceFile.read(temp) < keySegmentSize)
            {
                throw new InvalidFile(filePath);
            }
        }
        catch (IOException e)
        {
            throw new InvalidFile(filePath);
        }

        offset = 0;

        int resourcesListSize = ResourceListEntrySize * fileNumber;

        byte[] resourcesList = new byte[resourcesListSize];

        try
        {
            if (sourceFile.read(resourcesList) < resourcesListSize)
            {
                throw new InvalidFile(filePath);
            }
        }
        catch (IOException e)
        {
            throw new InvalidFile(filePath);
        }

        i = 0;

        resources = new Vector(fileNumber);
        int fileOffset = 0;
        while (i < fileNumber)
        {
            String resRef;

            try
            {
                resRef =
                    new String(
                        temp,
                        offset,
                        16,
                        Charset.forName(currentLanguage.getString(KeysHolder.BinaryCarset)).name());
                offset += 16;
            }
            catch (java.io.UnsupportedEncodingException e)
            {
                throw new InvalidFile(filePath);
            }

            int stringEnd = resRef.indexOf("\0"); //$NON-NLS-1$

            if (stringEnd >= 0)
            {
                resRef = resRef.substring(0, stringEnd);
            }

            int resId = (int) ReadInt(temp, offset);
            offset += 4;

            if (resId != i)
            {
                throw new InvalidFile(filePath);
            }

            DataInputStream dataStream =
                new DataInputStream(new ByteArrayInputStream(temp, offset, 4));

            long fileType;
            try
            {
                fileType = (long) dataStream.readUnsignedByte();
                fileType |= (long) dataStream.readUnsignedByte() << 8;
            }
            catch (IOException e)
            {
                throw new InvalidFile(filePath);
            }

            offset += 4;
            ++i;
            int curentFileOffset = (int) ReadInt(resourcesList, fileOffset);
            fileOffset += 4;
            int fileSize = (int) ReadInt(resourcesList, fileOffset);
            fileOffset += 4;
            resources.add(
                new ERFResourceEntry(fileSize, this, curentFileOffset, resRef, (int) fileType));

            fileState.setProgress((float) ((((float) i) / ((float) fileNumber)) * 1));

            setChanged();
            notifyObservers(fileState);
        }

        fileState.setOperation(currentLanguage.getString(KeysHolder.LoadOperationDone));

        fileState.setProgress((float) 0.0);
        setChanged();
        notifyObservers(fileState);
        clearChanged();
    }

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * </p><p>
     * 
     * 
     * 
     * @param resource ...
     * </p>
     */
    public void removeFile(ERFResourceEntry resource)
    {
        if (resources != null)
        {
            resources.remove(resource);
            setModified(true);
            fileNumber = resources.size();
        }
    } // end removeFile        

    public void removeStringListElement(StringListElement element)
    {
        if (stringListElement != null)
        {
            stringListElement.remove(element);
            setModified(true);
            stringTableElementNumber = stringListElement.size();
        }
    }

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * </p>
     */
    public void save() throws java.io.IOException
    {

        saveAs(file);

    } // end save        

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * </p><p>
     * 
     * 
     * 
     * @param file ...
     * </p>
     */
    public void saveAs(File _file) throws java.io.IOException
    {

        System.gc();

        if (!file.equals(_file))
        {
            modified = true;
        }
        Iterator element = stringListElement.iterator();

        int totalStringListsize = 0;

        while (element.hasNext())
        {
            StringListElement el = (StringListElement) element.next();

            modified = modified || el.isModified();
            totalStringListsize += el.getString().length() + 8;
        }

        element = resources.iterator();

        File tempFile = null;

        if (modified)
        {
            fileState.setModified(modified);
            fileState.setOperation(currentLanguage.getString(KeysHolder.SaveOperation));
            fileState.setProgress(0.0f);

            setChanged();
            notifyObservers(fileState);

            tempFile = File.createTempFile("ERF", "temp"); //$NON-NLS-1$ //$NON-NLS-2$
            //tempFile.deleteOnExit();

            FileInputStream sourceFile = new FileInputStream(file);
            FileImageOutputStream destFile = new FileImageOutputStream(tempFile);

            FileChannel sourceChanel = sourceFile.getChannel();
            MappedByteBuffer source =
                sourceChanel.map(FileChannel.MapMode.READ_ONLY, 0, file.length());

            destFile.setByteOrder(ByteOrder.LITTLE_ENDIAN);

            //Ecriture du header:

            destFile.write(
                fileType.getBytes(currentLanguage.getString(KeysHolder.BinaryCarset)),
                0,
                4);

            destFile.write(
                DefaultVersion.getBytes(currentLanguage.getString(KeysHolder.BinaryCarset)),
                0,
                4);

            destFile.writeInt(stringListElement.size());
            destFile.writeInt(totalStringListsize);
            destFile.writeInt(resources.size());

            destFile.writeInt(HeaderSize);
            destFile.writeInt(HeaderSize + totalStringListsize);

            long offsetToResourceList =
                (long) HeaderSize
                    + (long) totalStringListsize
                    + (long) KeyStructureSize * (long) resources.size();

            destFile.writeInt((int) offsetToResourceList);

            Calendar calendar = Calendar.getInstance();

            //calendar.setTime(new Date(System.currentTimeMillis()));

            destFile.writeInt(calendar.get(Calendar.YEAR) - 1900);
            destFile.writeInt(calendar.get(Calendar.DAY_OF_YEAR));
            destFile.writeInt(resRef);

            byte[] temp = new byte[116];

            destFile.write(temp);

            //Ecriture des chanes
            fileState.setProgress(fileState.getProgress() + 1f / 6f);
            setChanged();
            notifyObservers(fileState);

            element = stringListElement.iterator();

            while (element.hasNext())
            {
                StringListElement el = (StringListElement) element.next();

                destFile.writeInt(el.getLanguageID() * 2 + el.getGender());
                destFile.writeInt(el.getString().length());
                destFile.write(
                    el.getString().getBytes(currentLanguage.getString(KeysHolder.BinaryCarset)));

            }

            fileState.setProgress(fileState.getProgress() + 1f / 6f);
            setChanged();
            notifyObservers(fileState);

            //Ecriture de la KeyList:
            element = resources.iterator();
            int resID = 0;

            byte[] strRef = new byte[16];

            while (element.hasNext())
            {
                ERFResourceEntry el = (ERFResourceEntry) element.next();

                byte[] tempERF =
                    el.getFileName().getBytes(currentLanguage.getString(KeysHolder.BinaryCarset));

                if (tempERF.length < strRef.length)
                {
                    for (int i = 0; i < tempERF.length; ++i)
                    {
                        strRef[i] = tempERF[i];
                    }

                    for (int i = tempERF.length; i < tempERF.length; ++i)
                    {
                        strRef[i] = 0;
                    }

                    tempERF = strRef;
                }

                destFile.write(tempERF, 0, 16);
                destFile.writeInt(resID++);
                destFile.writeShort((short) el.getFileType());
                destFile.writeShort(0);
            }

            fileState.setProgress(fileState.getProgress() + 1f / 6f);
            setChanged();
            notifyObservers(fileState);

            //Ecriture de la ResourceList

            element = resources.iterator();
            long offsetToData = offsetToResourceList + ResourceListEntrySize * resources.size();
            long offset = offsetToData;

            while (element.hasNext())
            {
                ERFResourceEntry el = (ERFResourceEntry) element.next();

                long size = (long) el.getSize();
                destFile.writeInt((int) offset);
                destFile.writeInt((int) size);
                offset += size;

            }

            fileState.setProgress(fileState.getProgress() + 1f / 6f);
            setChanged();
            notifyObservers(fileState);

            //Ecriture des donnes
            //source

            element = resources.iterator();
            System.gc();
            source.load();

            destFile.close();

            FileOutputStream destination = new FileOutputStream(tempFile, true);
            FileChannel destinationChannel = destination.getChannel();

            //destinationChannel.position(offsetToData);
            offset = offsetToData;
            while (element.hasNext())
            {
                ERFResourceEntry el = (ERFResourceEntry) element.next();

                if (el.isInOriginalFile())
                {
                    source.position(el.getOffset());
                    source.limit(el.getOffset() + el.getSize());
                    destinationChannel.write(source);
                }
                else
                {
                    FileInputStream dataFile = new FileInputStream(el.getFile());

                    FileChannel dataChanel = dataFile.getChannel();
                    MappedByteBuffer data =
                        dataChanel.map(FileChannel.MapMode.READ_WRITE, 0, file.length());
                    data.load();
                    destinationChannel.write(data);
                    dataFile.close();
                    el.setInOriginalFile(true);
                }

                el.setOffset((int) offset);
                offset += (long) el.getSize();
            }

            fileState.setProgress(fileState.getProgress() + 1f / 6f);
            setChanged();
            notifyObservers(fileState);

            sourceFile.close();
            destinationChannel.close();
            destination.close();

            sourceFile = new FileInputStream(tempFile);
            FileOutputStream newFile = new FileOutputStream(_file);

            sourceChanel = sourceFile.getChannel();
            source = sourceChanel.map(FileChannel.MapMode.READ_ONLY, 0, tempFile.length());
            source.load();

            destinationChannel = newFile.getChannel();

            destinationChannel.write(source);

            destinationChannel.close();
            newFile.close();
            sourceFile.close();

            fileState.setOperation(currentLanguage.getString(KeysHolder.SaveOperationEnd));
            fileState.setProgress(100.0f);
            setChanged();
            notifyObservers(fileState);

            file = _file;
        }
        fileState.setOperation(""); //$NON-NLS-1$
        fileState.setProgress(0.0f);
        setModified(false);
        System.gc();

        if (tempFile != null && tempFile.exists())
        {
            if (!tempFile.delete())
            {
                System.err.println(
                    currentLanguage.getString(KeysHolder.CannotDeleteFile) + tempFile.toString());
            }
        }
    } // end saveAs        

    /**
     * <p>
     * Represents ...
     * </p>
     */
    public void setFileName(String _fileName)
    {

        setModified(true);

        fileName = _fileName;
    } // end setFileName        

    /**
     * <p>
     * Does ...
     * </p><p>
     * 
     * 
     * 
     * 
     * @param _fileType ...
     * </p><p>
     * 
     * </p>
     */
    public void setFileType(String _fileType) throws BadFileType
    {

        int i = 0;
        boolean ok = false;

        Collator compare = Collator.getInstance();

        compare.setStrength(Collator.PRIMARY);

        while (i < KownERFFileTypes.length && !ok)
        {
            ok = (compare.compare(_fileType, KownERFFileTypes[i]) == 0);
        }
        if (ok)
        {
            fileType = _fileType;
        }
        else
        {
            throw new BadFileType(_fileType);
        }
    } // end setFileType        

    protected void setModified(boolean _modified)
    {
        modified = _modified;
        fileState.setModified(modified);
        if (_modified)
        {
            setChanged();
            notifyObservers(fileState);
        }
        else
        {
            setChanged();
            notifyObservers(fileState);
        }
    }

    public void setResRef(int _resRef)
    {
        setModified(true);
        resRef = _resRef;
    }

} // end ERFFile
