/*
 *   Copyright 2003 Alexandre Brunel
 * 
 *   This file is part of Ghanima beta version 1.
 * 
 *   Ghanima is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package NWNFiles.GFFFieldTypes;
import java.nio.ByteBuffer;
import java.util.HashMap;
import java.util.Locale;
import java.util.ResourceBundle;

import NWNFiles.language.KeysHolder;


//import java.lang.reflect.*;

/**
 * This class is an abstract class which is used to disign GFF file field types.
 */
public abstract class FieldType
{
    /**Data type*/
    public static final long BYTE_TYPE = 0;
    /**Data type*/
    public static final long CExoLocString_TYPE = 12;
    /**Data type*/
    public static final long CExoString_TYPE = 10;
    /**Data type*/
    public static final long CHAR_TYPE = 1;

    private static final String ClassName = "NWNFiles.GFFFieldTypes.FieldType"; //$NON-NLS-1$
    protected static ResourceBundle CurrentLanguage =
        ResourceBundle.getBundle(KeysHolder.LanguagesBaseName);
    /**Data type*/
    public static final long DOUBLE_TYPE = 9;
    /**Data type*/
    public static final long DWORD_TYPE = 4;
    /**Data type*/
    public static final long DWORD64_TYPE = 6;

    private static HashMap FieldTypes = null;
    /**Data type*/
    public static final long FLOAT_TYPE = 8;
    /**Data type*/
    public static final long INT_TYPE = 5;
    /**Data type*/
    public static final long INT64_TYPE = 7;
    /**Data type*/
    public static final long List_TYPE = 15;
    /**Data type*/
    public static final long RESREF_TYPE = 11;
    /**Data type*/
    public static final long SHORT_TYPE = 3;
    /**Data type*/
    public static final long Struct_TYPE = 14;
    /**Data type*/
    public static final long VOID_TYPE = 13;
    /**Data type*/
    public static final long WORD_TYPE = 2;

    /**
     * REturns the local settings used by all FildTypeClasses
     * @return A ResourceBundle.
     */
    public static ResourceBundle getCurrentLanguage()
    {
        return CurrentLanguage;
    }

    /**
      * Returns an instance of the class representing the data type.
      * @param type Data type.
      */
    public static FieldType GetInstance(long type) throws InvalidType
    {

        String typeName = GetTypeName(type);

        if (typeName != null)
        {
            String className = ClassName + typeName;

            try
            {
                Class typeClass = Class.forName(className);

                return (FieldType) typeClass.getDeclaredConstructor(null).newInstance(null);
            }
            catch (Exception e)
            {
                System.err.println(e);
                throw new InvalidType(type);
            }

        }
        else
        {
            throw new InvalidType(type);
        }

    }

    /**
      * Returns the data type name.
      * @param type Data type.
      */
    public static String GetTypeName(long type)
    {
        if (FieldTypes == null)
        {
            FieldTypes = new HashMap();

            FieldTypes.put(new Long(0), "BYTE"); //$NON-NLS-1$
            FieldTypes.put(new Long(1), "CHAR"); //$NON-NLS-1$
            FieldTypes.put(new Long(2), "WORD"); //$NON-NLS-1$
            FieldTypes.put(new Long(3), "SHORT"); //$NON-NLS-1$
            FieldTypes.put(new Long(4), "DWORD"); //$NON-NLS-1$
            FieldTypes.put(new Long(5), "INT"); //$NON-NLS-1$
            FieldTypes.put(new Long(6), "DWORD64"); //$NON-NLS-1$
            FieldTypes.put(new Long(7), "INT64"); //$NON-NLS-1$
            FieldTypes.put(new Long(8), "FLOAT"); //$NON-NLS-1$
            FieldTypes.put(new Long(9), "DOUBLE"); //$NON-NLS-1$
            FieldTypes.put(new Long(10), "CExoString"); //$NON-NLS-1$
            FieldTypes.put(new Long(11), "RESREF"); //$NON-NLS-1$
            FieldTypes.put(new Long(12), "CExoLocString"); //$NON-NLS-1$
            FieldTypes.put(new Long(13), "VOID"); //$NON-NLS-1$
            FieldTypes.put(new Long(14), "Struct"); //$NON-NLS-1$
            FieldTypes.put(new Long(15), "List"); //$NON-NLS-1$

        }

        return (String) FieldTypes.get(new Long(type));
    }

    /**
     * Change the locale settings
     * @param locale New locale
     */
    public static void setCurrentLanguage(Locale locale)
    {
        CurrentLanguage = ResourceBundle.getBundle(KeysHolder.LanguagesBaseName, locale);
    }
    private boolean isCplx;

    private String typeName;
    private long typeNumber;

    /**
      * Constructs a complex data type.
      * @param _typeNumber Data type.
      */
    public FieldType(long _typeNumber)
    {
        typeName = GetTypeName(_typeNumber);
        typeNumber = _typeNumber;
        isCplx = false;
    }

    /**
      * Constructor
      * @param _typeNumber Data type.
      * @param _isCplx A boolean telling if the data type is complex.
      */
    public FieldType(long _typeNumber, boolean _isCplx)
    {
        typeName = GetTypeName(_typeNumber);
        typeNumber = _typeNumber;
        isCplx = _isCplx;
    }
    /**
      * Returns byte array containing the data.
      * @param data The object representing the data.
      */
    public abstract byte[] getData(Object data);

    /**
      * Returns the data type name.
      */
    public final String getTypeName()
    {
        return typeName;
    }
    /**
    * Returns the data type.
    */
    public long getTypeNumber()
    {
        return typeNumber;
    }

    /**
      * Returns a object representing the data stored in the buffer.
      * @param buffer A buffer containing the data. The current position of the buffer should be the biginning of the data.
      */
    public abstract Object interpretData(ByteBuffer buffer);
    /**
      * Tells if the data type is complex.
      * @return true if the data type is complex.
      */
    public final boolean isCplx()
    {
        return isCplx;
    }

    /**
     * @param data
     * @return
     */
    public Object cloneData(Object data)
    {
    	return data;
    }
    

}
