/*
 *   Copyright 2003 Alexandre Brunel
 * 
 *   This file is part of Ghanima beta version 1.
 * 
 *   Ghanima is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package NWNFiles.GFFFieldTypes;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.util.Hashtable;
import java.util.Vector;

import NWNFiles.StringListElement;
import NWNFiles.language.KeysHolder;

/** FieldTypeCExoLocString data type*/
public class FieldTypeCExoLocString extends FieldType {
	/**Represents a CExoLocString data */
	public static class CExoLocString {

		private Hashtable byLanguage;
		private Vector byGender;
		private long stringRef;
		private NWNFiles.StringListElement subStrings[];

		/** Returns the stringRef*/
		public long getStringRef() {
			return stringRef;
		}

		/** Returns the string count*/
		public int getStringCount() {
			return subStrings.length;
		}

		/** Returns the (index+1)th string
		 * @param index index of the stringth
		 */
		public NWNFiles.StringListElement getString(int index) {
			if (index < subStrings.length) {
				return subStrings[index];
			} else {
				return null;
			}
		}

		public String toString() {
			String retval = "["; //$NON-NLS-1$
			for (int i = 0; i < subStrings.length; ++i) {
				retval += " " + subStrings[i].getString(); //$NON-NLS-1$
			}
			retval += " ]"; //$NON-NLS-1$

			return retval;
		}

		/**
		  * Construct a CExoLocString containing the localized strings.
		  * @param _subStrings The strings to store.
		  * @param _stringRef stringRef.
		  */
		public CExoLocString(
			long _stringRef,
			NWNFiles.StringListElement[] _subStrings) {
			byGender = new Vector(2);
			byGender.add(new Hashtable());
			byGender.add(new Hashtable());

			byLanguage = new Hashtable();

			if (_subStrings == null) {
				subStrings = new NWNFiles.StringListElement[0];
			} else {
				subStrings = _subStrings;
			}

			stringRef = _stringRef;

			for (int i = 0; i < subStrings.length; ++i) {

				StringListElement current = subStrings[i];

				Long laguageKey = new Long(current.getLanguageID());

				((Hashtable) byGender.get(current.getGender())).put(
					laguageKey,
					current);

				if (byLanguage.containsKey(laguageKey)) {
					((Vector) byLanguage.get(laguageKey)).set(
						current.getGender(),
						current);
				} else {
					Vector temp = new Vector(2);
					byLanguage.put(laguageKey, temp);
					temp.setSize(2);
					temp.set(current.getGender(), current);
				}
			}
		}
		
		public StringListElement getString(long languageId,int gender){
			return (StringListElement) ((Hashtable)byGender.get(gender%2)).get(new Long(languageId));
		}

	}

	FieldTypeCExoLocString() {
		super(FieldType.CExoLocString_TYPE, true);
	}

	/**
	  * Returns a CExoLocString containing the strings.
	  * @param buffer A buffer containing the data. The current position of the buffer should be the biginning of the data.
	  * @return A CExoLocString object.
	  */
	public Object interpretData(ByteBuffer buffer) {
		byte[] read = new byte[4];

		buffer.get(read);

		long totalByteSize = ((long) read[0]) & 0xFF;
		totalByteSize |= (((long) read[1]) & 0xFF) << 8;
		totalByteSize |= (((long) read[2]) & 0xFF) << 16;
		totalByteSize |= (((long) read[3]) & 0xFF) << 24;

		buffer.get(read);

		long strRef = ((long) read[0]) & 0xFF;
		strRef |= (((long) read[1]) & 0xFF) << 8;
		strRef |= (((long) read[2]) & 0xFF) << 16;
		strRef |= (((long) read[3]) & 0xFF) << 24;

		buffer.get(read);

		long strCount = ((long) read[0]) & 0xFF;
		strCount |= (((long) read[1]) & 0xFF) << 8;
		strCount |= (((long) read[2]) & 0xFF) << 16;
		strCount |= (((long) read[3]) & 0xFF) << 24;

		StringListElement[] strings = new StringListElement[(int) strCount];

		for (long i = 0; i < strCount; ++i) {
			buffer.get(read);

			long strID = ((long) read[0]) & 0xFF;
			strID |= (((long) read[1]) & 0xFF) << 8;
			strID |= (((long) read[2]) & 0xFF) << 16;
			strID |= (((long) read[3]) & 0xFF) << 24;

			buffer.get(read);

			long stringLength = ((long) read[0]) & 0xFF;
			stringLength |= (((long) read[1]) & 0xFF) << 8;
			stringLength |= (((long) read[2]) & 0xFF) << 16;
			stringLength |= (((long) read[3]) & 0xFF) << 24;

			byte[] temp = new byte[(int) stringLength];

			buffer.get(temp);

			String string;

			try {

				string =
					new String(
						temp,
						0,
						(int) stringLength,
						Charset.forName(CurrentLanguage.getString(KeysHolder.BinaryCarset)).name());
			} catch (Exception e) {
				string = null;
			}

			if (string != null) {
				try {
					strings[(int) i] =
						new StringListElement(string, (int) strID);
				} catch (NWNFiles.NWNFilesExceptions.InvalidLanguageId e) {
					strings[(int) i] = null;
				}
			} else {
				strings[(int) i] = null;
			}

		}

		return new CExoLocString(strRef, strings);
	}

	/**
	  * Returns an array containing the data, ready to be stored in the GFF file.
	  * @param data Should be a CExoLocString containing the strings to be stored.
	  */
	public byte[] getData(Object data) {
		CExoLocString cExoLocString = (CExoLocString) data;

		long totalByteSize = 12;

		for (int i = 0; i < cExoLocString.getStringCount(); ++i) {
			totalByteSize
				+= (long) cExoLocString.getString(i).getString().length()
				+ 8;
		}

		byte[] retval = new byte[(int) totalByteSize];

		ByteBuffer dataBlock = ByteBuffer.wrap(retval);

		dataBlock.put((byte) totalByteSize);
		dataBlock.put((byte) (totalByteSize >> 8));
		dataBlock.put((byte) (totalByteSize >> 16));
		dataBlock.put((byte) (totalByteSize >> 24));

		long temp = cExoLocString.getStringRef();

		dataBlock.put((byte) temp);
		dataBlock.put((byte) (temp >> 8));
		dataBlock.put((byte) (temp >> 16));
		dataBlock.put((byte) (temp >> 24));

		int stringCount = cExoLocString.getStringCount();

		dataBlock.put((byte) stringCount);
		dataBlock.put((byte) (stringCount >> 8));
		dataBlock.put((byte) (stringCount >> 16));
		dataBlock.put((byte) (stringCount >> 24));

		for (int i = 0; i < stringCount; ++i) {
			StringListElement string = cExoLocString.getString(i);

			int stringId = string.getLanguageID() * 2 + string.getGender();

			temp = (long) stringId;

			dataBlock.put((byte) temp);
			dataBlock.put((byte) (temp >> 8));
			dataBlock.put((byte) (temp >> 16));
			dataBlock.put((byte) (temp >> 24));

			temp = (long) string.getString().length();

			dataBlock.put((byte) temp);
			dataBlock.put((byte) (temp >> 8));
			dataBlock.put((byte) (temp >> 16));
			dataBlock.put((byte) (temp >> 24));

			try {
				dataBlock.put(
					string.getString().getBytes(
						Charset.forName(CurrentLanguage.getString(KeysHolder.BinaryCarset)).name()));
			} catch (java.io.UnsupportedEncodingException e) {
				dataBlock.put(new byte[string.getString().length()]);
			}
		}

		return retval;
	}
}
