/*
 *   Copyright 2003 Alexandre Brunel
 * 
 *   This file is part of Ghanima beta version 1.
 * 
 *   Ghanima is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package NWNFiles.GFFFieldTypes;
import java.nio.ByteBuffer;

import NWNFiles.GFFFile;
import NWNFiles.GFFFile.Structure;

/**
 * 15 juil. 2003
 * @author Alexandre Brunel
 *
 * Struct data type.
 */
public class FieldTypeStruct extends FieldType
{

    /**
     * 15 juil. 2003
     * @author Alexandre Brunel
     *
     * Describe a structure block from a GFF file.
     */
    public final class StructDescriptor
    {
        private long type;
        private long offset;
        private long fieldNumber;

        private StructDescriptor(long _type, long _offset, long _fieldNumber)
        {
            type = _type;
            offset = _offset;
            fieldNumber = _fieldNumber;
        }

        /**
         * Returns the type of the strucutre.
         * @return The structure type.
         */
        public long getType()
        {
            return type;
        }
        /**
         * Return the data offset of the field list or the index of the field in the field array.
         * @return The offset or index.
         */
        public long getOffset()
        {
            return offset;
        }
        /**
         * Returns the number of field in this structure.
         * @return The field number.
         */
        public long getFieldNumber()
        {
            return fieldNumber;
        }

    }

    FieldTypeStruct()
    {
        super(FieldType.Struct_TYPE, true);
    }

    /**
     * Returns a {@link NWNFiles.GFFFieldTypes.FieldTypeStruct.StructDescriptor StructDescriptor} that describe thestructure.
     * 
     * @param buffer The ByteBuffer containing the structure block.
     * @return A {@link NWNFiles.GFFFieldTypes.FieldTypeStruct.StructDescriptor StructDescriptor} object.
     * @see NWNFiles.GFFFieldTypes.FieldType#interpretData(java.nio.ByteBuffer)
     */
    public Object interpretData(ByteBuffer buffer)
    {
        byte[] read = new byte[4];
        buffer.get(read);

        long type;

        type = ((long) read[0]) & 0xFF;
        type |= (((long) read[1]) & 0xFF) << 8;
        type |= (((long) read[2]) & 0xFF) << 16;
        type |= (((long) read[3]) & 0xFF) << 24;

        buffer.get(read);

        long offset;

        offset = ((long) read[0]) & 0xFF;
        offset |= (((long) read[1]) & 0xFF) << 8;
        offset |= (((long) read[2]) & 0xFF) << 16;
        offset |= (((long) read[3]) & 0xFF) << 24;

        buffer.get(read);

        long fieldNumber;

        fieldNumber = ((long) read[0]) & 0xFF;
        fieldNumber |= (((long) read[1]) & 0xFF) << 8;
        fieldNumber |= (((long) read[2]) & 0xFF) << 16;
        fieldNumber |= (((long) read[3]) & 0xFF) << 24;

        return new StructDescriptor(type, offset, fieldNumber);
    }

    /**
     * Returns a byte array representing the structure.
     * @param data Should be a {@link NWNFiles.GFFFile.Structure Structure} object.
     * @return A byte array. 
     * @see NWNFiles.GFFFieldTypes.FieldType#getData(java.lang.Object)
     */
    public byte[] getData(Object data)
    {
        Structure structure = (Structure) data;

        byte[] retval = new byte[(int) GFFFile.StructBlockSize];

        ByteBuffer structBlock = ByteBuffer.wrap(retval);

        long temp = structure.getType();

        structBlock.put((byte) temp);
        structBlock.put((byte) (temp >> 8));
        structBlock.put((byte) (temp >> 16));
        structBlock.put((byte) (temp >> 24));

        structBlock.putInt(0);

        temp = (long) structure.getFieldCount();

        structBlock.put((byte) temp);
        structBlock.put((byte) (temp >> 8));
        structBlock.put((byte) (temp >> 16));
        structBlock.put((byte) (temp >> 24));

        return retval;
    }
    /* (non-Javadoc)
     * @see NWNFiles.GFFFieldTypes.FieldType#cloneData(java.lang.Object)
     */
    public Object cloneData(Object data)
    {

        return ((GFFFile.Structure) data).clone();

    }

}
