
/*
 *   Copyright 2003 Alexandre Brunel
 * 
 *   This file is part of Ghanima beta version 1.
 * 
 *   Ghanima is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */
package NWNFiles;

import java.io.*;
import java.nio.ByteBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;
import java.util.*;

import NWNFiles.GFFFieldTypes.FieldType;
import NWNFiles.GFFFieldTypes.FieldTypeStruct;
import NWNFiles.GFFFieldTypes.InvalidType;
import NWNFiles.NWNFilesExceptions.FileDoesNotExist;
import NWNFiles.NWNFilesExceptions.InvalidFile;
import NWNFiles.NWNFilesExceptions.NameTooLong;
import NWNFiles.language.KeysHolder;




/**
 * 14 juil. 2003
 * @author Alexandre Brunel
 *
 * This class handles GFF files.
 */
public class GFFFile extends NWNFile
{

    /**
     * @author Alexandre Brunel
     *
     * This class represents a field rom a GFF file structure.
     */
    public final class Field implements Cloneable
    {

        private Object data;

        private String label;

        private Structure owner;

        private long type;

        /**
         * Constructs a Field object.
         * 
         * _data can be null.
         * 
         * @param _type Type of the field. See the  {@link FieldType FieldType} class for a list of types. 
         * @param _label The label of the field
         * @param _data An object representing the data of the field. See the {@link FieldType FieldType} sub-classes for further information
         * @throws NameTooLong If the label is longer than 16 characters. 
         */
        public Field(long _type, String _label, Object _data) throws NameTooLong
        {
            type = _type;
            owner = null;
            setLabel(_label);
            data = _data;

        }

        /**
         * Constructs a Field object and adds it to the _owner structure. 
         * 
         * @param _type Type of the field. See the  {@link FieldType FieldType} class for a list of types.
         * @param _owner The structure containing the field. 
         * @param _label The label of the field
         * @throws NameTooLong If the label is longer than 16 characters. 
         */
        protected Field(long _type, Structure _owner, String _label) throws NameTooLong
        {
            type = _type;
            owner = _owner;
            setLabel(_label);
            data = null;
        }

        /**
         * Constructs a Field object and adds it to the _owner structure. 
         * 
         * _data can be null.
         * 
         * @param _type Type of the field. See the  {@link FieldType FieldType} class for a list of types.
         * @param _owner The structure containing the field. 
         * @param _label The label of the field
         * @param _data An object representing the data of the field. See the {@link FieldType FieldType} sub-classes for further information
         * @throws NameTooLong If the label is longer than 16 characters. 
         */
        protected Field(long _type, Structure _owner, String _label, Object _data)
            throws NameTooLong
        {
            type = _type;
            owner = _owner;
            setLabel(_label);
            data = _data;
        }
        /**
         * Returns the data contained by the field.
         * @return See the {@link FieldType FieldType} class for more details on returned object.
         */
        public Object getData()
        {
            return data;
        }
        /**
         * Returns the label of the field
         * @return A String containing the label.
         */
        public String getLabel()
        {
            return label;
        }

        /**
         * Returns the Structure containing the field.
         * @return The {@link GFFFile.Structure} Structure object containing the field.
         */
        public Structure getOwner()
        {
            return owner;
        }
        /**
         * Returns the type of the field.
         * @return A long containing the field type value. See the {@link FieldType FieldType} class for more deatails.
         */
        public long getType()
        {
            return type;
        }

        /**
         * Sets the data of the field.
         * 
         * Changes are propagated to the GFFFile object containing the field.
         * 
         * @param data The new data.
         */
        public void setData(Object data)
        {
            this.data = data;
            if (owner != null && owner.getOwner() != null)
                owner.getOwner().setIsModified(true);
        }

        /**
         * Sets the label of the field.
         * 
         * Changes are propagated to the GFFFile object containing the field.
         * @param label New label.
         * @throws NameTooLong If the label is longer than 16 characters.
         */
        public void setLabel(String label) throws NameTooLong
        {
            if (label.length() <= 16)
            {
                this.label = label;
                if (owner != null && owner.getOwner() != null)
                {
                    owner.getOwner().setIsModified(true);
                    this.label = owner.getOwner().addLabel(label);
                }
            }
            else
            {
                throw new NameTooLong(label);
            }
        }

        /**
         * Sets the structure cotaining the field.
         * 
         * If the field was in an other structure before the call of that methode, no change is made to that structure.
         * 
         * @param _owner
         */
        protected void setOwner(Structure _owner)
        {
            owner = _owner;
        }

        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        public String toString()
        {
            if (data != null)
                return label + " " //$NON-NLS-1$
                +FieldType.GetTypeName(type) + " " //$NON-NLS-1$
                +data.toString();
            else
                return label + " " + FieldType.GetTypeName(type); //$NON-NLS-1$
        }

        /* (non-Javadoc)
         * @see java.lang.Object#clone()
         */
        public Object clone()
        {
            Field clone;
            try
            {
                clone = (Field) super.clone();
            }
            catch (CloneNotSupportedException e1)
            {
                clone = null;
            }

            if (clone != null)
            {
                clone.label = label;
                clone.owner = null;
                clone.type = type;

                try
                {
                    clone.data = FieldType.GetInstance(type).cloneData(data);
                }
                catch (InvalidType e)
                {
                    clone.data = null;
                }
            }
            return clone;
        }

        /**
         * Clones the field, the return object should be used in a specific GFFFile object.
         * @param target Target GFFFile object
         * @return a copy of this object.
         */
        public Field clone(GFFFile target)
        {
            Field clone;
            try
            {
                clone = target.new Field(type, null, label, null); //(Field) super.clone();
            }
            catch (NameTooLong e)
            {
                clone = null;
                e.printStackTrace();
            }

            if (clone != null)
            {
                try
                {
                    clone.data = FieldType.GetInstance(type).cloneData(data);
                }
                catch (InvalidType e)
                {
                    clone.data = null;
                }
            }
            return clone;
        }

    }

    /**
     * @author Alexandre Brunel
     *
     * This class represents a GFF file structure.
     */
    public class Structure implements Cloneable
    {
        private Collection field;

        private GFFFile owner;

        private long type;

        /**
         * Constructs a Structure object.
         * 
         * If fields is null, then the structure will be empty.
         * 
         * @param _type Type of the structure.
         * @param fields A list of the field in the structure.
         */
        public Structure(long _type, Field[] fields)
        {
            type = _type;
            owner = null;
            field = null;

            if (fields != null)
            {
                for (int i = 0; i < fields.length; ++i)
                {
                    addField(fields[i]);
                }
            }
        }

        /**
         * Constructs a Structure object.
         * 
         * If fields is null, then the structure will be empty.
         * 
         * @param _type Type of the structure.
         * @param _owner The gff file to which belong the structure.
         * @param fields A list of the field in the structure.
         */
        protected Structure(long _type, GFFFile _owner, Field[] fields)
        {
            type = _type;
            owner = _owner;
            field = null;

            if (fields != null)
            {
                for (int i = 0; i < fields.length; ++i)
                {
                    addField(fields[i]);
                }
            }
        }
        /**
         * Adds a field to the structures.
         * 
         * Changes to a structure are notified to its owner. The new field is added to the GFF file field list,
         * so is its label.
         * 
         * @param newField The field to be added to the structure.
         */
        public void addField(Field newField)
        {
            newField.setOwner(this);
            if (field == null)
            {
                field = new Vector();
            }

            field.add(newField);

            if (owner != null)
            {
                owner.setIsModified(true);
                //owner.setFieldCount(owner.getFieldCount()+1);
                owner.addField(newField);
                newField.label = owner.addLabel(newField.getLabel());
            }

        }

        /**
         * Creates and adds a field to the structures.
         * 
         * Changes to a structure are notified to its owner. The new field is added to the GFF file field list,
         * so is its label. _data can be null.
         * 
         * @param _type The field type, see the {@link FieldType FieldType} class for further information.  
         * @param _label The field label.
         * @param _data The data contained by the field.
         * @return Returns the newly created field.
         * @throws NameTooLong if the label is longer than 16 characters.
         */
        public Field addField(long _type, String _label, Object _data) throws NameTooLong
        {
            Field newField = new Field(_type, this, _label, _data);

            addField(newField);

            return newField;

        }
        /**
         * REturns the index+1th field.
         *   
         * @param index index of the field to be returned.
         * @return A Field object or null if index is gretter than the number of field in the structure. 
         */
        public Field getField(int index)
        {
            if (field != null && !field.isEmpty() && field.size() > index)
            {
                return (Field) ((field.toArray())[index]);
            }
            else
            {
                return null;
            }
        }
        /**
         * Returns the field whose label is label.
         * 
         * @param label the label of the field to be returned.
         * @return A Field object or null if no field matches the given label.
         */
        public Field getField(String label)
        {
            if (field != null && !field.isEmpty())
            {
                Iterator i = field.iterator();

                boolean found = false;

                Field cur = null;

                while (i.hasNext() && !found)
                {
                    cur = (Field) i.next();
                    if (cur.getLabel().equals(label))
                    {
                        found = true;
                    }
                }
                if (found)
                {
                    return cur;
                }
                else
                {
                    return null;
                }
            }
            else
            {
                return null;
            }
        }
        public int getFieldCount()
        {
            if (field != null)
                return field.size();
            else
                return 0;
        }

        public GFFFile getOwner()
        {
            return owner;
        }
        public long getType()
        {
            return type;
        }
        protected void setOwner(GFFFile _owner)
        {
            owner = _owner;
        }

        public String toString()
        {
            return field.toString();
        }
        /* (non-Javadoc)
         * @see java.lang.Object#clone()
         */
        public Object clone()
        {
            Structure clone;
            try
            {
                clone = (Structure) super.clone();
            }
            catch (CloneNotSupportedException e)
            {
                clone = null;
            }

            if (clone != null)
            {

                if (field != null)
                {
                    clone.field = new Vector(field.size());

                    Iterator i = field.iterator();

                    while (i.hasNext())
                    {
                        Field element = (Field) i.next();
                        Field newField = (Field) element.clone();

                        if (newField != null)
                        {
                            clone.field.add(newField);
                        }
                    }
                }
                else
                {
                    clone.field = null;
                }

                clone.owner = null;
                clone.type = type;
            }

            return clone;
        }

        /**
         * Creates a copy of this object to be used in the given GFFFile. 
         * @param target Target GFFFile.
         * @return A copy of this object
         */
        public Object clone(GFFFile target)
        {
            Structure clone = target.new Structure(type, null);

            if (clone != null)
            {

                if (field != null)
                {
                    clone.field = new Vector(field.size());

                    Iterator i = field.iterator();

                    while (i.hasNext())
                    {
                        Field element = (Field) i.next();
                        Field newField = (Field) element.clone();

                        if (newField != null)
                        {
                            clone.field.add(newField);
                        }
                    }
                }
                else
                {
                    clone.field = null;
                }

                clone.owner = null;

            }

            return clone;
        }

    }

    /**
     * Default GFF file version.
     */
    public static final String DefaultVersion = "V3.2"; //$NON-NLS-1$

    /**
     * Sise, in bytes, of a field block in a gff file.
     */
    public static final long FieldBlockSize = 12;

    /**
     * Size, in bytes, of a gff file header.
     */
    public static final long HeaderSize = 56;

    /**
     * List of the known gff file extensions. Extensions are in uppercasse.
     */
    public static final String[] KnownGFFExtensions = { "GFF", //$NON-NLS-1$
        "IFO", //$NON-NLS-1$
        "ARE", //$NON-NLS-1$
        "GIC", //$NON-NLS-1$
        "GIT", //$NON-NLS-1$
        "UTC", //$NON-NLS-1$
        "UTD", //$NON-NLS-1$
        "UTE", //$NON-NLS-1$
        "UTI", //$NON-NLS-1$
        "UTP", //$NON-NLS-1$
        "UTS", //$NON-NLS-1$
        "UTM", //$NON-NLS-1$
        "UTT", //$NON-NLS-1$
        "UTW", //$NON-NLS-1$
        "DLG", //$NON-NLS-1$
        "JRL", //$NON-NLS-1$
        "FAC", //$NON-NLS-1$
        "ITP", //$NON-NLS-1$
        "PTM", //$NON-NLS-1$
        "PTT", //$NON-NLS-1$
        "BIC" }; //$NON-NLS-1$

    /**
     * Size, in bytes, of a label block in the gff file.
     */
    public static final long LabelBlockSize = 16;

    /**
     * Size, in bytes, of a sctructure block in the gff file.
     */
    public static final long StructBlockSize = 12;

    private static long ReadInt(byte[] data)
    {
        return ReadInt(data, 0);
    }
    private static long ReadInt(byte[] data, int offset)
    {
        DataInputStream dataStream = new DataInputStream(new ByteArrayInputStream(data, offset, 4));
        try
        {
            long temp = (long) dataStream.readUnsignedByte();

            temp |= ((long) dataStream.readUnsignedByte()) << 8;
            temp |= ((long) dataStream.readUnsignedByte()) << 16;
            temp |= ((long) dataStream.readUnsignedByte()) << 24;

            return temp;
        }
        catch (java.io.IOException e)
        {
            return 0;
        }
    }

    private Hashtable labelRefCount;
    private long fieldCount;

    private long fieldDataCount;

    private long fieldIndiciesCount;
    private Collection fieldList;

    private File filePath;

    private String fileType;

    private String fileVersion;

    private long labelCount;
    private Collection labelList;

    private long listIndiciesCount;

    private boolean modified;

    private long structCount; //Tjrs >= 1

    private Collection structs;

    /**
     * Constructs GFFFile object, this file has the default root structure of type 0xFFFFFFFF 
     * 
     */
    public GFFFile()
    {
        super();

        fileType = "GFF "; //$NON-NLS-1$

        fileVersion = DefaultVersion;

        structCount = 0;

        fieldCount = 0;

        labelCount = 0;

        fieldDataCount = 0;

        fieldIndiciesCount = 0;

        listIndiciesCount = 0;

        structs = new Vector();
        labelList = new Vector();
        labelRefCount = new Hashtable();
        fieldList = new Vector();

        addStructure(0xFFFFFFFFl, null);
    }

    /**
     * Constructs a GFFFile object.
     * 
     * The file represented by the argument of the constructor is parsed to get the gff file contence.
     * 
     * @param file The file to parse.
     * @throws FileDoesNotExist If the file cannot be found.
     * @throws InvalidFile If the file is not a valid gff file.
     * @throws NameTooLong If the file is not a valid gff file.
     * @throws FileNotFoundException If the file cannot be found.
     * @throws java.io.IOException If the file is not a valid gff file.
     * @throws InvalidType If the file is not a valid gff file.
     */
    public GFFFile(File file)
        throws
            FileDoesNotExist,
            InvalidFile,
            NameTooLong,
            FileNotFoundException,
            java.io.IOException,
            InvalidType
    {
        super();

        load(file);
        setIsModified(false);
    }

    /**
     * Constructs a GFFFile object.
     * 
     * The file represented by the argument of the constructor is parsed to get the gff file contence.
     * An observer can be assigned to the file to monitor its modifications. The GFFFile can pass a
     * {@link NWNFile.FileState FileState} object when it's modified.
     * 
     * @param file The file to parse.
     * @param observer An observer.
     * 
     * @throws FileDoesNotExist If the file cannot be found.
     * @throws InvalidFile If the file is not a valid gff file.
     * @throws NameTooLong If the file is not a valid gff file.
     * @throws FileNotFoundException If the file cannot be found.
     * @throws java.io.IOException If the file is not a valid gff file.
     * @throws InvalidType If the file is not a valid gff file.
     */
    public GFFFile(File file, Observer observer)
        throws
            FileDoesNotExist,
            InvalidFile,
            NameTooLong,
            FileNotFoundException,
            java.io.IOException,
            InvalidType
    {
        super();
        addObserver(observer);
        load(file);
        setIsModified(false);
    }

    /**
     * Constructs a GFFFile object, this file has the default root structure of type 0xFFFFFFFF 
     * 
     * An observer can be assigned to the file to monitor its modifications. The GFFFile can pass a
     * {@link NWNFile.FileState FileState} object when it's modified.
     * 
     * @param observer An observer.
     * 
     * @throws FileDoesNotExist If the file cannot be found.
     * @throws InvalidFile If the file is not a valid gff file.
     * @throws NameTooLong If the file is not a valid gff file.
     * @throws FileNotFoundException If the file cannot be found.
     * @throws java.io.IOException If the file is not a valid gff file.
     * @throws InvalidType If the file is not a valid gff file.
     */
    public GFFFile(Observer observer)
    {
        super();

        fileType = "GFF "; //$NON-NLS-1$

        fileVersion = DefaultVersion;

        structCount = 0;

        fieldCount = 0;

        labelCount = 0;

        fieldDataCount = 0;

        fieldIndiciesCount = 0;

        listIndiciesCount = 0;

        structs = new Vector();
        labelList = new Vector();
        labelRefCount = new Hashtable();
        fieldList = new Vector();

        addObserver(observer);

        addStructure(0xFFFFFFFFl, null);
    }

    protected void addField(Field newField)
    {
        fieldList.add(newField);

        fieldCount = fieldList.size();
    }
    protected String addLabel(String label)
    {
        String retval = label;
        if (labelList == null)
        {
            labelList = new Vector();
            labelList.add(label);
            labelRefCount = new Hashtable();
            labelRefCount.put(label, new Long(1));
        }
        else
        {
            int index = 0;
            if ((index = ((Vector) labelList).indexOf(label)) == -1)
            {
                labelList.add(label);
                labelRefCount.put(label, new Long(1));
            }
            else
            {
                retval = (String) ((Vector) labelList).get(index);
                labelRefCount.put(
                    label,
                    new Long(((Long) labelRefCount.get(label)).longValue() + 1l));
            }
        }

        labelCount = labelList.size();

        return retval;
    }
    public Structure addStructure(long type, Field[] fields)
    {
        Structure newStructure = new Structure(type, fields);

        addStructure(newStructure);

        return newStructure;
    }

    public void addStructure(Structure newStruct)
    {
        if (structs == null)
        {
            structs = new Vector();
        }

        structs.add(newStruct);
        structCount = structs.size();
        //fieldCount+= newStruct.getFieldCount();

        for (int i = 0; i < newStruct.getFieldCount(); ++i)
        {
            Field newStructField = newStruct.getField(i);

            fieldList.add(newStructField);

            newStruct.getField(i).label = addLabel(newStruct.getField(i).getLabel());
        }
        fieldCount = fieldList.size();
        newStruct.setOwner(this);
        setIsModified(true);
    }
    public long getFieldCount()
    {
        return fieldCount;
    }
    public long getFieldDataCount()
    {
        return fieldDataCount;
    }
    public long getFieldIndiciesCount()
    {
        return fieldIndiciesCount;
    }

    public File getFilePath()
    {
        return filePath;
    }

    public String getFileType()
    {
        return fileType;
    }
    public String getFileVersion()
    {
        return fileVersion;
    }
    public long getLabelCount()
    {
        return labelCount;
    }
    public long getListIndiciesCount()
    {
        return listIndiciesCount;
    }
    public Structure getStruct(int index)
    {
        if (structs != null && !structs.isEmpty() && index < structs.size())
        {

            return (Structure) (structs.toArray())[index];

        }
        else
        {
            return null;
        }
    }
    public long getStructCount()
    {
        return structCount;
    }

    public boolean isModified()
    {
        return modified;
    }

    public void load(File file)
        throws
            FileDoesNotExist,
            InvalidFile,
            NameTooLong,
            FileNotFoundException,
            java.io.IOException,
            InvalidType
    {
        filePath = file;

        if (!filePath.exists())
        {
            throw new FileDoesNotExist(filePath);
        }
        if (!filePath.isFile())
        {
            throw new InvalidFile(filePath);
        }

        FileInputStream sourceFile = new FileInputStream(filePath);
        FileChannel sourceChannel = sourceFile.getChannel();

        MappedByteBuffer header;

        header = sourceChannel.map(FileChannel.MapMode.READ_ONLY, 0, HeaderSize);

        structs = new Vector();
        labelList = new Vector();
        labelRefCount = new Hashtable();
        fieldList = new Vector();

        byte[] temp = new byte[4];
        try
        {
            header.get(temp);
            fileType =
                new String(
                    temp,
                    0,
                    4,
                    Charset.forName(currentLanguage.getString(KeysHolder.BinaryCarset)).name());
        }
        catch (java.io.UnsupportedEncodingException e)
        {
            throw new InvalidFile(filePath);
        }

        try
        {
            header.get(temp);
            fileVersion =
                new String(
                    temp,
                    0,
                    4,
                    Charset.forName(currentLanguage.getString(KeysHolder.BinaryCarset)).name());
        }
        catch (java.io.UnsupportedEncodingException e)
        {
            throw new InvalidFile(filePath);
        }
        header.get(temp);

        long structOffset = ReadInt(temp, 0);

        if (structOffset < HeaderSize)
        {
            throw new InvalidFile(filePath);
        }

        header.get(temp);
        structCount = ReadInt(temp, 0);

        header.get(temp);
        long fieldOffset = ReadInt(temp, 0);

        if (fieldOffset < structOffset + structCount * StructBlockSize)
        {
            throw new InvalidFile(filePath);
        }

        header.get(temp);
        fieldCount = ReadInt(temp, 0);

        header.get(temp);
        long labelOffset = ReadInt(temp, 0);

        if (labelOffset < fieldOffset + fieldCount * FieldBlockSize)
        {
            throw new InvalidFile(filePath);
        }

        header.get(temp);
        labelCount = ReadInt(temp, 0);

        header.get(temp);
        long fieldDataOffset = ReadInt(temp, 0);

        if (fieldDataOffset < labelOffset + labelCount * LabelBlockSize)
        {
            throw new InvalidFile(filePath);
        }

        header.get(temp);
        fieldDataCount = ReadInt(temp, 0);

        header.get(temp);
        long fieldIndiciesOffset = ReadInt(temp, 0);

        if (fieldIndiciesOffset < fieldDataOffset + fieldDataCount)
        {
            throw new InvalidFile(filePath);
        }

        header.get(temp);
        fieldIndiciesCount = ReadInt(temp, 0);

        header.get(temp);
        long listIndiciesOffset = ReadInt(temp, 0);

        if (listIndiciesOffset < fieldIndiciesOffset + fieldIndiciesCount)
        {
            throw new InvalidFile(filePath);
        }

        header.get(temp);
        listIndiciesCount = ReadInt(temp, 0);

        if (listIndiciesOffset + listIndiciesCount > filePath.length())
        {
            throw new InvalidFile(filePath);
        }

        MappedByteBuffer structArray =
            sourceChannel.map(
                FileChannel.MapMode.READ_ONLY,
                structOffset,
                structCount * StructBlockSize);
        MappedByteBuffer fieldArray =
            sourceChannel.map(
                FileChannel.MapMode.READ_ONLY,
                fieldOffset,
                fieldCount * FieldBlockSize);
        MappedByteBuffer labelArray =
            sourceChannel.map(
                FileChannel.MapMode.READ_ONLY,
                labelOffset,
                labelCount * LabelBlockSize);
        MappedByteBuffer fieldDataBlock =
            sourceChannel.map(FileChannel.MapMode.READ_ONLY, fieldDataOffset, fieldDataCount);
        MappedByteBuffer filedIndiciesArray =
            sourceChannel.map(
                FileChannel.MapMode.READ_ONLY,
                fieldIndiciesOffset,
                fieldIndiciesCount);
        MappedByteBuffer listIndiciesArray =
            sourceChannel.map(FileChannel.MapMode.READ_ONLY, listIndiciesOffset, listIndiciesCount);

        //structArray.get(temp);

        structCount = fieldCount = labelCount = 0;

        FieldTypeStruct.StructDescriptor curentDescriptor =
            (FieldTypeStruct.StructDescriptor) FieldType.GetInstance(
                FieldType.Struct_TYPE).interpretData(
                structArray);

        Structure root = addStructure(curentDescriptor.getType(), null);
        Structure curent = root;

        Stack structureStack = new Stack();
        Stack descriptorstack = new Stack();
        Stack progressStack = new Stack();

        boolean finished = false;
        int structFieldCount = (int) curentDescriptor.getFieldNumber();
        int fieldIndex = 0;

        while (!finished)
        {
            if (fieldIndex == structFieldCount)
            {
                if (structureStack.empty())
                {
                    finished = true;
                }
                else
                {
                    curent = (Structure) structureStack.pop();
                    curentDescriptor = (FieldTypeStruct.StructDescriptor) descriptorstack.pop();
                    fieldIndex = ((Integer) progressStack.pop()).intValue();
                    structFieldCount = (int) curentDescriptor.getFieldNumber();
                }
            }
            else
            {
                ByteBuffer fieldLocation;

                if (structFieldCount == 1)
                {
                    fieldLocation =
                        (ByteBuffer) fieldArray.position(
                            (int) (curentDescriptor.getOffset() * FieldBlockSize));
                }
                else
                {
                    (
                        (ByteBuffer) filedIndiciesArray.position(
                            (int) (curentDescriptor.getOffset() + 4 * fieldIndex))).get(
                        temp);

                    long offset = ReadInt(temp);

                    fieldLocation =
                        (ByteBuffer) fieldArray.position((int) (offset * FieldBlockSize));
                }

                fieldLocation.get(temp);

                FieldType fieldType = FieldType.GetInstance(ReadInt(temp));
                fieldLocation.get(temp);

                long labelIndex = ReadInt(temp);

                String label;

                try
                {
                    byte[] tempString = new byte[(int) LabelBlockSize];
                    labelArray.position((int) (labelIndex * LabelBlockSize));
                    labelArray.get(tempString);
                    label =
                        new String(
                            tempString,
                            0,
                            (int) LabelBlockSize,
                            Charset
                                .forName(currentLanguage.getString(KeysHolder.BinaryCarset))
                                .name());
                }
                catch (java.io.UnsupportedEncodingException e)
                {
                    throw new InvalidFile(filePath);
                }

                int stringEnd;

                if ((stringEnd = label.indexOf("\0")) >= 0) //$NON-NLS-1$
                {
                    label = label.substring(0, stringEnd);
                }

                if (fieldType.isCplx())
                {
                    fieldLocation.get(temp);

                    long dataOffset = ReadInt(temp);

                    if (fieldType.getTypeNumber() == FieldType.Struct_TYPE)
                    {
                        structureStack.push(curent);
                        descriptorstack.push(curentDescriptor);
                        progressStack.push(new Integer(fieldIndex + 1));

                        curentDescriptor =
                            (FieldTypeStruct.StructDescriptor) fieldType.interpretData(
                                (ByteBuffer) structArray.position(
                                    (int) (dataOffset * StructBlockSize)));

                        curent =
                            (Structure) curent
                                .addField(
                                    fieldType.getTypeNumber(),
                                    label,
                                    addStructure(curentDescriptor.getType(), null))
                                .getData();

                        structFieldCount = (int) curentDescriptor.getFieldNumber();
                        fieldIndex = 0;

                    }
                    else if (fieldType.getTypeNumber() == FieldType.List_TYPE)
                    {
                        Vector fieldStructList =
                            (Vector) fieldType.interpretData(
                                (ByteBuffer) listIndiciesArray.position((int) dataOffset));

                        Iterator i = fieldStructList.iterator();

                        //Structure owner = curent;

                        Vector fieldData = new Vector(fieldStructList.size());

                        curent.addField(fieldType.getTypeNumber(), label, fieldData);
                        fieldIndex++;
                        while (i.hasNext())
                        {
                            structureStack.push(curent);
                            descriptorstack.push(curentDescriptor);
                            progressStack.push(new Integer(fieldIndex));

                            long offset = ((Long) i.next()).longValue();

                            curentDescriptor =
                                (FieldTypeStruct.StructDescriptor) FieldType.GetInstance(
                                    FieldType.Struct_TYPE).interpretData(
                                    (ByteBuffer) structArray.position(
                                        (int) (offset * StructBlockSize)));
                            curent = addStructure(curentDescriptor.getType(), null);

                            fieldData.add(curent);

                            structFieldCount = (int) curentDescriptor.getFieldNumber();
                            fieldIndex = 0;
                        }
                    }
                    else
                    {
                        curent.addField(
                            fieldType.getTypeNumber(),
                            label,
                            fieldType.interpretData(
                                (ByteBuffer) fieldDataBlock.position((int) dataOffset)));
                        fieldIndex++;

                    }
                }
                else
                {
                    curent.addField(
                        fieldType.getTypeNumber(),
                        label,
                        fieldType.interpretData(fieldLocation));
                    fieldIndex++;
                }
            }
        }

        header =
            structArray =
                fieldArray =
                    labelArray = fieldDataBlock = filedIndiciesArray = listIndiciesArray = null;

        sourceChannel = null;
        sourceFile.close();

        System.gc();
    }

    public void save() throws java.io.IOException, InvalidType
    {
        save(filePath);
    }

    public void save(File newFilePath) throws java.io.IOException, InvalidType
    {
        if (!newFilePath.equals(filePath))
        {
            modified = true;
        }

        File tempFile = null;

        if (modified)
        {
            fileState.setModified(modified);
            fileState.setOperation(currentLanguage.getString(KeysHolder.SaveOperation));
            fileState.setProgress(0.0f);
            setChanged();
            notifyObservers(fileState);

            tempFile = File.createTempFile("GFF", "temp"); //$NON-NLS-1$ //$NON-NLS-2$
            //tempFile.deleteOnExit();

            //FileInputStream sourceFile = new FileInputStream(filePath);
            FileOutputStream destFile = new FileOutputStream(tempFile);
            FileChannel destChannel = destFile.getChannel();
            //FileChannel sourceChanel = sourceFile.getChannel();
            //MappedByteBuffer source = sourceChanel.map(FileChannel.MapMode.READ_ONLY,0,file.length());

            byte[] headerBlock = new byte[(int) HeaderSize];

            ByteBuffer headerBuffer = ByteBuffer.wrap(headerBlock);

            headerBuffer.put(
                fileType.getBytes(currentLanguage.getString(KeysHolder.BinaryCarset)),
                0,
                4);
            headerBuffer.put(
                fileVersion.getBytes(currentLanguage.getString(KeysHolder.BinaryCarset)),
                0,
                4);

            FieldType DWORDType = FieldType.GetInstance(FieldType.DWORD_TYPE);

            long offset = HeaderSize;

            headerBuffer.put(DWORDType.getData(new Long(HeaderSize)));
            headerBuffer.put(DWORDType.getData(new Long((long) structs.size())));
            offset += (long) structs.size() * StructBlockSize;

            headerBuffer.put(DWORDType.getData(new Long(offset)));
            headerBuffer.put(DWORDType.getData(new Long((long) fieldList.size())));
            offset += (long) fieldList.size() * FieldBlockSize;

            headerBuffer.put(DWORDType.getData(new Long(offset)));
            headerBuffer.put(DWORDType.getData(new Long((long) labelList.size())));
            offset += (long) labelList.size() * LabelBlockSize;

            headerBuffer.put(DWORDType.getData(new Long(offset)));

            ByteBuffer structBlock =
                ByteBuffer.wrap(new byte[(int) (structs.size() * StructBlockSize)]);
            ByteBuffer fieldBlock =
                ByteBuffer.wrap(new byte[(int) (fieldList.size() * FieldBlockSize)]);
            ByteBuffer labelBlock =
                ByteBuffer.wrap(new byte[(int) (labelList.size() * LabelBlockSize)]);

            ByteBuffer fieldIndiciesBlock = ByteBuffer.wrap(new byte[fieldList.size() * 4]);
            ByteBuffer listIndiciesBlock = ByteBuffer.wrap(new byte[structs.size() * 8]);

            ByteArrayOutputStream fieldDataBlock = new ByteArrayOutputStream();

            long fieldDataBlockOffset = 0;
            //long fieldIndiciesBlockOffset = 0;
            //long listIndiciesBlockOffset  = 0;

            Iterator structIterator = structs.iterator();

            FieldType structType = FieldType.GetInstance(FieldType.Struct_TYPE);

            long fieldIndex = 0;
            int writtenField = 0;
            while (structIterator.hasNext())
            {
                Structure cur = (Structure) structIterator.next();

                byte[] temp = structType.getData(cur);

                ByteBuffer structure = (ByteBuffer) ByteBuffer.wrap(temp).position(4);

                if (cur.getFieldCount() > 1)
                {
                    structure.put(
                        DWORDType.getData(new Long((long) fieldIndiciesBlock.position())));

                    structBlock.put(temp);

                    for (int i = 0; i < cur.getFieldCount(); ++i)
                    {
                        fieldIndiciesBlock.put(DWORDType.getData(new Long(fieldIndex++)));

                        Field field = cur.getField(i);
                        FieldType fieldType = FieldType.GetInstance(field.getType());

                        fieldBlock.put(DWORDType.getData(new Long(field.getType())));
                        fieldBlock.put(
                            DWORDType.getData(
                                new Long((long) ((Vector) labelList).indexOf(field.getLabel()))));

                        writtenField++;

                        if (fieldType.isCplx())
                        {
                            if (fieldType.getTypeNumber() == FieldType.Struct_TYPE)
                            {
                                fieldBlock.put(
                                    DWORDType.getData(
                                        new Long(
                                            (long) ((Vector) structs).indexOf(field.getData()))));
                            }
                            else if (fieldType.getTypeNumber() == FieldType.List_TYPE)
                            {
                                fieldBlock.put(
                                    DWORDType.getData(
                                        new Long((long) listIndiciesBlock.position())));
                                Vector list = new Vector(((Vector) field.getData()).size());

                                Iterator listIterator = ((Vector) field.getData()).iterator();

                                while (listIterator.hasNext())
                                {
                                    list.add(
                                        new Long(
                                            (long) ((Vector) structs).indexOf(
                                                listIterator.next())));
                                }

                                byte[] listData = fieldType.getData(list);
                                if (listIndiciesBlock.remaining() < listData.length)
                                {
                                    byte[] tempIndicies = listIndiciesBlock.array();

                                    byte[] newSupport = new byte[tempIndicies.length * 2 + 1];

                                    int position = listIndiciesBlock.position();

                                    for (int j = 0; j < tempIndicies.length; j++)
                                    {
                                        newSupport[j] = tempIndicies[j];
                                    }

                                    listIndiciesBlock = ByteBuffer.wrap(newSupport);
                                    listIndiciesBlock.position(position);
                                }

                                listIndiciesBlock.put(listData);
                            }
                            else
                            {
                                fieldBlock.put(DWORDType.getData(new Long(fieldDataBlockOffset)));

                                temp = fieldType.getData(field.getData());

                                fieldDataBlock.write(temp);
                                fieldDataBlockOffset += temp.length;
                            }
                        }
                        else
                        {
                            byte[] simpleData = fieldType.getData(field.getData());

                            fieldBlock.put(simpleData);

                            if (simpleData.length < 4)
                            {
                                fieldBlock.position(fieldBlock.position() + 4 - simpleData.length);
                            }
                        }
                    }
                }
                else
                {
                    structure.put(DWORDType.getData(new Long(fieldIndex++)));
                    structBlock.put(temp);

                    Field field = cur.getField(0);
                    FieldType fieldType = FieldType.GetInstance(field.getType());

                    fieldBlock.put(DWORDType.getData(new Long(field.getType())));
                    fieldBlock.put(
                        DWORDType.getData(
                            new Long((long) ((Vector) labelList).indexOf(field.getLabel()))));

                    writtenField++;

                    if (fieldType.isCplx())
                    {
                        if (fieldType.getTypeNumber() == FieldType.Struct_TYPE)
                        {
                            fieldBlock.put(
                                DWORDType.getData(
                                    new Long((long) ((Vector) structs).indexOf(field.getData()))));
                        }
                        else if (fieldType.getTypeNumber() == FieldType.List_TYPE)
                        {
                            fieldBlock.put(
                                DWORDType.getData(new Long((long) listIndiciesBlock.position())));
                            Vector list = new Vector(((Vector) field.getData()).size());

                            Iterator listIterator = ((Vector) field.getData()).iterator();

                            while (listIterator.hasNext())
                            {
                                list.add(
                                    new Long(
                                        (long) ((Vector) structs).indexOf(listIterator.next())));
                            }

                            byte[] listData = fieldType.getData(list);
                            if (listIndiciesBlock.remaining() < listData.length)
                            {
                                byte[] tempIndicies = listIndiciesBlock.array();

                                byte[] newSupport = new byte[tempIndicies.length * 2 + 1];

                                int position = listIndiciesBlock.position();

                                for (int i = 0; i < tempIndicies.length; i++)
                                {
                                    newSupport[i] = tempIndicies[i];
                                }

                                listIndiciesBlock = ByteBuffer.wrap(newSupport);
                                listIndiciesBlock.position(position);
                            }
                            listIndiciesBlock.put(listData);
                        }
                        else
                        {
                            fieldBlock.put(DWORDType.getData(new Long(fieldDataBlockOffset)));

                            temp = fieldType.getData(field.getData());

                            fieldDataBlock.write(temp);
                            fieldDataBlockOffset += temp.length;
                        }
                    }
                    else
                    {
                        byte[] simpleData = fieldType.getData(field.getData());

                        fieldBlock.put(simpleData);

                        if (simpleData.length < 4)
                        {
                            fieldBlock.position(fieldBlock.position() + 4 - simpleData.length);
                        }
                    }
                }
            }

            fileState.setProgress(0.25f);
            setChanged();
            notifyObservers(fileState);

            headerBuffer.put(DWORDType.getData(new Long(fieldDataBlockOffset)));
            offset += fieldDataBlockOffset;

            headerBuffer.put(DWORDType.getData(new Long(offset)));
            headerBuffer.put(DWORDType.getData(new Long((long) fieldIndiciesBlock.position())));
            offset += (long) fieldIndiciesBlock.position();

            headerBuffer.put(DWORDType.getData(new Long(offset)));
            headerBuffer.put(DWORDType.getData(new Long((long) listIndiciesBlock.position())));
            offset += (long) listIndiciesBlock.position();

            Iterator labelIterator = labelList.iterator();

            while (labelIterator.hasNext())
            {
                byte[] temp =
                    ((String) labelIterator.next()).getBytes(
                        currentLanguage.getString(KeysHolder.BinaryCarset));

                if (temp.length < LabelBlockSize)
                {
                    ByteBuffer tempBuffer = ByteBuffer.wrap(new byte[16]);

                    tempBuffer.put(temp);

                    for (int i = temp.length; i < LabelBlockSize; ++i)
                    {
                        tempBuffer.put((byte) 0);
                    }

                    labelBlock.put(tempBuffer.array());
                }
                else
                {
                    labelBlock.put(temp, 0, 16);
                }

            }

            fileState.setProgress(50.0f);
            setChanged();
            notifyObservers(fileState);

            destChannel.write((ByteBuffer) headerBuffer.position(0));
            destChannel.write((ByteBuffer) structBlock.position(0));
            destChannel.write((ByteBuffer) fieldBlock.position(0));
            destChannel.write((ByteBuffer) labelBlock.position(0));

            destChannel.write(
                (ByteBuffer) ByteBuffer.wrap(fieldDataBlock.toByteArray()).limit(
                    (int) fieldDataBlockOffset));

            destChannel.write(
                (ByteBuffer) fieldIndiciesBlock.limit(
                    (int) fieldIndiciesBlock.position()).position(
                    0));
            destChannel.write(
                (ByteBuffer) listIndiciesBlock.limit((int) listIndiciesBlock.position()).position(
                    0));

            headerBuffer = null;
            structBlock = null;
            fieldBlock = null;
            labelBlock = null;

            fieldDataBlock = null;

            fieldIndiciesBlock = null;
            listIndiciesBlock = null;

            destChannel.close();
            destChannel = null;
            destFile.close();

            fileState.setProgress(75.0f);
            setChanged();
            notifyObservers(fileState);

            destFile = new FileOutputStream(newFilePath);
            destChannel = destFile.getChannel();

            FileInputStream source = new FileInputStream(tempFile);
            FileChannel sourceChannel = source.getChannel();

            MappedByteBuffer sourceBuffer =
                sourceChannel.map(FileChannel.MapMode.READ_ONLY, 0, sourceChannel.size());
            sourceBuffer.load();

            destChannel.write(sourceBuffer);
            sourceBuffer = null;
            sourceChannel.close();
            destChannel.close();
            destFile.close();
            source.close();

            sourceChannel = null;
            destChannel = null;
            destFile = null;
            source = null;

            fileState.setOperation(currentLanguage.getString(KeysHolder.SaveOperationEnd));
            fileState.setProgress(100.0f);
            setChanged();
            notifyObservers(fileState);
        }

        System.gc();

        fileState.setOperation(""); //$NON-NLS-1$
        fileState.setProgress(0.0f);
        setIsModified(false);

        if (tempFile != null)
        {
            if (!tempFile.delete())
            {
                System.err.println(
                    currentLanguage.getString(KeysHolder.CannotDeleteFile) + tempFile.toString());
            }
        }

    }

    protected final void setFieldCount(long _fieldCount)
    {
        fieldCount = _fieldCount;
        setIsModified(true);
    }
    public void setFileType(String fileType)
    {
        this.fileType = fileType;
        setIsModified(true);
    }
    protected void setIsModified(boolean modified)
    {
        this.modified = modified;

        fileState.setModified(modified);

        if (modified)
        {
            setChanged();
            notifyObservers(fileState);
        }
        else
        {
            setChanged();
            notifyObservers(fileState);
        }
    }
    /**
     * Removes a structure from the GFF file.
     * 
     * If other structures were refering to that strucutre, the GFF file may not be valid anymore. 
     * All field of the structures are removed from the file.
     * 
     * @param structure To remove from the file.
     */
    public void removeStruct(Structure structure)
    {
        if (structs.contains(structure))
        {

            structs.remove(structure);

            for (int i = 0; i < structure.getFieldCount(); i++)
            {
                removeField(structure.getField(i));
            }
            structCount = structs.size();
            structure.setOwner(null);
        }
    }
    /**
     * Remove a field.
     * 
     * If structures still use this field after the call, the file won't be valid anymore.
     * This operation removes also all unused label.
     *  
     * @param field Field to remove.
     */
    private void removeField(Field field)
    {
        if (fieldList.contains(field))
        {

            Long refCount = (Long) labelRefCount.get(field.getLabel());

            long value = refCount.longValue() - 1;

            if (value == 0)
            {
                labelList.remove(field.getLabel());
                labelCount = labelList.size();

                labelRefCount.remove(field.getLabel());

            }
            else
            {
                labelRefCount.put(field.getLabel(), new Long(value));
            }

            fieldList.remove(field);
            fieldCount = fieldList.size();
        }
    }

} // end GFFFile
